/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "socket.h"
#include "ftpsocket.h"

#include "kftpconfiguration.h"
#include "kftpotpgenerator.h"
#include "kcrcchecker.h"

#include <qapplication.h>
#include <qdir.h>

#include <errno.h>
#include <sys/ioctl.h>

#include <ksocks.h>
#include <ksockaddr.h>
#include <klocale.h>
#include <kio/job.h>

#define TELNET_IAC      255
#define TELNET_IP       244
#define TELNET_SYNCH    242

namespace KFTPNetwork {

FtpSocket::FtpSocket(QObject *parent)
 : Socket(parent, "ftp"),
   m_controlSSL(0),
   m_dataSSL(0),
   m_clientCert(0)
{
}

void FtpSocket::initConfig()
{
  Socket::initConfig();
  
  m_clientCert = 0L;
}

void FtpSocket::get(KURL source, KURL destination)
{
  ftpGet(source.path(), destination.path());
}

void FtpSocket::put(const KURL &source, const KURL &destination)
{
  ftpPut(source.path(), destination.path());
}

void FtpSocket::remove(const KURL &url)
{
  /* Discover if path is file or directory and call the right method */
  if (checkIsDir(url)) {
    setBusy(true, 1);

    recursiveDelete(url);
    ftpRmdir(url.path());

    setBusy(false, 1);
    return;
  }

  ftpDelete(url.path());
}

void FtpSocket::rename(const KURL &source, const KURL &destination)
{
  ftpRename(source.path(), destination.path());
}

void FtpSocket::chmod(const KURL &url, int mode)
{
  ftpChmod(mode, url.path());
}

void FtpSocket::mkdir(const KURL &url)
{
  ftpMkdir(url.path());
}

void FtpSocket::fxpTransfer(const KURL &source, const KURL &destination, Socket *client)
{
  FtpSocket *socket = static_cast<FtpSocket*>(client);
  
  setBusy(true);
  socket->setBusy(true);

  int retr = ftpFxpTransfer(source.path(), destination.path(), socket);

  setBusy(false);
  socket->setBusy(false);

  if (retr < 0) {
    emit sigLogUpdate(4, i18n("FXP transfer failed."));
    emit socket->sigLogUpdate(4, i18n("FXP transfer failed."));
  } else {
    emit sigLogUpdate(3, i18n("FXP transfer completed."));
    emit socket->sigLogUpdate(3, i18n("FXP transfer completed."));
  }
}

void FtpSocket::stat(const KURL &url)
{
  // Get info about a file
  FTPEntry entry;
  bool isDir = checkIsDir(url);

  QString parentDir = url.directory();
  QString listParams;

  if (isDir) {
    // It is a directory
    entry.type = 'd';
    entry.name = url.fileName();
    entry.permissions = "rwxrwxrwx";

    emit sigStatResult(entry);
    return;
  } else {
    // It is a file, go to parent dir and then send list command for
    // only this file
    listParams = url.fileName();
  }

  // FIXME read below
  // If in TLS, use 'C'lear text mode for directory listing, this
  // should be fixed by not using file functions on the data
  // socket.
  ftpChangeTLSDataProt('C');

  if (!ftpDirChange(parentDir, listParams)) {
    ftpChangeTLSDataProt(getConfig("prot_mode"));
    
    // We have failed - signal that as well!
    FTPEntry fail;
    emit sigStatResult(fail);
    return;
  }

  setBusy(true);

  FTPEntry *e;
  while ((e = ftpReadList())) {
    // Parse the whole dir listing
    if (e->name == url.fileName()) {
      entry = *e;
    }

    if (m_abort)
      break;
  }

  (void) ftpDirClose();
  setBusy(false);

  // Revert back to default data mode for TLS
  ftpChangeTLSDataProt(getConfig("prot_mode"));

  // We have all the data, signal that!
  emit sigStatResult(entry);
}

int FtpSocket::getFeatures()
{
  return SF_FXP_TRANSFER | SF_RAW_COMMAND;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////  SSL ENCRYPTION  /////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

KSSL *FtpSocket::SSLNewControl(int fd)
{
  // Create a new KSSL object for the specified file descriptor
  if (m_controlSSL != 0L) {
    m_controlSSL->reInitialize();
  } else {
    m_controlSSL = new KSSL();
    m_controlSSL->TLSInit();
  }
  
  // Use the client certificate if required
  if (m_clientCert != 0L) {
    m_controlSSL->setClientCertificate(m_clientCert);
  }
  
  m_sslLastResult = (m_controlSSL->connect(fd) == 1) ? true : false;
  
  return m_controlSSL;
}

KSSL *FtpSocket::SSLNewData(int fd)
{
  // Create a new KSSL object for the specified file descriptor
  if (m_dataSSL != 0L) {
    m_dataSSL->reInitialize();
  } else {
    m_dataSSL = new KSSL();
    m_dataSSL->TLSInit();
  }
  
  // Use the client certificate if required
  if (m_clientCert != 0L) {
    m_dataSSL->setClientCertificate(m_clientCert);
  }
  
  m_sslLastResult = (m_dataSSL->connect(fd) == 1) ? true : false;
  
  return m_dataSSL;
}

KSSL *FtpSocket::SSLGetObject()
{
  // Return the current control KSSL object (if any)
  return m_controlSSL;
}

void FtpSocket::SSLSetClientCert(KSSLPKCS12 *cert)
{
  m_clientCert = cert;
}

void FtpSocket::SSLClose()
{
  // We won't be using SSL for this connection anymore
  if (m_controlSSL != 0L) {
    m_controlSSL->close();
    
    delete m_controlSSL;
    m_controlSSL = 0L;
  }
  
  if (m_dataSSL != 0L) {
    delete m_dataSSL;
    m_dataSSL = 0L;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////  FTP PROTOCOL  ///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int FtpSocket::ftpReadLine(char *buf, int max)
{
  if (max == 0)
    return 0;

  char read_buf[1024];
  char *c = 0L;
  int read_bytes = -1;
  QCString s_buf;
  QString tmp = "";

  if (m_readBuf.find('\n') == -1) {
    while (!c) {
      int b_read;
      memset(read_buf, 0, sizeof(read_buf));
  
      if (getConfig("using_ssl") == 1) {
        /* We are in TLS mode */
        if ((b_read = m_controlSSL->read(read_buf, sizeof(read_buf))) == -1) {
          qDebug("[ERROR] Read failed (TLS).");
          return -1;
        }
      } else {
        /* Normal cleartext mode */
        if ((b_read = KSocks::self()->recv(m_control, read_buf, sizeof(read_buf), 0)) == -1) {
          qDebug("[ERROR] Read failed.");
          return -1;
        }
        
        // The peer has closed the connection
        if (b_read == 0) {
          qDebug("[ERROR] Peer closed connection.");
          return -1;
        }
      }

      read_bytes += b_read;

      // Search for the \n (if it exists)
      c = strchr(read_buf, '\n');
      s_buf += read_buf;
    }

    tmp = s_buf;
  }

  if (m_readBuf.length() > 0)
    tmp.prepend(m_readBuf);

  int n_pos = tmp.find('\n');
  memset(buf, 0, max);
  strcpy(buf, tmp.mid(0, n_pos+1).ascii());
  read_bytes = QCString(buf).length();

  // Save the remaining bytes
  tmp.remove(0, n_pos+1);
  m_readBuf = tmp;

  return read_bytes;
}

char FtpSocket::ftpReadResponse()
{
  // Read the response into the response buffer and return the first
  // char of the buffer for error code recognition
  int size;
  char match[5];
  
  if ((size = ftpReadLine(m_responseBuf, 256)) == -1) {
    return '\0';
  }

  if (m_responseBuf[3] == '-') {
    // Multiline text
    strncpy(match, m_responseBuf, 3);
    match[3] = ' ';
    match[4] = '\0';

    m_lastMultiResp = "";

    emit sigLogUpdate(2, m_responseBuf);

    while (strncmp(m_responseBuf, match, 4)) {
      // While the response code is the same as above, read the multiline
      // response from server

      if (ftpReadLine(m_responseBuf, 256) == -1) {
        return '\0';
      }

      if (m_responseBuf[3] == '-' || m_responseBuf[0] == ' ') {
        emit sigLogUpdate(2, m_responseBuf);
        m_lastMultiResp += m_responseBuf;
      }
    }
  } else {
    // Single line response
  }

  // For log update
  emit sigLogUpdate(0, m_responseBuf);

  // Return the first character
  return m_responseBuf[0];
}

bool FtpSocket::ftpIsRespCode(const QString &code)
{
  // Get the current response code and compare
  QCString curResp;
  curResp += m_responseBuf[0];
  curResp += m_responseBuf[1];
  curResp += m_responseBuf[2];

  return QCString(code) == curResp;
}

int FtpSocket::ftpSendCommand(const QString &command)
{
  // Send a command to the server, if it fails try to retry
  // or reconnect

  QCString buffer(command + "\r\n");
  char response = '\0';

  // For log update
  emit sigLogUpdate(1, command);

  int retr;

  if (getConfig("using_ssl") == 1) {
    /* We are in TLS mode */
    retr = m_controlSSL->write(buffer.data(), buffer.length());
  } else {
    /* Normal mode */
    retr = KSocks::self()->write(m_control, buffer.data(), buffer.length());
  }

  if (retr > 0)
    response = ftpReadResponse();

  if (response == '\0' || ftpIsRespCode("421")) {
    // If response is NULL(\0) or the response code is 421
    // we are disconnected/timed out
    Socket::disconnect();

    // Reconnect if requested
    if (getConfig("do_retry"))
      Socket::connect(getClientInfoUrl());

    return false;
  }

  return true;
}

bool FtpSocket::ftpCanRead()
{
  // Check if there is anything to read on the control socket
  if (getConfig("using_ssl") == 1) {
    return m_controlSSL->pending() > 0;
  } else {
    int numBytes = 0;
    ioctl(m_control, FIONREAD, &numBytes);
  
    return numBytes > 0;
  }
}

void FtpSocket::ftpChangeTLSDataProt(char prot)
{
  if (getConfig("using_ssl") == 1 && getConfig("prot_mode") != 'C') {
    QString cmd = "PROT ";
    cmd += prot;

    if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
      emit sigLogUpdate(4, i18n("Protection mode setup failed, fallback to unencrypted connection."));
      setConfig("prot_mode", 'C');
    }
  }
}

int FtpSocket::protoConnect()
{
  // Support decimal IPs
  QString addr = getClientInfoUrl().host();
  bool isDecimal;
  unsigned long decIP = addr.toULong(&isDecimal);
  if (isDecimal) {
    // It is a decimal formatted IP, let's convert
    QString hex;
    hex.setNum(decIP, 16);
    addr = "";
    
    addr += QString::number(hex.mid(0, 2).toInt(0, 16));
    addr += ".";
    
    addr += QString::number(hex.mid(2, 2).toInt(0, 16));
    addr += ".";
    
    addr += QString::number(hex.mid(4, 2).toInt(0, 16));
    addr += ".";
    
    addr += QString::number(hex.mid(6, 2).toInt(0, 16));
  }
  
  struct sockaddr_in sa;
  struct hostent *he;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(getClientInfoUrl().port());
  int on = 1;
  
  int ret = inet_aton(addr.ascii(), &sa.sin_addr);
  if (ret == 0) {
    if ((he = ::gethostbyname(addr.ascii())) == NULL) {
      sigLogUpdate(4, i18n("Uknown host '%1'.").arg(addr));
      
      return -2;
    }
    
    memcpy((char *)&sa.sin_addr, he->h_addr, he->h_length);
  }
  
  m_socketAddress = sa;
  m_control = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (m_control < 0) {
    return -1;
  }
  
  if (setsockopt(m_control, SOL_SOCKET, SO_REUSEADDR, (char*) &on, sizeof(on)) == -1) {
    return -1;
  }
  
  if (::connect(m_control, (struct sockaddr*) &sa, sizeof(sa)) < 0) {
    sigLogUpdate(4, i18n("Connection failed (%1).").arg(strerror(errno)));
    
    return -2;
  }
  
  // Clear the buffer
  m_readBuf = "";
  
  // If we are required to use implicit SSL, do so now
  if (getConfig("use_implicit") == 1) {
    SSLNewControl(m_control);
    setConfig("using_ssl", 0);
    setConfig("prot_mode", 'P');

    if (m_sslLastResult != 1) {
      sigLogUpdate(4, i18n("Unable to establish implicit SSL connection!"));
      return -5;
    } else {
      sigLogUpdate(3, i18n("Established implicit SSL connection."));
      setConfig("using_ssl", 1);
      setConfig("use_tls", 0);
    }
  }

  if (ftpReadResponse() != '2') {
    // Response code should be 2xx, but it is NOT
    qDebug("[ERROR] Unable to connect: %s", QString(m_responseBuf + 3).stripWhiteSpace().ascii());

    return -5;
  }

  return 1;
}

int FtpSocket::protoLogin()
{
  if (m_isLoggedIn) return 1;

  // If both is empty, do anonymous
  if (!getClientInfoUrl().hasUser() && !getClientInfoUrl().hasPass()) {
    getClientInfoUrl().setUser("anonymous");
    getClientInfoUrl().setPass(FTPConfig().m_anonMail);
  }

  // Shall we use TLS ?
  if (getConfig("use_tls") == 1) {
    if (ftpSendCommand("AUTH TLS") && m_responseBuf[0] != '2') {
      // Failed, disable TLS usage on this connection
      setConfig("use_tls", 0);
    } else {
      SSLNewControl(m_control);
      setConfig("using_ssl", 0);
      setConfig("prot_mode", 'P');

      if (m_sslLastResult != 1)
        emit sigLogUpdate(4, i18n("Unable to establish TLS connection!"));
      else
        setConfig("using_ssl", 1);
    }
  }

  // Actualy send the login commands
  ftpSendCommand("USER " + getClientInfoUrl().user());

  if (ftpIsRespCode("331")) {
    // Send the password
    if (!strncmp(m_responseBuf, "331 Response to otp-", 20) ||
        !strncmp(m_responseBuf, "331 Response to s/key", 21)) {
      // OTP: 331 Response to otp-md5 41 or4828 ext required for foo.
      QString tmp = m_responseBuf;
      tmp = tmp.section(' ', 3, 5);

      KFTPOTPGenerator g(tmp, getClientInfoUrl().pass());
      ftpSendCommand("PASS " + g.generateOTP());
    } else {
      ftpSendCommand("PASS " + getClientInfoUrl().pass());
    }
  }

  if (!ftpIsRespCode("230")) {
    // Login failed
    return -1;
  }

  /* Set some stuff if in TLS mode */
  if (getConfig("using_ssl") == 1) {
    ftpSendCommand("PBSZ 0");
    ftpChangeTLSDataProt(getConfig("prot_mode"));
  }

  // Now we should be logged in, switch to UNIX-like dir listing if this is
  // IIS 4
  if (ftpSendCommand("SYST") && m_responseBuf[0] == '2') {
    if (!strncmp(m_responseBuf, "215 Windows_NT version", 22)) {
      ftpSendCommand("SITE DIRSTYLE");

      if (!strncmp(m_responseBuf, "200 MSDOS-like directory output is on", 37)) {
        // It was UNIX before, repair what we broke =)
        ftpSendCommand("SITE DIRSTYLE");
      }
    }
  }

  // Send FEAT command to discover extended features ;)
  if (ftpSendCommand("FEAT") && m_responseBuf[0] == '2') {
      QStringList p_featList = QStringList::split("\n", m_lastMultiResp);
      for (QStringList::iterator i = p_featList.begin();i != p_featList.end();++i) {
        QString p_feat = *i;
        p_feat = p_feat.stripWhiteSpace();

        // Check for FEAT output type
        if (p_feat.left(4) == "211-") {
          p_feat.remove(0, 4);
        }

        if (p_feat == "PRET") {
          // This server looks like a distributed FTP server, enable the
          // option if not already enabled so we will send the right commands
          // before transfers.
          setConfig("feat_pret", 1);
        } else if (p_feat == "SSCN") {
          // Server supports SSCN mode, if we are using SSL connection, we should
          // enable support
          setConfig("feat_sscn", 1);
        } else if (p_feat == "CPSV") {
          // Server supports CPSV (similar to SSCN)
          setConfig("feat_cpsv", 1);
        } else if (p_feat.startsWith("XCRC")) {
          // Server supports XCRC checks
          setConfig("feat_xcrc", 1);
        } else if (p_feat.startsWith("SITE")) {
          // Let's see some SITE commands that are supported
          p_feat.remove(0, 5);
          QStringList siteList = QStringList::split(";", p_feat);
          
          for (QStringList::iterator j = siteList.begin(); j != siteList.end(); ++j) {
            QString siteFeat = *j;
            siteFeat = siteFeat.stripWhiteSpace();
            
            if (siteFeat == "SET") {
              // This is a very usefull options, since SITE SET TRANSFERPROGRESS ON may be
              // implemented -- this is needed for showing status of FXP transfers
              setConfig("feat_site_set", 1);
            }
          }
        } else if (p_feat.startsWith("UTF8")) {
          // TODO enable utf8 and send it right after the feat is done
          setConfig("feat_utf8", 1);
        }
      }
  }

  // Check where we are
  if (!ftpSendCommand("PWD") || m_responseBuf[0] != '2') {
    // Ugh ?! Unable to send PWD command... wierd
    return -2;
  }
  
  // 257 "/home/default/path"
  m_defDir = QString(m_responseBuf);
  int first = m_defDir.find('"') + 1;
  m_defDir = m_defDir.mid(first, m_defDir.findRev('"') - first);

  m_isLoggedIn = true;
  return 1;
}

void FtpSocket::protoDisconnect()
{
  /* Free the SSL stuff */
  if (getConfig("using_ssl") == 1) {
    SSLClose();
  }
  
  if (m_control != 0) {
    close(m_control);
    m_control = 0;
  }

  m_isLoggedIn = false;
}

void FtpSocket::ftpCloseDataSockets()
{
  if (m_dataConn != 0) {
    if (getConfig("using_ssl") == 1 && m_dataSSL) {
      m_dataSSL->close();
    } else {
      shutdown(m_dataConn, 2);
      ::close(m_dataConn);
    }
    
    m_dataConn = 0;
  }

  if (m_data != 0) {
    ::close(m_data);
    m_data = 0;
  }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  CONNECTION TYPE METHODS  ///////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool FtpSocket::ftpPASVConnection()
{
  KExtendedSocket kSocket;
  int ip[6];
  struct linger lng = { 1, 120 };

  // Can we actually do PASV - only via IPv4 !
  if (m_socketAddress.sin_family != AF_INET)
    return false;

  // Check if PASV mode is supported
  if (getConfig("can_pasv") == 0)
    return false;

  if (!ftpSendCommand("PASV") || !ftpIsRespCode("227")) {
    if (m_responseBuf[0] == '5') {
      // Unknown command - disable PASV mode
      setConfig("can_pasv", 0);
    }

    return false;
  }

  // Ok PASV command successfull - let's parse the result
  char *begin = strchr(m_responseBuf, '(');

  // Some stinky servers don't respect RFC and do it on their own
  if (!begin)
    begin = strchr(m_responseBuf, '=');

  if (!begin || (sscanf(begin, "(%d,%d,%d,%d,%d,%d)",&ip[0], &ip[1], &ip[2], &ip[3], &ip[4], &ip[5]) != 6 &&
                 sscanf(begin, "=%d,%d,%d,%d,%d,%d",&ip[0], &ip[1], &ip[2], &ip[3], &ip[4], &ip[5]) != 6)) {
    // Unable to parse ?
    return false;
  }

  // Convert to string
  QString host;
  int port;

  host.sprintf("%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
  port = ip[4] << 8 | ip[5];

  // If the reported IP address is from a private IP range, this might be because the
  // remote server is not properly configured. So we just use the server's real IP instead
  // of the one we got (if the host is really local, then this should work as well).
  if (host.startsWith("192.168.") || host.startsWith("10.") || host.startsWith("172.16."))
    kSocket.setAddress(inet_ntoa(m_socketAddress.sin_addr), port);
  else
    kSocket.setAddress(host, port);
    
  kSocket.setSocketFlags(KExtendedSocket::noResolve);
  kSocket.setTimeout(10);

  if (kSocket.connect() < 0) {
    return false;
  }

  // Get the socket's file descriptor and set some opts
  int on = 1;
  m_data = kSocket.fd();

  if ((setsockopt(m_data, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)) == -1) || (m_data < 0)) {
    return false;
  }
  if ((setsockopt(m_data, SOL_SOCKET, SO_KEEPALIVE, (char*)&on, sizeof(on)) < 0)) {
    qDebug("Keepalive not allowed");
  }
  if (setsockopt(m_data, SOL_SOCKET, SO_LINGER, (char*)&lng, sizeof(lng)) < 0) {
    qDebug("Unable to set linger mode");
  }

  kSocket.release();
  return true;
}

bool FtpSocket::ftpEPSVConnection()
{
  KExtendedSocket kSocket;
  struct linger lng = { 1, 120 };

  // Check if EPSV mode is supported
  if (getConfig("can_epsv") == 0)
    return false;

  if (!ftpSendCommand("EPSV") || m_responseBuf[0] != '2') {
    if (m_responseBuf[0] == '5') {
      // Unknown command - disable EPSV mode
      setConfig("can_epsv", 0);
    }

    return false;
  }

  // 229 Entering Extended Passive Mode (|||55016|)
  char *begin = strchr(m_responseBuf, '(');
  int port;

  if (!begin || sscanf(begin, "(|||%d|)", &port) != 1) {
    // Unable to parse
    return false;
  }

  kSocket.setAddress(inet_ntoa(m_socketAddress.sin_addr), port);
  kSocket.setSocketFlags(KExtendedSocket::noResolve);
  kSocket.setTimeout(10);

  if (kSocket.connect() < 0) {
    return false;
  }

  // Get the socket's file descriptor and set some opts
  int on = 1;
  m_data = kSocket.fd();

  if ((setsockopt(m_data, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)) == -1) || (m_data < 0)) {
    return false;
  }
  if (setsockopt(m_data, SOL_SOCKET, SO_KEEPALIVE, (char*)&on, sizeof(on)) < 0) {
    qDebug("Keepalive not allowed");
  }
  if (setsockopt(m_data, SOL_SOCKET, SO_LINGER, (char*)&lng, sizeof(lng)) < 0) {
    qDebug("Unable to set linger mode");
  }

  kSocket.release();
  return true;
}

bool FtpSocket::ftpEPRTConnection()
{
  // TODO implement EPRT connections ;)
  //        EPRT |1|132.235.1.2|6275|
  //        EPRT |2|1080::8:800:200C:417A|5282|
  return false;
}

bool FtpSocket::ftpActiveConnection()
{
  union
  {
    struct sockaddr sa;
    struct sockaddr_in in;
  } sin;

  struct linger lng = { 0, 0 };
  ksocklen_t l;
  char buf[64];
  int on = 1;

  // Get control's ip
  l = sizeof(sin);
  if (KSocks::self()->getsockname(m_control, &sin.sa, &l) < 0)
    return false;

  // PORT is only for IPv4 connections
  if (sin.sa.sa_family != PF_INET)
    return false;

  m_data = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (m_data == 0)
    return false;

  if (setsockopt(m_data, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)) == -1) {
    ::close(m_data);
    return false;
  }

  if (setsockopt(m_data, SOL_SOCKET, SO_LINGER, (char*)&lng, sizeof(lng)) == -1) {
    ::close(m_data);
    return false;
  }

  sin.in.sin_port = 0;
  if (KSocks::self()->bind(m_data, &sin.sa, sizeof(sin)) == -1) {
    ::close(m_data);
    m_data = 0;
    return false;
  }

  if (KSocks::self()->listen(m_data, 1) < 0) {
    ::close(m_data);
    return false;
  }

  if (KSocks::self()->getsockname(m_data, &sin.sa, &l) < 0)
    return false;

  // If the user has chosen to set a different ip
  if (FTPConfig().m_portForceIp) {
    // Get the port we used when binding
    KSocketAddress *p_tmp = KSocketAddress::newAddress(&sin.sa, l);

    // Resolve IP first - using the port above
    int error;
    QPtrList<KAddressInfo> p_list;

    p_list = KExtendedSocket::lookup(FTPConfig().m_portIp, p_tmp->serviceName(), 0, &error);
    p_list.setAutoDelete(true);
    if (error < 0) {
      // Well, we are unable to resolve the name, so we should use what we got
      // from control socket
    } else {
      // The name has been resolved and we have the address, so we should
      // use it
      sin.sa = *p_list.at(0)->address()->address();
      l = p_list.at(0)->address()->size();
    }

    delete p_tmp;
  }

  // Send the apropriate PORT command
  sprintf(buf,"PORT %d,%d,%d,%d,%d,%d",
          (unsigned char)sin.sa.sa_data[2],(unsigned char)sin.sa.sa_data[3],
          (unsigned char)sin.sa.sa_data[4],(unsigned char)sin.sa.sa_data[5],
          (unsigned char)sin.sa.sa_data[0],(unsigned char)sin.sa.sa_data[1]);

  if (!ftpSendCommand(buf) || m_responseBuf[0] != '2') {
    ::close(m_data);
    return false;
  }

  return true;
}

bool FtpSocket::ftpInitDataConnection()
{
  // Init a data connection
  bool result = false;

  if (getConfig("epsv_mode") == 1) {
    // We are already in EPSV mode, so the only thing we can init
    // is EPSV data connection.
    result = ftpEPSVConnection();
  } else {
    // Otherwise go one by one
    if (getConfig("can_epsv") == 1)
      result = ftpEPSVConnection();

    if (!result && getConfig("can_pasv") == 1)
      result = ftpPASVConnection();

    if (!result && getConfig("can_eprt") == 1)
      result = ftpEPRTConnection();

    if (result)
      setConfig("passive_mode", 1);

    if (!result) {
      result = ftpActiveConnection();

      if (result)
        setConfig("passive_mode", 0);
    }
  }

  return result;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  DATA CONNECTION METHODS  ///////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

size_t FtpSocket::ftpDataRead(void *buf, long size)
{
  /* Use SSL if in TLS mode and PROTO is not cleartext */
  if (getConfig("using_ssl") == 1 && getConfig("prot_mode") != 'C') {
    return m_dataSSL->read(buf, size);
  } else {
    return KSocks::self()->read(m_dataConn, buf, size);
  }

  return 0;
}

size_t FtpSocket::ftpDataWrite(void *buf, long size)
{
  /* Use SSL if in TLS mode and PROTO is not cleartext */
  if (getConfig("using_ssl") == 1 && getConfig("prot_mode") != 'C') {
    return m_dataSSL->write(buf, size);
  } else {
    return KSocks::self()->write(m_dataConn, buf, size);
  }

  return 0;
}

int FtpSocket::ftpPrepareTLS(int fd)
{
  if (getConfig("using_ssl") != 1)
    return fd;

  m_dataSSL = SSLNewData(fd);
  if (m_sslLastResult != 1) {
    ::close(fd);
    return -1;
  }

  return fd;
}

int FtpSocket::ftpInitDataSocket(bool noTLS)
{
  int data_sock;
  struct sockaddr addr;
  ksocklen_t l;
  fd_set mask;

  FD_ZERO(&mask);
  FD_SET(m_data, &mask);

  // If passive socket, return the currently connected socket
  if (getConfig("passive_mode") == 1)
    return noTLS ? m_data : ftpPrepareTLS(m_data);

  if (KSocks::self()->select(m_data + 1, &mask, NULL, NULL, 0L) == 0) {
    ::close(m_data);
    return -1;
  }

  l = sizeof(addr);
  if ((data_sock = KSocks::self()->accept(m_data, &addr, &l)) > 0)
    return noTLS ? data_sock : ftpPrepareTLS(data_sock);

  ::close(m_data);
  return -1;
}

bool FtpSocket::ftpDataCommand(const QString &command, const QString &path, char mode, bool noTLS,
                                    filesize_t offset)
{
  if (!m_isLoggedIn)
    return false;

  // If the protection is set to 'C'-lear then we shouldn't use TLS
  if (getConfig("prot_mode") == 'C')
    noTLS = true;
    
  // First, set the transfer mode
  QString cmd = "TYPE ";
  cmd += mode;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    return false;
  }

  // If in distributed FTPD mode, we must first send PRET <command> so the
  // server will know where to send us.
  if (getConfig("feat_pret") == 1) {
    QString p_cmd = "PRET ";
    p_cmd += command;
    if (!path.isEmpty()) {
      p_cmd += " ";
      p_cmd += path;
    }

    if (!ftpSendCommand(p_cmd) || m_responseBuf[0] != '2') {
      // PRET not supported or just permission/fs problems - have to check
      if (ftpIsRespCode("530") || ftpIsRespCode("550")) {
        // yep, the file could not be fetched, return false and don't even
        // try to init the transfer
        return false;
      }

      setConfig("feat_pret", 0);
    }
  }

  // Now try to init a data connection
  if (!ftpInitDataConnection()) {
    qDebug("Init failed.");
    return false;
  }

  // Check if we have an offset (for RETR/STOR)
  if (offset > 0) {
    char tmp[100];
    sprintf(tmp, "REST %ld", offset);

    if (!ftpSendCommand(tmp) || m_responseBuf[0] != '3') {
      // Disable resuming
      setConfig("feat_resume", 0);
    }
  }

  // If path is set, add it after the command
  QString p_command = command;
  if (!path.isEmpty()) {
    p_command += " ";
    p_command += path;
  }

  if (!ftpSendCommand(p_command) || m_responseBuf[0] != '1') {
    return false;
  }

  if ((m_dataConn = ftpInitDataSocket(noTLS)) < 0) {
    return false;
  }

  return true;
}

bool FtpSocket::ftpCloseDataCommand()
{
  ftpCloseDataSockets();

  if (ftpReadResponse() != '2') {
    // No transfer complete msg ?!
    return false;
  }

  return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  DIR LIST PARSING METHODS  //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool FtpSocket::protoDirList(const KURL &url)
{
  if (!m_isLoggedIn) {
    // We are not logged in
    return false;
  }

  QString path = url.path();

  // FIXME read below
  // If in TLS, use 'C'lear text mode for directory listing, this
  // should be fixed by not using file functions on the data
  // socket.
  ftpChangeTLSDataProt('C');

  if (!ftpDirChange(path)) {
    // If we are still logged in, change transfer mode back
    if (m_isLoggedIn)
      ftpChangeTLSDataProt(getConfig("prot_mode"));
    return false;
  }

  setBusy(true);
  m_lastDirList.clear();

  FTPEntry *entry;
  while ((entry = ftpReadList())) {
    // Parse the whole dir listing
    m_lastDirList.append(FTPDirectoryItem(*entry));

    if (m_abort)
      break;
  }

  (void) ftpDirClose();
  setBusy(false);

  // Revert back to default data mode for TLS
  ftpChangeTLSDataProt(getConfig("prot_mode"));

  sigLogUpdate(3, i18n("Directory listing complete."));
  return true;
}

bool FtpSocket::ftpDirChange(const QString &path, const QString &params)
{
  if (!m_isLoggedIn)
    return false;

  if (!ftpCwd(path)) {
    // Error while accessing the directory
    return false;
  }

  // TODO: use a config entry to specify a custom list command
  if (!ftpDataCommand("LIST -a", params, 'A', true)) {
    // List command & data connection failed
    return false;
  }

  m_list = fdopen(m_dataConn, "r");
  if (!m_list)
    return false;

  return true;
}

bool FtpSocket::ftpDirClose()
{
  if (m_list) {
    if (!ftpCloseDataCommand())
      return false;

    fclose(m_list);
    m_list = 0L;
  }

  return true;
}

FTPEntry *FtpSocket::ftpReadList()
{
  char buffer[1024];

  while (fgets(buffer, sizeof(buffer), m_list) != 0) {
    FTPEntry *entry = ftpParseDirListing(buffer);
    if (entry)
      return entry;
  }

  return 0L;
}

FTPEntry *FtpSocket::ftpParseDirListing(char *raw)
{
  /*
    PERMISIONS  ??? USER     (GROUP) (DEV)SIZE     DATE     FILENAME
    total 123456
    drwxr-xr-x    3 kostko   users        4096 Jul 16 13:46 test
    -rw-r--r--    1 kostko   users           0 Jul 17 10:41 test.txt
    crw-rw-rw-    1 root     root       1,   5 Jan  1  1970 zero
    drwxr-xr-x    2 kostko   kostko       4096 Jan 19 19:32 internet stuff
    lrwxr-xr-x    3 kostko   users        4096 Jul 16 13:46 test -> /nekje
    drwxr-xr-x    3 kostko                4096 Jul 16 2001 krneki
    ----10----  -?- ---?---  ---?---  ----?--- ------?---- --toend>-------
  */

  static FTPEntry entry;
  QString line(raw);

  if (line.left(5) == "total")
    return 0L;

  /* Get permissions and type */
  entry.type = line.at(0).latin1() == '-' ? 'f' : line.at(0).latin1();
  entry.permissions = line.mid(1, 9);

  if (entry.type == 'c') {
    qDebug("[WARNING] Entry type 'c' - device, not implemented! Skipping...");
    return 0L;
  }

  /* Get owner */
  line.remove(0, 10);
  line = line.stripWhiteSpace();
  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  // "kostko   users        4096 Jul 16 13:46 test"
  // "kostko                4096 Jul 16 13:46 test"
  entry.owner = line.mid(0, line.find(' ')).latin1();
  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  // "users        4096 Jul 16 13:46 test"
  // "4096 Jul 16 13:46 test"

  /* Some servers don't send group, so there is no group, but we can determine
     that when we parse the date */

  QString unknown_1 = line.mid(0, line.find(' '));
  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  // "4096 Jul 16 13:46 test"
  // "Jul 16 13:46 test"
  QString unknown_2 = line.mid(0, line.find(' ')); // size or month
  bool isdigit;
  unknown_2.toInt(&isdigit);

  if (isdigit) {
    // This is size, so unknown_1 was group
    entry.group = unknown_1.latin1();
    entry.size = unknown_2.toInt();

    line.remove(0, line.find(' ') + 1);
    line = line.stripWhiteSpace();
  } else {
    // This is month, so unknown_1 was size
    entry.size = unknown_1.toInt();
  }

  time_t currentTime = time(0L);
  struct tm *tmp = gmtime(&currentTime);
  int currentMonth = tmp->tm_mon;

  // Reset time fields
  tmp->tm_sec = 0;
  tmp->tm_min = 0;
  tmp->tm_hour = 0;

  /* Parse month */
  // "Jul 16 13:46 test"
  // "Jul 16 13:46 test -> /dev/null"
  static const char *months[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
  QString month = line.mid(0, line.find(' '));
  for (register int i = 0; i < 12; i++) {
    if (strcmp(months[i], month.ascii()) == 0) {
      tmp->tm_mon = i;
      break;
    }
  }

  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  /* Parse day */
  // 4 16:49 krneki.c
  // 28  2003 lftp-2.6.5
  tmp->tm_mday = line.mid(0, line.find(' ')).toInt();
  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  /* Parse time or year */
  // 16:49 krneki.c
  // 2003 lftp-2.6.5
  QString unknown_3 = line.mid(0, line.find(' '));
  unknown_3.toInt(&isdigit);

  if (isdigit && unknown_3.length() == 4) {
    // Year
    tmp->tm_year = unknown_3.toInt() - 1900;
  } else {
    // Hour (hh:mm)
    tmp->tm_hour = unknown_3.section(':', 0, 0).toInt();
    tmp->tm_min = unknown_3.section(':', 1, 1).toInt();

    if (tmp->tm_mon > currentMonth + 1)
      tmp->tm_year--;
  }

  entry.date = mktime(tmp);

  line.remove(0, line.find(' ') + 1);
  line = line.stripWhiteSpace();

  /* Parse filename */
  // loleq
  // krneki -> /dev/null

  if (entry.type == 'l') {
    // Symlink
    entry.name = line.mid(0, line.findRev("->")).stripWhiteSpace();
    entry.link = line.mid(line.findRev("->")+3).stripWhiteSpace();
  } else {
    // File or directory
    entry.name = line.stripWhiteSpace();
    entry.link = QString::null;
  }

  /* Do not output . or .. */
  if (entry.name == "." || entry.name == "..")
    return 0L;

  return &entry;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////  FTP COMMAND IMPLEMENTATION  ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int FtpSocket::ftpCwd(const QString &dir)
{
  return (ftpSendCommand(QString("CWD %1").arg(dir)) && m_responseBuf[0] == '2') ? 1 : 0;
}

int FtpSocket::ftpSize(const QString &file, char mode)
{
  // Set the datatype
  QString cmd = "TYPE ";
  cmd += mode;
  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2')
    return -1;

  cmd = "SIZE ";
  cmd += file;
  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2')
    return -1;

  // Return the size
  m_lastSize = atol(m_responseBuf + 4);
  return 1;
}

int FtpSocket::ftpGet(const QString &source, const QString &destination)
{
  FILE *f;
  filesize_t offset = 0;
  char mode = 'I';
  QDir fs;

  /* Get the mode for the desired file (ascii/binary) */
  mode = FTPConfig().ftpMode(source);

  /* Check the sizes (and if the file exists) */
  filesize_t size_local = 0;
  if (ftpSize(source, mode) == -1) {
    // File does not exist or it is a directory
    return -1;
  }

  filesize_t size_remote = m_lastSize;

  /* Does destination (local) file exist ? */
  if (fs.exists(destination) && getConfig("feat_resume") == 1) {
    // Check if we can resume
    f = fopen(destination, "a");
    if (f != NULL) {
      if (fseek(f, 0, SEEK_END) != 0) {
        fclose(f);
        return -1;
      }
      
      size_local = ftell(f);
  
      if (size_local < size_remote && size_local > 0) {
        offset = size_local;
      }
    } else {
      return -1;
    }
  } else {
    // Check if the file path exists, if not, create it
    QString dest_dir = fs.cleanDirPath(destination.mid(0, destination.findRev('/')));

    // Is the path available ?
    if (!fs.exists(dest_dir)) {
      // Create all dirs
      QString full_path;
      for(register int i = 1;i<=dest_dir.contains('/');i++) {
        full_path += "/" + dest_dir.section('/', i, i);
        if (!fs.exists(full_path)) {
          fs.mkdir(full_path);
        }
      }
    }

    // Create the file
    f = fopen(destination, "w");
    
    if (f == NULL) {
      return -1;
    }
  }

  if (size_remote == size_local && getConfig("feat_resume") == 1) {
    /* File is already downloaded - skip it if enabled */
    fclose(f);
    processedSize(size_local);

    return 1;
  }

  // Get the file's date-time via MDTM
  time_t p_modtime = 0;
  if (getConfig("feat_mdtm") == 1) {
    QString cmd = "MDTM ";
    cmd += source;

    if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
      setConfig("feat_mdtm", 0);
    } else {
      // 213 20031222105809
      // 213 YYYYMMDDhhmmss
      time_t curtime = time(0L);
      struct tm *p_dt = gmtime(&curtime);
      QString tmp(m_responseBuf);

      tmp.remove(0, 4);
      p_dt->tm_year = tmp.left(4).toInt() - 1900;
      p_dt->tm_mon = tmp.mid(4, 2).toInt();
      p_dt->tm_mday = tmp.mid(6, 2).toInt();
      p_dt->tm_hour = tmp.mid(8, 2).toInt();
      p_dt->tm_min = tmp.mid(10, 2).toInt();
      p_dt->tm_sec = tmp.mid(12, 2).toInt();
      p_modtime = mktime(p_dt);
    }
  }

  // Init data conn
  if (!ftpDataCommand("RETR", source, mode, false, offset)) {
    return -1;
  }

  // Open the file and grab the data
  // TODO implement variable buffer size
  
  char buffer[4096];
  filesize_t bytes_receieved = offset;

  setBusy(true);
  setOffset(offset);
  emit sigResumedOffset(offset);

  while (int b_read = ftpDataRead(buffer, sizeof(buffer))) {
    fwrite(buffer, b_read, 1, f);
    bytes_receieved += b_read;
    
    if (b_read == 0 && getSpeed() == 0) {
      // We are stalled...
      // FIXME should we abort transfer after timeout ?
    }

    processedSize(b_read);

    if (m_abort)
      break;
  }
  fflush(f);
  fclose(f);
  
  setBusy(false);

  // Set the modification date we got at MDTM
  if (getConfig("feat_mdtm") == 1) {
    // Access time is actualy this moment ;) and modification time
    // should be set to what we got when doing MDTM

    utimbuf p_timbuf;
    p_timbuf.actime = time(0L);
    p_timbuf.modtime = p_modtime;
    utime(destination.latin1(), &p_timbuf);
  }
  
  if (!ftpCloseDataCommand()) {
    /* An error while closing ? */
  }
  
  if (bytes_receieved != size_remote) {
    /* Incomplete file download */
    qDebug("[WARNING] File download is incomplete");
  } else {
    ftpCheckCRC(source, destination);
  }

  emit sigLogUpdate(3, i18n("Transfered %1 bytes.").arg(bytes_receieved));

  return 1;
}

int FtpSocket::ftpPut(const QString &source, const QString &destination)
{
  int fd;
  filesize_t offset = 0;
  char mode = 'I';
  QDir fs;

  /* Get the mode for the desired file (ascii/binary) */
  mode = FTPConfig().ftpMode(source);

  /* Check the sizes (and if the file exists) */
  filesize_t size_local;
  filesize_t size_remote = 0;

  if (!fs.exists(source)) {
    return -1;
  } else {
    /* Get local filesize */
    fd = open(source, O_RDONLY);
    
    if (fd < 0) {
      return -1;
    } else {
      size_local = lseek(fd, 0, SEEK_END);
      lseek(fd, 0, SEEK_SET);
    }
  }

  /* Does destination (remote) file exist ? */
  if (ftpSize(destination, mode) == 1 && getConfig("feat_resume") == 1) {
    // Check if we can resume
    size_remote = m_lastSize;

    if (size_remote < size_local) {
      offset = size_remote;
    }
  }

  if (size_remote == size_local && getConfig("feat_resume") == 1) {
    /* File is already on the server - skip it if enabled */
    close(fd);
    processedSize(size_remote);
    return 1;
  }
  
  // Does the destination directory exist ?
  if (size_remote == 0 && !ftpCwd(KURL(destination).directory())) {
    // It doesn't - create it
    QString dest_dir = KURL(destination).directory();
    QString full_path;
    for(register int i = 1; i <= dest_dir.contains('/'); i++) {
      full_path += "/" + dest_dir.section('/', i, i);
      
      // Create the directory
      ftpSendCommand(QString("MKD %1").arg(full_path));
    }
  }

  // Init data conn
  if (!ftpDataCommand("STOR", destination, mode, false, offset)) {
    return -1;
  }

  // Open the file and grab the data
  char buffer[4096];
  filesize_t bytes_sent = offset;

  setBusy(true);
  setOffset(offset);
  emit sigResumedOffset(offset);

  while (int b_read = read(fd, buffer, 4096)) {
    ftpDataWrite(buffer, b_read);
    bytes_sent += b_read;

    processedSize(b_read);

    if (m_abort)
      break;
  }
  close(fd);

  setBusy(false);

  if (!ftpCloseDataCommand()) {
    /* An error while closing ? */
  }
  
  if (bytes_sent != size_local) {
    /* Incomplete file upload */
    qDebug("[WARNING] File upload is incomplete");
  } else {
    ftpCheckCRC(destination, source);
  }
  
  emit sigLogUpdate(3, i18n("Transfered %1 bytes.").arg(bytes_sent));

  return 1;
}

int FtpSocket::ftpRename(const QString &source, const QString &destination)
{
  QString cmd = "RNFR ";
  cmd += source;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '3') {
    /* File does not exist */
    return -1;
  }

  cmd.sprintf("RNTO %s", destination.ascii());

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    /* Permission denied ? */
    return -1;
  }

  return 1;
}

int FtpSocket::ftpDelete(const QString &path)
{
  QString cmd = "DELE ";
  cmd += path;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    /* Unable to delete */
    return -1;
  }

  return 1;
}

int FtpSocket::ftpChmod(int mode, const QString &path)
{
  QString cmd;
  cmd.sprintf("SITE CHMOD %.3d %s", mode, path.ascii());

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    /* Unable to chmod */
    return -1;
  }

  return 1;
}

int FtpSocket::ftpMkdir(const QString &dir)
{
  QString cmd = "MKD ";
  cmd += dir;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    /* Unable to create directory */
    return -1;
  }

  return 1;
}

int FtpSocket::ftpRmdir(const QString &dir)
{
  QString cmd = "RMD ";
  cmd += dir;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    /* Unable to remove directory */
    return -1;
  }

  return 1;
}

int FtpSocket::ftpFxpTransfer(const QString &source, const QString &destination, FtpSocket *client)
{
  int errCode = 0;
  int pos;
  QString cmd;
  QString ip_txt;
  
  if (m_socketAddress.sin_family != client->m_socketAddress.sin_family) {
    // Different families, will fail
    return -11;
  }

  if (   (getConfig("using_ssl") == 1 || client->getConfig("using_ssl") == 1) 
      && (getConfig("feat_sscn") == 0 || client->getConfig("feat_sscn") == 0)
      && (getConfig("feat_cpsv") == 0 || client->getConfig("feat_cpsv") == 0) ) {
    // We are in SSL mode and we don't have SSCN/CPSV support - SSL transfer is not possible
    return -12;
  }
  
  bool sentSSCN = false;
  bool sendCPSV = false;
  
  if (getConfig("using_ssl") == 1 && client->getConfig("using_ssl") == 1) {
    if (getConfig("feat_sscn") == 1 && client->getConfig("feat_sscn") == 1) {
      // Ok, both servers support SSCN (aka new extension to do FXP transfers over SSL), send the
      // proper command to enable SSCN
      if (!ftpSendCommand("SSCN ON") || m_responseBuf[0] != '2')
        return -1;
        
      if (!client->ftpSendCommand("SSCN ON") || client->m_responseBuf[0] != '2')
        return -1;
        
      sentSSCN = true;
    } else if (getConfig("feat_cpsv") == 1 && client->getConfig("feat_cpsv") == 1) {
      // Ok, server support CPSV, send it instead of PASV
      sendCPSV = true;
    }
  }

  // Do a FXP transfer - source is always this client
  char mode = FTPConfig().ftpMode(source);

  filesize_t offset = 0;
  filesize_t srcSize = 0;
  filesize_t dstSize = 0;

  // Does the source file exist ?
  if (ftpSize(source, mode) == 1) {
    srcSize = m_lastSize;
  }

  if (client->ftpSize(destination, mode) == 1 && getConfig("feat_resume") == 1) {
    // Check if we can resume
    dstSize = client->m_lastSize;

    if (dstSize < srcSize) {
      offset = dstSize;
    }
  }
  
  // Does the destination directory exist ?
  if (dstSize == 0 && !client->ftpCwd(KURL(destination).directory())) {
    // It doesn't - create it
    QString dest_dir = KURL(destination).directory();
    QString full_path;
    for(register int i = 1; i <= dest_dir.contains('/'); i++) {
      full_path += "/" + dest_dir.section('/', i, i);
      
      // Create the directory
      client->ftpSendCommand(QString("MKD %1").arg(full_path));
    }
  }

  if (dstSize == srcSize && getConfig("feat_resume") == 1) {
    errCode = 1;
    goto abort;
  }

  if (offset > 0) {
    QString tmp;
    tmp.sprintf("REST %ld", offset);

    if (!ftpSendCommand(tmp) || m_responseBuf[0] != '3') {
      errCode = -1;
      goto abort;
    }
    
    if (!client->ftpSendCommand(tmp) || client->m_responseBuf[0] != '3') {
      errCode = -1;
      goto abort;
    }
  }

  qDebug("[FXP] Initing both FTP servers...");
  // Set TYPE for both clients
  cmd = "TYPE ";
  cmd += mode;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    errCode = -1;
    goto abort;
  }

  if (!client->ftpSendCommand(cmd) || client->m_responseBuf[0] != '2') {
    errCode = -1;
    goto abort;
  }

  // Init the PASV mode
  if (sendCPSV)
    cmd = "CPSV";
  else
    cmd = "PASV";
  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '2') {
    errCode = -2;
    goto abort;
  }

  ip_txt = m_responseBuf;
  pos = ip_txt.find('(')+1;
  ip_txt = ip_txt.mid(pos, ip_txt.find(')') - pos);

  // Init the PORT mode
  cmd = "PORT ";
  cmd += ip_txt;

  if (!client->ftpSendCommand(cmd) || client->m_responseBuf[0] != '2') {
    errCode = -2;
    goto abort;
  }

  qDebug("[FXP] Starting file transfer...");
  // STOR on destination, RETR on source
  cmd = "STOR ";
  cmd += destination;

  if (!client->ftpSendCommand(cmd) || client->m_responseBuf[0] != '1') {
    errCode = -3;
    goto abort;
  }

  cmd = "RETR ";
  cmd += source;

  if (!ftpSendCommand(cmd) || m_responseBuf[0] != '1') {
    errCode = -3;
    goto abort;
  }

  qDebug("[FXP] Waiting for file transfer to complete...");

  // Now wait, until the transfer is complete
  while (1) {
    // Check the server's response every 0.5 sec
    ::usleep(500000);
    
    // If there is anything to read - read it!
    if (ftpCanRead() || client->ftpCanRead()) {
      if (ftpReadResponse() != '\0') {
        // Read the last response from the other
        client->ftpReadResponse();
        break;
      } else if (client->ftpReadResponse() != '\0') {
        // Read the last response from this one
        ftpReadResponse();
        break;
      }
    }
    
    // Check for abortion
    if (m_abort) {
      char msg[4] = {TELNET_IAC, TELNET_IP, TELNET_IAC, TELNET_SYNCH};
      int fl = fcntl(m_control, F_GETFL);
      fcntl(m_control, F_SETFL, fl &~O_NONBLOCK);
      
      if (getConfig("using_ssl") == 1) {
        m_controlSSL->write(msg, sizeof(msg));
      } else {
        send(m_control, msg, 1, MSG_OOB);
        send(m_control, msg + 1, sizeof(msg) - 1, 0);
      }
      
      fcntl(m_control, F_SETFL, fl);
      ftpSendCommand("ABOR");
      ftpReadResponse();
      break;
    }
  }

  if (m_abort) {
    errCode = -4;
    goto abort;
  }

  // Finished ? Check the responses...
  if (m_responseBuf[0] != '2' || client->m_responseBuf[0] != '2') {
    errCode = -5;
    goto abort;
  }

  // No errors :)
  errCode = 1;

abort:
  if (errCode < 0) {
    // Abort transfer
    qDebug("[FXP] Transfer has negative return code.");
    if (client->ftpCanRead())
      client->ftpReadResponse();
      
    if (ftpCanRead())
      ftpReadResponse();
  }
  
  // Disable SSCN if it was activated
  if (sentSSCN) {
    ftpSendCommand("SSCN OFF");
    client->ftpSendCommand("SSCN OFF");
  }
  
  return errCode;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////  AUXILARY METHODS ///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool FtpSocket::ftpCheckCRC(const QString &remote, const QString &local)
{
  // Check hash if supported
  if (getConfig("feat_xcrc") == 1 && ftpSendCommand(QString("XCRC %1").arg(remote)) && m_responseBuf[0] == '2') {
    // 250 361F12...
    long l_fileCRC = KCRCChecker::getFileCRC(local);
    QString fileCRC;
    fileCRC.sprintf("%08lX", l_fileCRC);
    
    QString serverCRC(m_responseBuf);
    serverCRC.remove(0, 4);
    serverCRC = serverCRC.stripWhiteSpace();
    
    if (fileCRC == serverCRC) {
      sigLogUpdate(3, i18n("CRC check matches - file integrity was verified."));
      return true;
    } else {
      sigLogUpdate(4, i18n("Warning: CRC check failed. Downloaded file may be corrupted."));
      return false;
    }
  }
  
  return false;
}

}
#include "ftpsocket.moc"
