/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "sftpsocket.h"

#include <qdir.h>

#include <klocale.h>

namespace KFTPNetwork {

SftpSocket::SftpSocket(QObject *parent)
 : Socket(parent, "sftp"), m_curDir(0)
{
}

int SftpSocket::getFeatures()
{
  return 0;
}

void SftpSocket::initConfig()
{
  Socket::initConfig();
}

void SftpSocket::get(KURL source, KURL destination)
{
  protoGet(source.path(), destination.path());
}

void SftpSocket::put(const KURL &source, const KURL &destination)
{
  protoPut(source.path(), destination.path());
}

void SftpSocket::remove(const KURL &url)
{
  /* Discover if path is file or directory and call the right method */
  if (checkIsDir(url)) {
    setBusy(true, 1);

    recursiveDelete(url);
    protoRmdir(url.path());

    setBusy(false, 1);
    return;
  }

  protoDelete(url.path());
}

void SftpSocket::rename(const KURL &source, const KURL &destination)
{
  protoRename(source.path(), destination.path());
}

void SftpSocket::chmod(const KURL &url, int mode)
{
}

void SftpSocket::mkdir(const KURL &url)
{
  protoMkdir(url.path());
}

void SftpSocket::stat(const KURL &url)
{
  FTPEntry entry;
  
  if (!protoDirList(url.upURL())) {
    // We have failed - signal that as well!
    FTPEntry fail;
    emit sigStatResult(fail);
    
    return;
  }
  
  for (FTPDirList::iterator i = m_lastDirList.begin(); i != m_lastDirList.end(); ++i) {
    if ((*i).name() == url.fileName()) {
      entry = (*i).m_ftpEntry;
      break;
    }
  }

  // We have all the data, signal that!
  emit sigStatResult(entry);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////  SFTP PROTOCOL  //////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int SftpSocket::protoConnect()
{
  // Set connection info
  m_sshOptions = options_new();
  options_set_username(m_sshOptions, (char*) getClientInfoUrl().user().ascii());
  options_set_host(m_sshOptions, getClientInfoUrl().host().ascii());
  options_set_port(m_sshOptions, getClientInfoUrl().port());
  
  // Connect to the remote host
  m_sftpSession = 0;
  m_sshSession = ssh_connect(m_sshOptions);
  
  if (!m_sshSession) {
    emit sigLogUpdate(4, i18n("Unable to establish SSH connection (%1)").arg(ssh_get_error(0)));
    return -1;
  }
  
  return 1;
}

void SftpSocket::protoDisconnect()
{
  if (m_curDir)
    protoCloseDir();
    
  if (m_sftpSession)
    sftp_free(m_sftpSession);
    
  ssh_disconnect(m_sshSession);
  m_sshSession = 0;
  
  m_isLoggedIn = false;
}

int SftpSocket::protoLogin()
{
  if (m_isLoggedIn) return 1;
  
  if (ssh_userauth_password(m_sshSession, NULL, (char*) getClientInfoUrl().pass().ascii()) != AUTH_SUCCESS) {
    return -1;
  }
  
  m_sftpSession = sftp_new(m_sshSession);
  if (!m_sftpSession) {
    emit sigLogUpdate(4, i18n("Unable to init SFTP channel!"));
    return -1;
  }
  
  if (sftp_init(m_sftpSession)) {
    emit sigLogUpdate(4, i18n("Unable to init SFTP!"));
    return -1;
  }
  
  emit sigLogUpdate(3, i18n("Connection established."));
  
  m_isLoggedIn = true;
  return 1;
}

char addPermStr(bool cond, char add)
{
  if (cond)
    return add;
  else
    return '-';
}

QString SftpSocket::posixToString(int permissions)
{
  QString str;
  
  // User
  str += addPermStr(permissions & S_IRUSR, 'r');
  str += addPermStr(permissions & S_IWUSR, 'w');
  str += addPermStr(permissions & S_IXUSR, 'x');
  
  // Group
  str += addPermStr(permissions & S_IRGRP, 'r');
  str += addPermStr(permissions & S_IWGRP, 'w');
  str += addPermStr(permissions & S_IXGRP, 'x');
  
  // Other
  str += addPermStr(permissions & S_IROTH, 'r');
  str += addPermStr(permissions & S_IWOTH, 'w');
  str += addPermStr(permissions & S_IXOTH, 'x');
  
  return str;
}

bool SftpSocket::protoDirList(const KURL &url)
{
  if (!m_isLoggedIn) {
    // We are not logged in
    return false;
  }
  
  if (!protoCwd(url.path()))
    return false;
  
  emit sigLogUpdate(0, i18n("SFTP Reading directory listing..."));
    
  setBusy(true);
  m_lastDirList.clear();
  
  // Read the specified directory
  SFTP_ATTRIBUTES *file;
  FTPEntry entry;
  while ((file = sftp_readdir(m_sftpSession, m_curDir))) {
    entry.name = file->name;
    
    if (entry.name != "." && entry.name != "..") {
      entry.owner = file->owner;
      entry.group = file->group;
      entry.date = file->mtime;
      entry.size = file->size;
      entry.link = "";
      entry.permissions = posixToString(file->permissions);
      
      if (file->permissions & S_IFDIR)
        entry.type = 'd';
      else
        entry.type = 'f';
      
      m_lastDirList.append(FTPDirectoryItem(entry));
    }
    
    sftp_attributes_free(file);
  }
  
  protoCloseDir();
  setBusy(false);
  
  return true;
}

void SftpSocket::protoCloseDir()
{
  if (m_curDir) {
    sftp_dir_close(m_curDir);
    m_curDir = 0L;
  }
}

int SftpSocket::protoCwd(const QString &dir)
{
  protoCloseDir();
  m_curDir = sftp_opendir(m_sftpSession, (char*) dir.ascii());
  
  if (!m_curDir)
    return 0;

  emit sigLogUpdate(0, i18n("SFTP Directory changed to '%1'").arg(dir));

  return 1;
}

int SftpSocket::protoRmdir(const QString &dir)
{
  return sftp_rmdir(m_sftpSession, (char*) dir.ascii()) < 0 ? -1 : 1;
}

int SftpSocket::protoDelete(const QString &path)
{
  return sftp_rm(m_sftpSession, (char*) path.ascii()) < 0 ? -1 : 1;
}

int SftpSocket::protoMkdir(const QString &dir)
{
  SFTP_ATTRIBUTES *attrs = static_cast<SFTP_ATTRIBUTES*>(malloc(sizeof(SFTP_ATTRIBUTES)));
  memset(attrs, 0, sizeof(*attrs));
  
  int ret = sftp_mkdir(m_sftpSession, (char*) dir.ascii(), attrs) < 0 ? -1 : 1;
  
  free(attrs);
  return ret;
}

int SftpSocket::protoRename(const QString &source, const QString &destination)
{
  return sftp_rename(m_sftpSession, (char*) source.ascii(), (char*) destination.ascii()) < 0 ? -1 : 1;
}

int SftpSocket::protoSize(const QString &file)
{
  QString parentDir = KURL(file).directory();
  QString fileName = KURL(file).fileName();
  if (!protoCwd(parentDir))
    return -1;
  
  SFTP_ATTRIBUTES *f;
  while ((f = sftp_readdir(m_sftpSession, m_curDir))) {
    if (QString(f->name) == fileName) {
      m_lastSize = f->size;
      break;
    }
  }
    
  protoCloseDir();
  return 1;
}

int SftpSocket::protoGet(const QString &source, const QString &destination)
{
  SFTP_FILE *rf;
  FILE *f;
  filesize_t offset = 0;
  QDir fs;

  /* Check the sizes (and if the file exists) */
  filesize_t size_local = 0;
  if (protoSize(source) == -1) {
    // File does not exist or it is a directory
    return -1;
  }

  filesize_t size_remote = m_lastSize;

  /* Does destination (local) file exist ? */
  if (fs.exists(destination) && getConfig("feat_resume") == 1) {
    // Check if we can resume
    f = fopen(destination, "a");
    if (f != NULL) {
      if (fseek(f, 0, SEEK_END) != 0) {
        fclose(f);
        return -1;
      }
      
      size_local = ftell(f);
  
      if (size_local < size_remote && size_local > 0) {
        offset = size_local;
      }
    } else {
      return -1;
    }
  } else {
    // Check if the file path exists, if not, create it
    QString dest_dir = fs.cleanDirPath(destination.mid(0, destination.findRev('/')));

    // Is the path available ?
    if (!fs.exists(dest_dir)) {
      // Create all dirs
      QString full_path;
      for(register int i = 1;i<=dest_dir.contains('/');i++) {
        full_path += "/" + dest_dir.section('/', i, i);
        if (!fs.exists(full_path)) {
          fs.mkdir(full_path);
        }
      }
    }
    
    // Create the file
    f = fopen(destination, "w");
    
    if (f == NULL) {
      return -1;
    }
  }
  
  if (size_remote == size_local && getConfig("feat_resume") == 1) {
    /* File is already downloaded - skip it if enabled */
    fclose(f);
    processedSize(size_local);

    return 1;
  }
  
  emit sigLogUpdate(3, i18n("Starting with '%1' file download").arg(source));
  
  // Download the file
  char buffer[4096];
  filesize_t bytes_receieved = offset;

  setBusy(true);
  setOffset(offset);
  emit sigResumedOffset(offset);
  
  rf = sftp_open(m_sftpSession, (char*) source.data(), O_RDONLY, 0);
  if (!rf) {
    fclose(f);
    processedSize(size_local);
    setBusy(false);
    
    return -1;
  }
  
  // Use the offset we are given
  if (offset > 0)
    sftp_seek(rf, offset);
  
  while (int b_read = sftp_read(rf, buffer, sizeof(buffer))) {
    fwrite(buffer, b_read, 1, f);
    bytes_receieved += b_read;
    
    if (b_read == 0 && getSpeed() == 0) {
      // We are stalled...
      // FIXME should we abort transfer after timeout ?
    }

    processedSize(b_read);

    if (m_abort)
      break;
  }
  
  fflush(f);
  fclose(f);
  sftp_file_close(rf);
  
  setBusy(false);
  
  emit sigLogUpdate(3, i18n("Transfered %1 bytes.").arg(bytes_receieved));
  
  return 1;
}

int SftpSocket::protoPut(const QString &source, const QString &destination)
{
  int fd;
  SFTP_FILE *rf;
  filesize_t offset = 0;
  QDir fs;

  /* Check the sizes (and if the file exists) */
  filesize_t size_local;
  filesize_t size_remote = 0;

  if (!fs.exists(source)) {
    return -1;
  } else {
    /* Get local filesize */
    fd = open(source, O_RDONLY);
    
    if (fd < 0) {
      return -1;
    } else {
      size_local = lseek(fd, 0, SEEK_END);
      lseek(fd, 0, SEEK_SET);
    }
  }

  /* Does destination (remote) file exist ? */
  if (protoSize(destination) == 1 && getConfig("feat_resume") == 1) {
    // Check if we can resume
    size_remote = m_lastSize;

    if (size_remote < size_local) {
      offset = size_remote;
    }
  }

  if (size_remote == size_local && getConfig("feat_resume") == 1) {
    /* File is already on the server - skip it if enabled */
    close(fd);
    processedSize(size_remote);
    return 1;
  }
  
  // Does the destination directory exist ?
  if (size_remote == 0 && !checkIsDir(KURL(destination).directory())) {
    // It doesn't - create it
    QString dest_dir = KURL(destination).directory();
    QString full_path;
    for(register int i = 1; i <= dest_dir.contains('/'); i++) {
      full_path += "/" + dest_dir.section('/', i, i);
      
      // Create the directory
      protoMkdir(full_path);
    }
  }
  
  emit sigLogUpdate(3, i18n("Starting with '%1' file upload").arg(source));
  
  // Open the file and grab the data
  char buffer[4096];
  filesize_t bytes_sent = offset;

  setBusy(true);
  setOffset(offset);
  emit sigResumedOffset(offset);
  
  rf = sftp_open(m_sftpSession, (char*) destination.data(), O_WRONLY | O_CREAT, 0);
  if (!rf) {
    close(fd);
    processedSize(size_remote);
    setBusy(false);
    
    return -1;
  }

  while (int b_read = read(fd, buffer, sizeof(buffer))) {
    sftp_write(rf, buffer, b_read);
    bytes_sent += b_read;

    processedSize(b_read);

    if (m_abort)
      break;
  }
  close(fd);
  sftp_file_close(rf);

  setBusy(false);
  
  emit sigLogUpdate(3, i18n("Transfered %1 bytes.").arg(bytes_sent));
  
  return 1;
}

}
#include "sftpsocket.moc"
