/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpbookmarks.h"
#include "kftpbookmarkaction.h"
#include "kftpgrabber.h"
#include "kftpqueue.h"
#include "kftpconfiguration.h"
#include "kftpapi.h"
#include "kftpwalletconnection.h"
#include "misc.h"
#include "desencryptor.h"
#include "kftpfiledirview.h"
#include "kftpfiledirlistview.h"
#include "kftpsession.h"
#include "kftpbookmarksshare.h"
#include "ftpsocket.h"

#include <qfile.h>

#include <klocale.h>
#include <kiconloader.h>
#include <kpassdlg.h>
#include <kmessagebox.h>

/* KSSL includes */
#include <ksslpkcs12.h>

using namespace KFTPGrabberBase;

KFTPBookmarks *FTPBookmarks;
static QCString des_pass = "";

KFTPBookmarks::KFTPBookmarks()
  : QObject(0)
{
  xmlDoc = QDomDocument("KFTPGrabber");
  m_bookmarksShare = new KFTPBookmarksShare(this);
}

KFTPBookmarks::KFTPBookmarks(const KFTPBookmarks &b)
  : QObject(0)
{
  xmlDoc = QDomDocument("KFTPGrabber");
  
  QDomNode tmp = xmlDoc.importNode(b.xmlDoc.documentElement(), true);
  xmlDoc.appendChild(tmp);
  
  m_bookmarksShare = new KFTPBookmarksShare(this);
}

void KFTPBookmarks::domReadFile(const QString &fileName)
{
  // Reads a file via te DOM parser
  m_xmlFile = fileName;
  QFile file(fileName);
  if (!file.open(IO_ReadOnly)) {
    qDebug("WARNING: Unable to open xml, creating a new one!");

    // Create new empty XML
    xmlDoc.setContent(QString("<sites></sites>"));

    return;
  }
  
  // Check if the file is encrpyted
  QCString content(file.readAll());
  if (content.left(6) != "<sites" && content.left(9) != "<!DOCTYPE") {
    // File is encrypted
pwd_entry:
    qDebug("Detected encrypted bookmark file...");
    int saveToWallet = 1;
    
    QCString p_pass(KFTPAPI::getInstance()->walletConnection()->getPassword("bookmarkDecryptPwd"));    
    if (QString(p_pass).isNull()) {
      // Ask the user for a password
      int ret = KPasswordDialog::getPassword(p_pass, i18n("This bookmark file is encrypted. Please enter key for decryption!"), &saveToWallet);
      
      if (ret != KPasswordDialog::Accepted) {
        // User pressed cancel
        p_pass = "";
      }
    }
    
    // Try to decrypt
    DESEncryptor des;
    des.setKey(p_pass);
    des.decrypt(content);
    
    if (des.output().left(6) != "<sites" && des.output().left(9) != "<!DOCTYPE") {
      // Decryption failed
      qDebug("Decryption failed.", des.output().ascii());
      
      // Clear any saved passwords
      KFTPAPI::getInstance()->walletConnection()->setPassword("bookmarkDecryptPwd", QString::null);
      
      if (KMessageBox::warningContinueCancel(
            KFTPAPI::getInstance()->mainWindow(), 
            i18n("<qt>Bookmark file decryption has failed with provided key! Do you want to <b>overwrite</b> bookmarks with an empty file ?<br><br><font color=\"red\"><b>Warning:</b> If you overwrite, all current bookmarks will be lost!</font></qt>"),
            i18n("Decryption failed"),
            KGuiItem(i18n("&Overwrite bookmarks"))
          ) != KMessageBox::Continue)
      {
        // Request the password again
        goto pwd_entry;
      }
      
      // Create new empty XML
      xmlDoc.setContent(QString("<sites></sites>"));
      
      file.close();
      return;
    }
    
    // Save the password for later encryption
    des_pass = p_pass; 
    content = des.output();
    qDebug("Decryption done.");
    
    if (saveToWallet) {
      KFTPAPI::getInstance()->walletConnection()->setPassword("bookmarkDecryptPwd", p_pass);
    }
  }
  
  xmlDoc.setContent(content);
  file.close();
  
  // We have just loaded the bookmarks, so update all the menus
  emit update();
}

void KFTPBookmarks::domSaveFile()
{
  // Save the new XML file
  if (m_xmlFile.isEmpty()) {
    qDebug("WARNING: No open XML file, will NOT save.");
    return;
  }

  QFile file(m_xmlFile);
  if (!file.open(IO_WriteOnly)) {
    qDebug("WARNING: Unable to open xml for writing!");
    return;
  }
  
  // Should we encrypt the data ?
  QString content = xmlDoc.toString();
  if (FTPConfig().m_encryptBookmarks) {
    DESEncryptor des;
    
    if (des_pass.isEmpty()) {
      // Ask the user for the password
      KPasswordDialog::getPassword(des_pass, i18n("Enter key for bookmark file encryption!"));
    }
    
    des.setKey(des_pass);
    des.encrypt(content);
    content = des.output();
  }

  // Write the XML data to the stream
  QTextStream fileStream(&file);
  fileStream << content;
  file.flush();
  file.close();
}

QDomNode KFTPBookmarks::domFindTagByName(const QString &tag, const QString &name)
{
  QDomElement rootElement = xmlDoc.documentElement();

  // Get all tags :)
  QDomNodeList tagNodes = rootElement.elementsByTagName(tag);
  for(unsigned int i=0;i<tagNodes.length();i++) {
    QDomNode tagNode = tagNodes.item(i);
    if (!tagNode.toElement().isNull()) {
      // Grab the name
      QString sName = tagNode.toElement().attribute("name", "");
      if (sName == name) return tagNode;
    }
  }

  return rootElement;
}

QDomNode KFTPBookmarks::domFindServerByHost(const QString &serverHost)
{
  // Finds a server by its host
  QDomElement rootElement = xmlDoc.documentElement();

  // Get all tags :)
  QDomNodeList tagNodes = rootElement.elementsByTagName("host");
  for(unsigned int i=0;i<tagNodes.length();i++) {
    QDomNode tagNode = tagNodes.item(i);
    if (!tagNode.toElement().isNull()) {
      // Grab the host
      QString host = tagNode.toElement().text();
      if (host == serverHost) return tagNode.parentNode();
    }
  }
  
  return QDomNode();
}

QDomNode KFTPBookmarks::domFindServerByName(const QString &serverName)
{
  // Finds a server by its name
  return domFindTagByName("server", serverName);
}

bool KFTPBookmarks::domServerExistsByName(const QString &serverName)
{
  // Finds a server by its name
  if (domFindServerByName(serverName).toElement().attribute("name", "") == serverName)
    return true;
  else
    return false;
}

QDomNode KFTPBookmarks::domFindServerByUrl(const KURL url, QDomNode parent)
{
  // Finds a server by its host, port, username, password
  QDomNode n = parent.firstChild();

  while (!n.isNull()) {
    if (n.toElement().tagName() == "category") {
      // Check the category
      QDomNode result = domFindServerByUrl(url, n);

      if (!result.isNull()) {
        // Search complete
        return result;
      }
    } else if (n.toElement().tagName() == "server") {
      // Check if the server matches
      if (domGetProperty(n, "host") == url.host() && domGetProperty(n, "port").toInt() == url.port() &&
          domGetProperty(n, "downuser") == url.user())
      {
        // The server has been found!
        return n;
      }
    }
    n = n.nextSibling();
  }

  // Nothing found
  n.clear();
  return n;
}

bool KFTPBookmarks::domServerExistsByUrl(const KURL url)
{
  if (!domFindServerByUrl(url, xmlDoc.documentElement()).toElement().isNull())
    return true;
  else
    return false;
}

bool KFTPBookmarks::domServerExistsByHost(const QString &serverHost)
{
  // Finds a server by its name
  if (!domFindServerByHost(serverHost).toElement().isNull())
    return true;
  else
    return false;
}

bool KFTPBookmarks::domCategoryExistsByName(const QString &categoryName)
{
  if (domFindCategoryByName(categoryName).toElement().attribute("name", "") == categoryName)
    return true;
  else
    return false;
}

QDomNode KFTPBookmarks::domFindCategoryByName(const QString &categoryName)
{
  // Finds a category by its name
  return domFindTagByName("category", categoryName);
}

void KFTPBookmarks::domDeleteServer(QDomNode server)
{
  server.parentNode().removeChild(server);

  update();
}

void KFTPBookmarks::domDeleteCategory(QDomNode category)
{
  category.parentNode().removeChild(category);

  update();
}

void KFTPBookmarks::domRenameCategory(QDomNode category, const QString &newName)
{
  category.toElement().setAttribute("name", newName);

  update();
}

void KFTPBookmarks::domCreateCategory(QDomNode parent, const QString &categoryName)
{
  QDomElement newCat = xmlDoc.createElement("category");
  newCat.setAttribute("name", categoryName);
  parent.appendChild(newCat);
  
  update();
}

void KFTPBookmarks::domAddProperty(QDomElement element, const QString &name, const QString &text)
{
  if (!text.isEmpty()) {
    QDomElement serverProperty = xmlDoc.createElement(name);
    element.appendChild(serverProperty);

    QDomText serverText = xmlDoc.createTextNode(text);
    serverProperty.appendChild(serverText);

    update();
  }
}

bool KFTPBookmarks::domDeleteProperty(QDomElement element, const QString &name)
{
  QDomNodeList tagNodes = element.elementsByTagName(name);
  if (tagNodes.length() > 0) {
    element.removeChild(tagNodes.item(0));
    return true;
  } else {
    return false;
  }
}

void KFTPBookmarks::domSetProperty(QDomElement element, const QString &name, const QString &text)
{
  domDeleteProperty(element, name);
  domAddProperty(element, name, text);
}

QString KFTPBookmarks::domGetProperty(QDomNode node, const QString &name)
{
  QDomNodeList tagNodes = node.toElement().elementsByTagName(name);
  if (tagNodes.length() > 0) {
    QString prop = tagNodes.item(0).toElement().text();
    prop.stripWhiteSpace();

    // Automagicly decode passwords from BASE64
    if (name.contains("pass") == 1)
      prop = decodePassword(prop);

    return prop;
  } else {
    return QString::null;
  }
}

QDomNode KFTPBookmarks::domFindNodeByXMLPath(const QString &path)
{
  QDomNode n = xmlDoc.documentElement().firstChild();
  
  if (path == "/") {
    // Return root node
    return xmlDoc.documentElement();
  }
  
  for (int i = 0; i < path.contains('/'); i++) {
    QString searchFor = path.section('/', i, i, QString::SectionSkipEmpty);
    
    while (!n.isNull()) {
      if (n.toElement().attribute("name") == searchFor)
        break;
        
      n = n.nextSibling();
    }
    
    // Have we found anything ?
    if (n.isNull())
       break;
    
    if (n.toElement().tagName() == "category" && i != path.contains('/') - 1)
      n = n.firstChild();
  }
  
  if (!n.isNull()) {
    return n;
  }
  
  return QDomNode();
}

void KFTPBookmarks::populateZeroconfMenu(KActionMenu *parentMenu)
{
  // Clear the menu
  parentMenu->popupMenu()->clear();
  
  #ifdef HAVE_HOWL
  // Populate
  QValueList<KFTPHowlService>::iterator i;
  QValueList<KFTPHowlService> list = KFTPAPI::getInstance()->howlInterface()->getServiceList();
  
  if (list.size() > 0) {
    for (i = list.begin(); i != list.end(); ++i) {
      KFTPZeroconfAction *newService = new KFTPZeroconfAction((*i).getServiceName(), "lan", KShortcut(), this, SLOT(slotZeroconfExecuted()), KFTPAPI::getInstance()->mainWindow()->actionCollection());
      newService->setSite(*i);
      
      parentMenu->insert(newService);
    }
  } else {
    KAction *disabledAction = new KAction(i18n("<No services published>"));
    disabledAction->setEnabled(false);
    parentMenu->insert(disabledAction);
  }
  #else
  KAction *disabledAction = new KAction(i18n("<Howl not available>"));
  disabledAction->setEnabled(false);
  parentMenu->insert(disabledAction);
  #endif
}

void KFTPBookmarks::populateBookmarksMenu(KActionMenu *parentMenu, QDomNode parentNode, bool base, QObject *data)
{
  QDomNode n = parentNode.firstChild();
  KActionMenu *menu = 0L;
  KFTPBookmarkAction *action = 0L;
  
  while (!n.isNull()) {
    QString name = n.toElement().attribute("name");
    
    if (n.toElement().tagName() == "category") {
      menu = new KActionMenu(name, "bookmark_folder", KFTPAPI::getInstance()->mainWindow()->actionCollection());
      parentMenu->insert(menu, base ? 4 : 0);
      
      // Fill the menu
      populateBookmarksMenu(menu, n, false, data);
    } else if (n.toElement().tagName() == "server") {
      action = new KFTPBookmarkAction(name, "ftp", KShortcut(), this, SLOT(slotBookmarkExecuted()), KFTPAPI::getInstance()->mainWindow()->actionCollection());
      action->setNode(n);
      action->setData(data);
      
      parentMenu->insert(action);
    }
    
    n = n.nextSibling();
  }
}

bool KFTPBookmarks::domDeleteOption(QDomElement element, const QString &name)
{
  bool itemDeleted = false;
  QDomNodeList tagNodes = element.elementsByTagName("option");
  if (tagNodes.length() > 0) {
    for(unsigned int i=0; i < tagNodes.length(); i++) {
      if (tagNodes.item(i).toElement().attribute("name", "") == name) {
        element.removeChild(tagNodes.item(0));
        itemDeleted = true;
      }
    }
  }
  
  return itemDeleted;
}

void KFTPBookmarks::domSetOption(QDomElement element, const QString &name, const QString &value)
{
  domDeleteOption(element, name);
  domAddOption(element, name, value);
}

void KFTPBookmarks::domSetOption(QDomElement element, const QString &name, int value)
{
  domDeleteOption(element, name);
  domAddOption(element, name, value);
}

void KFTPBookmarks::domAddOption(QDomElement element, const QString &name, const QString &value)
{
  if (!value.isEmpty()) {
    QDomElement serverProperty = xmlDoc.createElement("option");
    element.appendChild(serverProperty);

    serverProperty.setAttribute("name", name);
    serverProperty.setAttribute("value", value);

    update();
  }
}

void KFTPBookmarks::domAddOption(QDomElement element, const QString &name, int value)
{
  QDomElement serverProperty = xmlDoc.createElement("option");
  element.appendChild(serverProperty);

  serverProperty.setAttribute("name", name);
  serverProperty.setAttribute("value", value);

  update();
}

QString KFTPBookmarks::domGetOption(QDomNode node, const QString &name)
{
  QDomNodeList tagNodes = node.toElement().elementsByTagName("option");
  if (tagNodes.length() > 0) {
    for(unsigned int i=0;i<tagNodes.length();i++) {
      if (tagNodes.item(i).toElement().attribute("name", "") == name)
        return tagNodes.item(i).toElement().attribute("value", "");
    }
  }
  
  return QString::null;
}

KURL KFTPBookmarks::urlFromServer(QDomNode server)
{
  KURL url;
  url.setProtocol("ftp");
  url.setUser(domGetProperty(server, "downuser"));
  url.setPass(domGetProperty(server, "downpass"));
  url.setHost(domGetProperty(server, "host"));
  url.setPort(domGetProperty(server, "port").toInt());

  return url;
}

KURL KFTPBookmarks::urlFromClientSocket(KFTPClientThr *client, const QString &path)
{
  KURL url;
  if (domServerExistsByHost(client->getClient()->getClientInfoUrl().host())) {
    // Server is in bookmarks, use the bookmarked data
    url = urlFromServer(domFindServerByHost(client->getClient()->getClientInfoUrl().host()));
  } else {
    // Server is not in bookmarks, use the socket data
    url = client->getClient()->getClientInfoUrl();
  }
  url.setPath(path);

  return url;
}

void KFTPBookmarks::setFTPOptions(QDomNode server, KFTPClientThr *client)
{
  int retryTime = domGetProperty(server, "retrytime").toInt();
  int retryCnt = domGetProperty(server, "retrycount").toInt();
  
  // First activate the correct socket and reset the old flags
  client->getSocketManager()->setProtocol(KURL(QString("%1://test/").arg(domGetProperty(server, "protocol"))));
  client->getClient()->initConfig();

  if (retryTime != 0 && retryCnt != 0) {
    client->getClient()->setConfig("do_retry", 1);
    client->getClient()->setConfig("max_retries", retryCnt);
    client->getClient()->setConfig("retry_delay", retryTime);
  } else {
    client->getClient()->setConfig("do_retry", 0);
  }

  client->getClient()->setConfig("use_tls", domGetProperty(server, "use_tls") == "true" ? 1 : 0);
  client->getClient()->setConfig("use_implicit", domGetProperty(server, "use_implicit") == "true" ? 1 : 0);
  client->getClient()->setConfig("prot_mode", domGetProperty(server, "tls_mode").at(0).latin1());
  client->getClient()->setConfig("feat_pret", domGetProperty(server, "dist_ftpd") == "true" ? 1 : 0);
  client->getClient()->setConfig("can_pasv", domGetOption(server, "disablePASV").toInt() == 1 ? 0 : 1);
  client->getClient()->setConfig("can_epsv", domGetOption(server, "disableEPSV").toInt() == 1 ? 0 : 1);
  client->getClient()->setConfig("local_dir", domGetProperty(server, "deflocalpath"));
  
  // Should we use a X509 certificate ?
  if (domGetProperty(server, "use_cert") == "true" && domGetProperty(server, "protocol") == "ftp") {
    // Ask the user for the decryption password
    QCString p_pass;
    KPasswordDialog::getPassword(p_pass, i18n("Please provide your X509 certificate decryption password."));
    
    static_cast<KFTPNetwork::FtpSocket*>(client->getClient())->SSLSetClientCert(KSSLPKCS12::loadCertFile(domGetProperty(server, "tls_cert_path"), p_pass));
  }
}

void KFTPBookmarks::setFTPOptions(const KURL &url, KFTPClientThr *client)
{
  if (domServerExistsByUrl(url)) {
    QDomNode p_server = domFindServerByUrl(url, xmlDoc.documentElement());
    setFTPOptions(p_server, client);
  }
}

void KFTPBookmarks::slotBookmarkExecuted()
{
  // Get the sender
  KFTPBookmarkAction *action = (KFTPBookmarkAction*) QObject::sender();
  QDomNode selServer = action->node();
  
  // Get the node data from bookmarks
  KURL serverURL;
  serverURL.setProtocol(FTPBookmarks->domGetProperty(selServer, "protocol"));
  serverURL.setHost(FTPBookmarks->domGetProperty(selServer, "host"));
  serverURL.setPort(FTPBookmarks->domGetProperty(selServer, "port").toInt());
  if (!FTPBookmarks->domGetProperty(selServer, "lookuser").isEmpty()) {
    // If there is look account set, use it
    serverURL.setUser(FTPBookmarks->domGetProperty(selServer, "lookuser"));
    serverURL.setPass(FTPBookmarks->domGetProperty(selServer, "lookpass"));
  } else {
    // Otherwise use download account
    serverURL.setUser(FTPBookmarks->domGetProperty(selServer, "downuser"));
    serverURL.setPass(FTPBookmarks->domGetProperty(selServer, "downpass"));
  }
  
  // Check something
  if (!serverURL.hasUser()) {
    KMessageBox::sorry(0, i18n("No username specified for '%1'!").arg(serverURL.host()), i18n("Error connecting"));
    return;
  }
  
  KFTPClientThr *client = 0L;
  if (action->data()) {
    // We have some additional data
    KFTPFileDirView *view = (KFTPFileDirView*) action->data();
    
    if (view->m_list->m_offlineMode) {
      // If in offline mode, just open the correct URL if in cache
      serverURL.setPath("/");
      if (!FTPCache().listGetFromCache(serverURL)) {
        KMessageBox::sorry(0, i18n("This URL is not in cache!"));
        return;
      }
  
      view->m_list->m_offlineURL = serverURL;
      view->m_list->openURL(serverURL);
      
      // Trick the session into thinking this is a remote url
      view->m_ftpClient->getClient()->setSiteUrl(serverURL);
      FTPSessionManager->find(view)->slotClientConnected(true);
      
      return;
    }
    
    // Set the correct client for connection
    client = view->m_ftpClient;
    view->m_list->setHomeURL(remoteUrl(domGetProperty(selServer, "defremotepath"), serverURL));
    
    // Now, connect to the server
    if (client->getClient()->isConnected()) {
      if (KMessageBox::warningYesNo(0, i18n("Do you want to drop current connection ?")) == KMessageBox::No)
        return;
  
      client->KFTPClientThr::disconnect();
    }
    
    // Connect to the server
    client->getClient()->setSiteUrl(serverURL);
    setFTPOptions(selServer, client);
    client->connectToHost(serverURL);
  } else {
    // Just spawn a new session
    KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, serverURL, selServer);
    setFTPOptions(selServer, session->getClient());
  }
}

void KFTPBookmarks::slotZeroconfExecuted()
{
  #ifdef HAVE_HOWL
  // Get the sender
  KFTPZeroconfAction *action = (KFTPZeroconfAction*) QObject::sender();
  KFTPHowlService service = action->getSite();
  
  KFTPAPI::getInstance()->mainWindow()->slotQuickConnect(service.getServiceName(), service.getServiceIp(), service.getServicePort());
  #else
  qDebug("POSSIBLE BUG DETECTED: slotZeroconfExecuted() but HAVE_HOWL not defined !!!");
  #endif
}

KActionMenu *KFTPBookmarks::getSendToMenu()
{
  return m_bookmarksShare->getContactList();
}

void KFTPBookmarks::doEmitUpdate()
{
  emit update();
}

void KFTPBookmarks::setBookmarks(KFTPBookmarks *b)
{
  xmlDoc = QDomDocument("KFTPGrabber");
  QDomNode tmp = xmlDoc.importNode(b->getXMLDocument().documentElement(), true);
  xmlDoc.appendChild(tmp); 
  
  emit update();
}

