/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2004 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */
  
#include <qlabel.h>
#include <qlayout.h>
#include <qtextedit.h>
#include <qtabwidget.h>
#include <qhbox.h>
#include <qthread.h>

#include <kapplication.h>
#include <kmdimainfrm.h>
#include <kmdichildview.h>
#include <kmenubar.h>
#include <kpopupmenu.h>
#include <ktoolbar.h>
#include <kiconloader.h>
#include <kdialogbase.h>
#include <kfiledetailview.h>
#include <kmessagebox.h>
#include <klineeditdlg.h>
#include <kstatusbar.h>
#include <klocale.h>
#include <ktabwidget.h>
#include <kio/job.h>
#include <kpassdlg.h>
#include <ksslpkcs12.h> 

#include "misc.h"
#include "kftpapi.h"

// Widgets
#include "kftpsystemtray.h"
#include "kftpbookmarksidebar.h"
#include "kftpbookmarkeditor.h"
#include "kftpbookmarkeditorlistview.h"
#include "kftplogview.h"
#include "kftpqueueview.h"
#include "kftpquickconnectdialog.h"
#include "kftpserverlineedit.h"
#include "kftpfiledirview.h"
#include "kopeteballoon.h"
#include "kftpqueueviewtoolbar.h"
#include "kftpzeroconflistview.h"

// Other KFTPGrabber stuff
#include "kftpconfiguration.h"
#include "kftpgrabber.h"
#include "kftpbookmarks.h"
#include "kftpqueue.h"
#include "kftpsession.h"
#include "kftpqueueconverter.h"
#include "kftppluginmanager.h"
#include "ftpsocket.h"

#ifdef HAVE_QSA
#include "qsa/kftpscripting.h"
#endif

#ifdef HAVE_HOWL
#include "kftphowlthread.h"
#endif

using namespace KFTPGrabberBase;

KFTPGrabber::KFTPGrabber()
    : KMdiMainFrm( 0, "KFTPGrabber", KMdi::IDEAlMode )
{
  // init the API
  KFTPAPI::getInstance()->m_mainWindow = this;

  // Set the shell's ui resource file
  setXMLFile("kftpgrabberui.rc");
  connect(KApplication::kApplication(), SIGNAL(shutDown()), this, SLOT(appShutdown()));

  // Restore size and position
  resize(FTPConfig().m_size);
  move(FTPConfig().m_position);
  setCaption("KFTPGrabber");
  
  // Load plugins
  KFTPAPI::getInstance()->pluginManager()->loadPlugins();

  // Create some objects
  FTPQueueManager = new KFTPQueueManager(this);
  FTPBookmarks = new KFTPBookmarks();

#ifdef HAVE_QSA
  KFTPAPI::getInstance()->scriptInterface()->loadFromFile();
#endif

#ifdef HAVE_HOWL
  // Init the Howl (Rendezvous) thread
  KFTPAPI::getInstance()->howlInterface()->start();
  connect(KFTPAPI::getInstance()->howlEvents(), SIGNAL(howlListUpdate()), this, SLOT(initBookmarkMenu()));
#endif

  // Load the cache
  FTPCache().cacheLoadFromFile(getStoreDir("cache"));
  
  // Create the actions object
  m_actions = new KFTPMainActions(this, "main actions");
  m_actions->initActions();

  // Create the systray icon
  s_sysTray = new KFTPSystemTray(m_actions, this);

  // Update actions everytime, when there is any kind of change
  connect(FTPQueueManager, SIGNAL(queueUpdate()), m_actions, SLOT(updateActions()));
  connect(FTPQueueManager, SIGNAL(queueUpdate()), this, SLOT(slotUpdateStatusBar()));
  
  connect(FTPBookmarks, SIGNAL(update()), this, SLOT(initBookmarkMenu()));
  
  m_bookmarkMenu = 0L;
  m_zeroconfMenu = 0L;
  
  // Init the gui system
  initMainView();
  initStatusBar();
  initBookmarkMenu();
  
  // auto-save toolbar/menubar/statusbar settings
  setAutoSaveSettings(QString::fromLatin1("MainWindow"), false);
    
  // Create base two sessions
  FTPSessionManager->spawnLocalSession(LEFT_SIDE);
  FTPSessionManager->spawnLocalSession(RIGHT_SIDE);
  
  // Load the saved queue
  FTPQueueManager->getConverter()->importQueue(getStoreDir("queue"));
  
  // Load bookmarks
  QTimer::singleShot(500, this, SLOT(slotReadBookmarks()));
  
  emit qsaAppStartup();
   
  createGUI(0);
}

void KFTPGrabber::slotReadBookmarks()
{
  FTPBookmarks->domReadFile(getStoreDir("bookmarks.xml"));
  
  // Update the bookmark menu
  initBookmarkMenu();
}

void KFTPGrabber::appShutdown()
{
#ifdef HAVE_QSA
  KFTPAPI::getInstance()->scriptInterface()->saveToFile();
#endif

  // Save the config data on shutdown  
  FTPConfig().write();

  // Save ftp client cache
  FTPCache().cacheSaveToFile(getStoreDir("cache"));
  
  // Save current queue
  FTPQueueManager->getConverter()->exportQueue(getStoreDir("queue"));
}

bool KFTPGrabber::queryClose()
{
  if (FTPConfig().m_confirmExit && FTPQueueManager->getNumRunning() > 0) {
    if (KMessageBox::questionYesNo(0, i18n("There are currently %1 transfers running.\nAre you sure you want to quit ?")
                                      .arg(FTPQueueManager->getNumRunning()),
                                      i18n("Quit")) == KMessageBox::No)
    {
      return false;
    }
  }
  
  // Save XML bookmarks here, because the user may be prompted for an encryption password
  FTPBookmarks->domSaveFile();
  
  emit qsaAppExit();

  return true;
}

KFTPGrabber::~KFTPGrabber()
{
  delete m_bookmarkMenu;
  delete m_zeroconfMenu;
}

void KFTPGrabber::showBookmarkEditor()
{
  KFTPBookmarkEditor *bookmarkEditor = new KFTPBookmarkEditor(this, "bookmark editor");
  bookmarkEditor->exec();
  delete bookmarkEditor;
  
  // Update the bookmarks globaly
  FTPBookmarks->doEmitUpdate();
}

void KFTPGrabber::initBookmarkMenu()
{
  // Bookmarks menu
  if (m_bookmarkMenu == 0L)
    m_bookmarkMenu = new KActionMenu(i18n("Bookmarks"), 0, actionCollection(), "bookmarks");
  else
    m_bookmarkMenu->popupMenu()->clear();
  
  // Zeroconf menu
  if (m_zeroconfMenu == 0L)
    m_zeroconfMenu = new KActionMenu(i18n("FTP Sites near me"));
  
  m_bookmarkMenu->popupMenu()->insertItem(loadSmallIcon("bookmark"), i18n("Edit bookmarks..."), 1);
  m_bookmarkMenu->popupMenu()->connectItem(1, this, SLOT(showBookmarkEditor()));
  m_bookmarkMenu->insert(FTPBookmarks->getSendToMenu());
  m_bookmarkMenu->insert(m_zeroconfMenu);
  m_bookmarkMenu->popupMenu()->insertSeparator();
  FTPBookmarks->populateBookmarksMenu(m_bookmarkMenu, FTPBookmarks->getXMLDocument().documentElement(), true);
  FTPBookmarks->populateZeroconfMenu(m_zeroconfMenu);
}

/*!
    \fn KFTPGrabber::initStatusBar()
 */
void KFTPGrabber::initStatusBar()
{
  statusBar()->insertItem("<b>" + i18n("Status") +":</b> " + i18n("idle"), 1, 1);
  statusBar()->setItemAlignment(1, Qt::AlignLeft);

  statusBar()->insertItem(
    i18n("Download: %1/s  Upload: %1/s")
    .arg(KIO::convertSize(FTPQueueManager->getDownloadSpeed()))
    .arg(KIO::convertSize(FTPQueueManager->getUploadSpeed())),
    2
  );

  statusBar()->setItemAlignment(2, Qt::AlignLeft);
}

/*!
    \fn KFTPGrabber::initMainView()
 */
void KFTPGrabber::initMainView()
{
  KMdiChildView *centralView = new KMdiChildView( i18n( "Central View" ), this );
  
  QHBoxLayout *tLayout = new QHBoxLayout(centralView, 0, 0, "tLayout");
  QSplitter *splitter = new QSplitter(centralView);

  // View list layouts
  QWidget *box_1 = new QWidget(splitter);
  QWidget *box_2 = new QWidget(splitter);

  QHBoxLayout *lay_1 = new QHBoxLayout(box_1);
  QHBoxLayout *lay_2 = new QHBoxLayout(box_2);

  KTabWidget *leftTabs = new KTabWidget(box_1);
  KTabWidget *rightTabs = new KTabWidget(box_2);
  
  leftTabs->setHoverCloseButton(true);
  rightTabs->setHoverCloseButton(true);

  lay_1->addWidget(leftTabs);
  lay_2->addWidget(rightTabs);

  tLayout->addWidget(splitter);

  addWindow( centralView );
  
  // Create the tabwidget which holds the logs
  QTabWidget *tabWidget = new QTabWidget(0, "log");
  
  // Create the session manager
  FTPSessionManager = new KFTPSessionManager(this, tabWidget, leftTabs, rightTabs);

  // Create the queue 
  QWidget *queueWidget = new QWidget( 0, "queue" );
    
  QVBoxLayout *queueViewLayout = new QVBoxLayout(queueWidget);
  KFTPQueueView *queueView = new KFTPQueueView(queueWidget);
  KFTPQueueViewToolbar *queueViewToolbar = new KFTPQueueViewToolbar(queueWidget, queueView);
  
  queueViewLayout->addWidget(queueViewToolbar);
  queueViewLayout->addWidget(queueView);

  // Add the bottom bar
  addToolView( KDockWidget::DockBottom, queueWidget, SmallIcon("document"), i18n("Queue"));
  addToolView( KDockWidget::DockBottom, tabWidget, SmallIcon("info"), i18n("Log Messages"));

  // Add the bookmarks sidebar
  m_bookmarkSidebar = new KFTPBookmarkSidebar(0, "bookmarks");
  addToolView( KDockWidget::DockLeft, m_bookmarkSidebar, SmallIcon("bookmark"), i18n("Bookmarks"));
  
  // Add the zeroconf sidebar
  m_zeroconfSidebar = new KFTPZeroConfListView(0, "zeroconf");
  addToolView(KDockWidget::DockLeft, m_zeroconfSidebar, SmallIcon("lan"), i18n("Sites near me"));
}

void KFTPGrabber::slotUpdateStatusBar()
{
  // Status bar
  statusBar()->changeItem(i18n("idle"), 1);
  statusBar()->changeItem(
    i18n("Download: %1/s  Upload: %1/s")
    .arg(KIO::convertSize(FTPQueueManager->getDownloadSpeed()))
    .arg(KIO::convertSize(FTPQueueManager->getUploadSpeed())),
    2
  );
}

KMdiToolViewAccessor *KFTPGrabber::addToolView(KDockWidget::DockPosition position, QWidget *widget, const QPixmap &icon, const QString &sname, const QString &tabToolTip, const QString &tabCaption)
{
  widget->setIcon(icon);
  widget->setCaption(sname);

  return addToolWindow(widget, position, getMainDockWidget(), 25, tabToolTip, tabCaption);
}

bool KFTPGrabber::removeToolView(QWidget *w)
{
  deleteToolWindow (w);
  return true;
}

bool KFTPGrabber::removeToolView(KMdiToolViewAccessor *accessor)
{
  deleteToolWindow (accessor);
  return true;
}

void KFTPGrabber::slotQuickConnect(QString title, QString host, int port)
{
  // Create/get the new dialog
  KFTPQuickConnectDialog *qcDialog = new KFTPQuickConnectDialog(this, "qc dialog");
  qcDialog->setCaption(title);
  
  qcDialog->setHost(host);
  qcDialog->setPort(port);
  if (!host.isEmpty())
    qcDialog->setFocusToUser();

  if (qcDialog->exec() == KDialogBase::Accepted) {
    // Get the url and connect
    KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, qcDialog->getServerURL());
    
    // Set the server's options
    if ( qcDialog->getFTPMode() == 1 || qcDialog->getFTPMode() == 2 )
    {
      KFTPClientThr *client = session->getClient();
      
      // Set TLS options
      client->getClient()->setConfig("prot_mode", qcDialog->getTLSMode().latin1());
      if (qcDialog->getFTPMode() == 2)
        client->getClient()->setConfig("use_tls", 1);
      else
        client->getClient()->setConfig("use_implicit", 1);

      // Should we use a X509 certificate ?
      if (qcDialog->isCertChecked())
      {
        // Ask the user for the decryption password
        QCString p_pass;
        KPasswordDialog::getPassword(p_pass, i18n("Please provide your X509 certificate decryption password."));
    
        static_cast<KFTPNetwork::FtpSocket*>(client->getClient())->SSLSetClientCert(KSSLPKCS12::loadCertFile(qcDialog->getCertPath(), p_pass));
      }
    }
  }
  delete qcDialog;
}

#include "kftpgrabber.moc"
