/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2004 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpmainactions.h"
#include <config.h>

#include "misc.h"
#include "kftpapi.h"
#include "kftpfiltereditor.h"
#include "kftpclientthread.h"
#include "kftpqueue.h"
#include "kftpbookmarks.h"
#include "kftpconfiguration.h"
#include "kftpconfigdialog.h"
#include "kftpsession.h"

#ifdef HAVE_QSA
#include "qsa/kftpscripting.h"
#endif

#include <kmainwindow.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kaboutapplication.h>
#include <kkeydialog.h>
#include <kedittoolbar.h>

using namespace KFTPGrabberBase;

KFTPMainActions::KFTPMainActions(QObject *parent, const char *name)
 : QObject(parent, name)
{
  if (!parent->inherits("KMainWindow")) {
    qDebug("WARNING: KFTPMainAction should be created by KMainWindow only!");
    return;
  }

  m_configDialog = 0L;
}

KActionCollection *KFTPMainActions::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

KMainWindow *KFTPMainActions::mainWidget()
{
  return KFTPAPI::getInstance()->mainWindow();
}

void KFTPMainActions::initActions()
{
  // setup File menu
  m_fileConnectAction = new KAction(i18n("Quick &connect"), "connect_creating", KShortcut(), mainWidget(), SLOT(slotQuickConnect()), actionCollection(), "file_quick_connect");
  m_newSessionAction = new KActionMenu(i18n("&New session"), "filenew", actionCollection(), "file_newsession");
  
  KAction *leftSide = new KAction(i18n("&Left side"), KShortcut(), this, SLOT(slotNewSessionLeft()), actionCollection());
  KAction *rightSide = new KAction(i18n("&Right side"), KShortcut(), this, SLOT(slotNewSessionRight()), actionCollection());
  m_newSessionAction->insert(leftSide);
  m_newSessionAction->insert(rightSide);
  m_newSessionAction->setStickyMenu(true);
  m_newSessionAction->setDelayed(false);
  KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());  
  
  // setup View menu
  new KAction(i18n("&Filter editor..."), KShortcut(), this, SLOT(slotViewFilters()), actionCollection(), "edit_filters");
#ifdef HAVE_QSA
  new KAction(i18n("&Script editor..."), KShortcut(), this, SLOT(slotViewScriptEditor()), actionCollection(), "edit_scripts");
#endif  

  // setup Settings menu
  //KStdAction::configureToolbars(this, SLOT(slotConfigureToolbars()), actionCollection());
  //KStdAction::keyBindings(this, SLOT(slotKeyBindings()), actionCollection());
  KStdAction::saveOptions(this, SLOT(slotSettingsSave()), actionCollection());
  KStdAction::preferences(this, SLOT(slotSettingsConfig()), actionCollection());
  mainWidget()->createStandardStatusBarAction();
  mainWidget()->setStandardToolBarMenuEnabled(true);
  
  // setup transfermode actions    
  m_transModeAction = new KActionMenu(i18n("&Transfer mode (Auto)"), "binary", actionCollection(), "transfermode");
  m_modeAscii = new KRadioAction(i18n("ASCII"), "ascii", KShortcut(), this, SLOT(slotModeAscii()), actionCollection());
  m_modeBinary = new KRadioAction(i18n("Binary"), "binary", KShortcut(), this, SLOT(slotModeBinary()), actionCollection());
  m_modeAuto = new KRadioAction(i18n("Auto"), KShortcut(), this, SLOT(slotModeAuto()), actionCollection());
  
  // Set grouping so only one action can be selected
  m_modeAscii->setExclusiveGroup("Transfer mode");
  m_modeBinary->setExclusiveGroup("Transfer mode");
  m_modeAuto->setExclusiveGroup("Transfer mode");
  
  // Insert the actions into the menu
  m_transModeAction->insert(m_modeAscii);
  m_transModeAction->insert(m_modeBinary);
  m_transModeAction->popupMenu()->insertSeparator();
  m_transModeAction->insert(m_modeAuto);
  m_transModeAction->setStickyMenu(true);
  m_transModeAction->setDelayed(false);

  updateActions();
}

void KFTPMainActions::updateActions()
{
}

void KFTPMainActions::slotModeAscii()
{
  m_transModeAction->setIcon("ascii");
  m_transModeAction->setText(i18n("&Transfer mode (ASCII)"));
  FTPConfig().m_transMode = 'A';
}

void KFTPMainActions::slotModeBinary()
{
  m_transModeAction->setIcon("binary");
  m_transModeAction->setText(i18n("&Transfer mode (Binary)"));
  FTPConfig().m_transMode = 'I';
}

void KFTPMainActions::slotModeAuto()
{
  m_transModeAction->setText(i18n("&Transfer mode (Auto)"));
  FTPConfig().m_transMode = 'X';
}

void KFTPMainActions::slotFileQuit()
{
  mainWidget()->close();
}

void KFTPMainActions::slotViewScriptEditor()
{
#ifdef HAVE_QSA
  KFTPScripting::getInstance()->showScriptEditor();
#endif
}

void KFTPMainActions::slotViewFilters()
{
  KFTPFilterEditor *m_filterEditor = new KFTPFilterEditor(mainWidget(), "filter editor");
  m_filterEditor->exec();
}

void KFTPMainActions::slotSettingsSave()
{
  FTPConfig().write();
}

void KFTPMainActions::slotSettingsConfig()
{
  if (!m_configDialog)
    m_configDialog = new KFTPConfigDialog(mainWidget());

  m_configDialog->exec();
}

void KFTPMainActions::slotNewSessionLeft()
{
  FTPSessionManager->spawnLocalSession(LEFT_SIDE, true);
}

void KFTPMainActions::slotNewSessionRight()
{
  FTPSessionManager->spawnLocalSession(RIGHT_SIDE, true);
}

void KFTPMainActions::slotKeyBindings()
{
  KKeyDialog::configure(actionCollection());
}

void KFTPMainActions::slotConfigureToolbars()
{
  KEditToolbar dlg(actionCollection());

  if (dlg.exec())
    mainWidget()->createGUI();
}

#include "kftpmainactions.moc"
