/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpqueueprocessor.h"
#include "kftpqueue.h"
#include "kftpsession.h"
#include "misc.h"

using namespace KFTPGrabberBase;

KFTPQueueProcessor::KFTPQueueProcessor(QObject *parent, const char *name)
 : QObject(parent, name)
{
  m_running = false;
  m_srcSession = 0L;
  m_dstSession = 0L;
}

KFTPQueueProcessor::~KFTPQueueProcessor()
{
}

bool KFTPQueueProcessor::isRunning()
{
  return m_running;
}

void KFTPQueueProcessor::processActiveTransfer()
{
  if (!m_activeTransfer) {
    qWarning("KFTPQueueProcessor: Active transfer is NULL ?!?");
    return;
  }
  
  if (!m_srcSession) {
    m_srcSession = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, m_activeTransfer->getSourceUrl());
    m_dstSession = FTPSessionManager->spawnRemoteSession(OPPOSITE_SIDE(m_srcSession->getSide()), m_activeTransfer->getDestUrl());
    
    connect(m_srcSession->getClient(), SIGNAL(loginComplete(bool)), this, SLOT(slotTransferLoggedIn(bool)));
    connect(m_dstSession->getClient(), SIGNAL(loginComplete(bool)), this, SLOT(slotTransferLoggedIn(bool)));
  } else {
    // Disconnect stuff
    if (FTPSessionManager->find(m_activeTransfer->getSourceUrl()) != m_srcSession)
      m_srcSession->reconnect(m_activeTransfer->getSourceUrl());
      
    if (FTPSessionManager->find(m_activeTransfer->getDestUrl()) != m_dstSession)
      m_dstSession->reconnect(m_activeTransfer->getDestUrl());
  }
  
  if (!m_srcSession->isConnected() || !m_dstSession->isConnected()) {
    // We have to wait for both sessions to be available
    return;
  }

  // Start the transfer
  m_activeTransfer->delayedExecute();
}

bool KFTPQueueProcessor::nextTransfer()
{
  // Select the first server (should be next) if there is none, we are done :)
  KFTPQueueTransfers *transfers = FTPQueueManager->getQueueTransferList();
  
  if (transfers->count() > 0 && m_running) {
    if (m_activeTransfer && m_activeTransfer->getId() == transfers->at(0)->getId()) {
      // Because this function can be called from slotTransferComplete and signal gets
      // emitted *before* transfer is removed we should use the second transfer on the list
      m_activeTransfer = transfers->at(1);
    } else {
      m_activeTransfer = transfers->at(0);
    }
  } else {
    m_activeTransfer = 0L;
  }
  
  if (m_activeTransfer) {
    // Connect the signals
    connect(m_activeTransfer, SIGNAL(transferComplete(long)), this, SLOT(slotTransferComplete()));
    
    return true;
  } else {
    // We are done, so we emit the proper signal
    m_running = false;
    
    emit queueComplete();
    return false;
  }
}

void KFTPQueueProcessor::startProcessing()
{
  m_running = true;
  m_srcSession = 0L;
  m_dstSession = 0L;
  
  // Select the transfer and process it
  if (nextTransfer()) {
    processActiveTransfer();
  }
}

void KFTPQueueProcessor::stopProcessing()
{
  // Stop the queue processing
  m_running = false;
  m_srcSession = 0L;
  m_dstSession = 0L;
  
  // Abort current transfer
  if (m_activeTransfer) {
    m_activeTransfer->abort();
  
    // Disconnect signals
    m_activeTransfer->QObject::disconnect(this);
  }
  
  m_activeTransfer = 0L;
}

void KFTPQueueProcessor::slotTransferComplete()
{
  // Transfer is complete, move to the next
  if (nextTransfer()) {
    processActiveTransfer();
  }
}

void KFTPQueueProcessor::slotTransferLoggedIn(bool success)
{
  if (success) {
    if (!m_srcSession || !m_dstSession || !m_srcSession->isConnected() || !m_dstSession->isConnected()) return;
    
    // Execute the transfer if both sessions are connected
    if (!m_activeTransfer->isRunning())
      m_activeTransfer->delayedExecute();
  } else {
    // Unable to login - skip the file
    slotTransferComplete();
  }
}

#include "kftpqueueprocessor.moc"

