/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpsession.h"
#include "kftpapi.h"
#include "kftpfiledirlistview.h"
#include "kftpbookmarks.h"
#include "kftpconfiguration.h"
#include "misc.h"
#include "kftpsystemtray.h"

#include <qdir.h>
#include <qobjectlist.h>

#include <kmessagebox.h>
#include <klocale.h>

using namespace KFTPGrabberBase;

KFTPSessionManager *FTPSessionManager;

////////////////////////////////////////////////////////
//////////////////    KFTPSession     //////////////////
////////////////////////////////////////////////////////

KFTPSession::KFTPSession(SessionSide side)
  : QObject(), m_side(side), m_remote(false), m_curTransfer(0L)
{
  // Register this session
  FTPSessionManager->registerSession(this);
  
  connect(m_client, SIGNAL(finished(CommandType)), this, SLOT(slotRemoteCommandDone(CommandType)));
}

KFTPSession::~KFTPSession()
{
  disconnect(m_client, SIGNAL(finished(CommandType)), this, SLOT(slotRemoteCommandDone(CommandType)));
}

bool KFTPSession::isConnected()
{
  if (!m_remote)
    return true;
    
  return m_client->getClient()->isConnected() || m_fileView->m_list->m_offlineMode;
}

bool KFTPSession::canLock()
{
  return m_curTransfer.isNull() && !m_fileView->m_list->m_offlineMode;
}

bool KFTPSession::lock(KFTPQueue::Transfer *transfer, bool lockOnly)
{
  // Lock current session to selected transfer -- only one transfer can
  // have a session locked
  if (!canLock()) return false;
  
  // Local sessions can't be locked
  if (!m_remote) return true;
  
  m_curTransfer = transfer;
  if (!lockOnly) {
    connect(transfer, SIGNAL(transferComplete(long)), this, SLOT(slotTransferCompleted()));
  }
  
  return true;
}

void KFTPSession::unlock()
{
  // Remove current transfer, so session becomes unlocked
  if (m_curTransfer) {
    disconnect(m_curTransfer, SIGNAL(transferComplete(long)), this, SLOT(slotTransferCompleted()));
    m_curTransfer = 0L;
  }
}

void KFTPSession::slotTransferCompleted()
{
  unlock();
}

void KFTPSession::slotClientConnected(bool success)
{
  if (success) {
    m_remote = true;
    
    FTPSessionManager->getTabs(m_side)->changeTab(m_fileView, m_client->getClient()->getClientInfoUrl().host());
    FTPSessionManager->getStatTabs()->changeTab(m_log, i18n("Log (%1)").arg(m_client->getClient()->getClientInfoUrl().host()));
  }
}

void KFTPSession::slotClientDisconnected()
{
  m_remote = false;
  
  FTPSessionManager->getTabs(m_side)->changeTab(m_fileView, i18n("Local session"));
  FTPSessionManager->getStatTabs()->changeTab(m_log, "[" + i18n("Log") + "]");
}

void KFTPSession::slotClientRetrySuccess()
{
  if (FTPConfig().m_showRetrySuccessBalloon) {
    s_sysTray->showBalloon(i18n("Connection with %1 has been successfully established.").arg(m_client->getClient()->getClientInfoUrl().host()));
  }
}

void KFTPSession::scanDirectory(KFTPQueue::Transfer *parent)
{
  // Go trough all files in path and add them as transfers that have parent as their parent
  // transfer
  KURL path = parent->getSourceUrl();
  
  if (path.isLocalFile()) {
    // Scan local
    QDir fs(path.path());
    const QFileInfoList *p_list = fs.entryInfoList();
    QFileInfoListIterator i(*p_list);
    QFileInfo *file;
    
    while ((file = i.current()) != 0) {
      ++i;
      
      if (file->fileName() == "." || file->fileName() == "..") continue;
      if (FTPConfig().m_skipEmptyFiles && file->size() == 0) continue;
      
      // Spawn transfer
      KURL destUrl = parent->getDestUrl();
      destUrl.addPath(file->fileName());
    
      // This is needed, since QFileInfo works with uint for the filesize
      filesize_t realSize = KFileItem(KFileItem::Unknown, KFileItem::Unknown, file->absFilePath()).size();
    
      FTPQueueManager->spawnTransfer(KURL(file->absFilePath()), destUrl, realSize, file->isDir(), parent->getTransferType(), false, false, parent);
    }
    
    emit dirScanDone();
  } else if (m_remote) {
    // Check if we are in offline mode
    if (m_fileView->getListView()->m_offlineMode) {
      emit dirScanDone();
      return;
    }
    
    // Lock the session to the specified transfer
    if (!canLock()) {
      emit dirScanDone();
      return;
    }
    lock(parent, true);

    m_client->QObject::disconnect(m_fileView, SLOT(slotDisableView()));
    m_client->QObject::disconnect(m_fileView, SLOT(slotEnableView()));
    
    m_fileView->slotDisableView();
    m_client->scanDirectory(parent);
  }
}

void KFTPSession::emitNewChildren(KFTPQueue::Transfer *parent)
{
  QPtrList<KFTPQueue::Transfer> children = static_cast<KFTPQueue::TransferDir*>(parent)->getChildrenList();
  
  KFTPQueue::Transfer *i;
  for (i = children.first(); i; i = children.next()) {
    emit FTPQueueManager->newTransfer(i);
    
    if (i->isDir() && i->children()) {
      emitNewChildren(i);
    }
  }
}

void KFTPSession::slotRemoteCommandDone(CommandType type)
{
  if (type == TCMD_SCAN) {
    // Scan is complete, go trough all child transfers and signal a new transfer
    if (m_curTransfer && m_curTransfer->children())
      emitNewChildren(m_curTransfer);
    
    connect(m_client, SIGNAL(operationStarted()), m_fileView, SLOT(slotDisableView()));
    connect(m_client, SIGNAL(operationEnded()), m_fileView, SLOT(slotEnableView()));
    
    m_fileView->slotEnableView();
    
    // Unlock the session
    unlock();
    emit dirScanDone();
  }
  
  // Refresh the view
  if (type == TCMD_GET || type == TCMD_PUT || type == TCMD_CHMOD || type == TCMD_MKDIR || type == TCMD_RENAME || type == TCMD_REMOVE ||
      type == TCMD_FXP) {
    m_fileView->getListView()->moveTo(KFTPFileDirListView::Reload);
  }
}

void KFTPSession::abort()
{
  if (!m_remote) return;
  
  // Abort transfer
  m_client->getClient()->abort();
}

void KFTPSession::reconnect(const KURL &url)
{
  abort();
  if (m_remote && m_client->getClient()->isConnected())
    m_client->KFTPClientThr::disconnect();
  
  m_fileView->getListView()->setHomeURL(remoteUrl("/", url));
  m_client->connectToHost(url);
}

////////////////////////////////////////////////////////
////////////////// KFTPSessionManager //////////////////
////////////////////////////////////////////////////////

KFTPSessionManager::KFTPSessionManager(QObject *parent, QTabWidget *stat, KTabWidget *left, KTabWidget *right)
  : QObject(parent), m_statTabs(stat), m_leftTabs(left), m_rightTabs(right)
{
  // Init some stuff
  m_leftActive = 0L;
  m_rightActive = 0L;
  
  // Connect some signals
  connect(left, SIGNAL(currentChanged(QWidget*)), this, SLOT(slotActiveChanged(QWidget*)));
  connect(right, SIGNAL(currentChanged(QWidget*)), this, SLOT(slotActiveChanged(QWidget*)));
  
  connect(left, SIGNAL(closeRequest(QWidget*)), this, SLOT(slotSessionCloseRequest(QWidget*)));
  connect(right, SIGNAL(closeRequest(QWidget*)), this, SLOT(slotSessionCloseRequest(QWidget*)));
}

void KFTPSessionManager::registerSession(KFTPSession *session)
{
  // Create some new stuff and assign it to the session
  session->m_client = new KFTPClientThr(session);
  session->m_fileView = new KFTPFileDirView(0L, "", session->m_client);
  session->m_log = new KFTPLogView();

  // Connect some signals
  connect(session->m_client, SIGNAL(logUpdate(int, const QString&)), session->m_log, SLOT(ftpLog(int, const QString&)));
  connect(session->m_client, SIGNAL(loginComplete(bool)), session, SLOT(slotClientConnected(bool)));
  connect(session->m_client, SIGNAL(disconnectDone()), session, SLOT(slotClientDisconnected()));
  connect(session->m_client, SIGNAL(retrySuccess()), session, SLOT(slotClientRetrySuccess()));
  
  // Assign GUI positions
  m_statTabs->addTab(session->m_log, "Log");
  getTabs(session->m_side)->addTab(session->m_fileView, KFTPGrabberBase::loadSmallIcon("system"), i18n("Session"));
  
  // Actually add the session
  m_sessionList.append(session);
}

void KFTPSessionManager::unregisterSession(KFTPSession *session)
{
  // Destroy all objects related to the session and remove it
  getTabs(session->m_side)->removePage(session->m_fileView);
  m_statTabs->removePage(session->m_log);
  
  if (session->m_client->getClient()->isConnected()) {
    session->abort();
    session->m_client->KFTPClientThr::disconnect();
  }
  
  // Delete objects
  session->m_fileView->deleteLater();
  session->m_log->deleteLater();
  
  // Actually remove the session
  m_sessionList.remove(session);
  delete session;
}

KFTPSession *KFTPSessionManager::find(KFTPClientThr *client)
{
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    if (i->m_client == client)
      return i;
  }
  
  return 0L;
}

KFTPSession *KFTPSessionManager::find(KFTPQueue::Transfer *transfer)
{
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    if (!i->m_curTransfer.isNull()) {
      KFTPQueue::Transfer *tmp = i->m_curTransfer;
      
      if (tmp == transfer)
        return i;
    }
  }
  
  return 0L;
}

KFTPSession *KFTPSessionManager::find(KFTPFileDirView *fileView)
{
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    if (i->m_fileView == fileView)
      return i;
  }
  
  return 0L;
}

KFTPSession *KFTPSessionManager::find(KFTPLogView *log)
{
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    if (i->m_log == log)
      return i;
  }
  
  return 0L;
}

KFTPSession *KFTPSessionManager::find(const KURL &url, bool mustUnlock)
{
  if (url.isLocalFile())
    return find(true);
    
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    KURL tmp = i->m_client->getClient()->getClientInfoUrl();
    tmp.setPath(url.path());
    
    if (tmp == url && i->isRemote() && i->isConnected() && (!mustUnlock || i->canLock()))
      return i;
  }
  return 0L;
}

KFTPSession *KFTPSessionManager::find(bool local)
{
  KFTPSession *i;
  
  for (i = m_sessionList.first(); i; i = m_sessionList.next()) {
    if (i->m_remote != local)
      return i;
  }
  
  return 0L;
}

KFTPSession *KFTPSessionManager::spawnLocalSession(SessionSide side, bool forceNew)
{
  // Creates a new local session
  KFTPSession *session = 0L;
   
  if (forceNew || (session = find(true)) == 0L || (session->m_side != side && side != IGNORE_SIDE)) {
    side = side == IGNORE_SIDE ? LEFT_SIDE : side;
    
    session = new KFTPSession(side);
    session->m_remote = false;
    getTabs(side)->changeTab(session->m_fileView, i18n("Local session"));
    getStatTabs()->changeTab(session->m_log, "[" + i18n("Log") + "]");
  }
  
  setActive(session);
  return session;
}

KFTPSession *KFTPSessionManager::spawnRemoteSession(SessionSide side, const KURL &remoteUrl, QDomNode site, bool mustUnlock)
{
  // Creates a new remote session and connects it to the correct server
  KFTPSession *session;
  
  if (remoteUrl.isLocalFile())
    return spawnLocalSession(side);
  
  if ((session = find(remoteUrl, mustUnlock)) == 0L || (session->m_side != side && side != IGNORE_SIDE)) {
    side = side == IGNORE_SIDE ? RIGHT_SIDE : side;
    session = new KFTPSession(side);
    
    // Set properties
    session->m_remote = true;
    if (!site.isNull()) {
      session->m_fileView->getListView()->setHomeURL(KFTPGrabberBase::remoteUrl(FTPBookmarks->domGetProperty(site, "defremotepath"),
                                                                                remoteUrl));
    } else {
      session->m_fileView->getListView()->setHomeURL(KURL());
      FTPBookmarks->setFTPOptions(remoteUrl, session->m_client);
    }
    session->m_client->connectToHost(remoteUrl);
  }
  
  setActive(session);
  return session;
}

void KFTPSessionManager::setActive(KFTPSession *session)
{
  // Make a session active on its own side ;)
  KFTPSession *oldActive = getActive(session->m_side);
  
  oldActive ? oldActive->m_active = false : 0;
  session->m_active = true;
  
  // Set dir view companions
  if (oldActive) oldActive->m_fileView->setCompanion(0L);
  
  if (getActive(OPPOSITE_SIDE(session->m_side))) {
    session->m_fileView->setCompanion(getActive(OPPOSITE_SIDE(session->m_side))->m_fileView);
    getActive(OPPOSITE_SIDE(session->m_side))->m_fileView->setCompanion(session->m_fileView);
  } else {
    session->m_fileView->setCompanion(0L);
  }
    
  switch (session->m_side) {
    case LEFT_SIDE: m_leftActive = session; break;
    case RIGHT_SIDE: m_rightActive = session; break;
    case IGNORE_SIDE: qDebug("Invalid side specified!"); return;
  }
  
  // Refresh the GUI
  getTabs(session->m_side)->showPage(session->m_fileView);
}

KFTPSession *KFTPSessionManager::getActive(SessionSide side)
{
  switch (side) {
    case LEFT_SIDE: return m_leftActive;
    case RIGHT_SIDE: return m_rightActive;
    case IGNORE_SIDE: qDebug("Invalid side specified!"); break;
  }
  
  return NULL;
}

KTabWidget *KFTPSessionManager::getTabs(SessionSide side)
{
  switch (side) {
    case LEFT_SIDE: return m_leftTabs;
    case RIGHT_SIDE: return m_rightTabs;
    case IGNORE_SIDE: qDebug("Invalid side specified!"); break;
  }
  
  return NULL;
}

void KFTPSessionManager::slotActiveChanged(QWidget *page)
{
  KFTPSession *session = find(static_cast<KFTPFileDirView*>(page));
  setActive(session);
}

void KFTPSessionManager::slotSessionCloseRequest(QWidget *page)
{
  KFTPSession *session = find(static_cast<KFTPFileDirView*>(page));
  
  if (getTabs(session->m_side)->count() == 1) {
    KMessageBox::error(0L, i18n("At least one session must remain open on each side!"));
    return;
  }
  
  if (session->m_curTransfer || (session->m_remote && session->m_client->getClient()->isBusy())) {
    KMessageBox::error(0L, i18n("Please finish all transfers before closing the session!"));
    return;
  } else {
    // Remove the session
    if (session->m_client->getClient()->isConnected()) {
      if (KMessageBox::questionYesNo(0L, i18n("This session is currently connected. Are you sure you wish to disconnect ?"), i18n("Close session")) == KMessageBox::No)
        return;
    }
    
    unregisterSession(session);
  }
}

#include "kftpsession.moc"
