/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#ifndef KFTPSESSION_H
#define KFTPSESSION_H

#include <qobject.h>
#include <qguardedptr.h>
#include <qdom.h>

#include <ktabwidget.h>

#include "kftpqueue.h"
#include "kftpclientthread.h"
#include "kftplogview.h"
#include "kftpfiledirview.h"

enum SessionSide {
  LEFT_SIDE,
  RIGHT_SIDE,
  IGNORE_SIDE
};

#define OPPOSITE_SIDE(x) (x == LEFT_SIDE ? RIGHT_SIDE : LEFT_SIDE)

/**
KFTPSession connects all the relevant elements of KFTPGrabber together. Together
with the session manager it allows many ftp sessions inside one program.

@author Jernej Kos
*/
class KFTPSession : public QObject
{
Q_OBJECT
friend class KFTPSessionManager;
friend class KFTPBookmarks;
friend class KFTPFileDirViewActions;
public:
    KFTPSession(SessionSide side);
    ~KFTPSession();

    /* Session description */
    bool isRemote() { return m_remote; }
    bool isActive() { return m_active; }
    bool isConnected();
    
    KFTPClientThr *getClient() { return m_client; }
    KFTPLogView *getLog() { return m_log; }
    KFTPFileDirView *getFileView() { return m_fileView; }
    SessionSide getSide() { return m_side; }
    
    /* Session locking */
    bool canLock();
    bool lock(KFTPQueue::Transfer *transfer, bool lockOnly = false);
    void unlock();
    
    /* Session functions */
    void reconnect(const KURL &url);
    void abort();
    void scanDirectory(KFTPQueue::Transfer *parent);
private:
    SessionSide m_side;
    bool m_remote;
    bool m_active;
    
    KFTPClientThr *m_client;
    KFTPLogView *m_log;
    KFTPFileDirView *m_fileView;
    QGuardedPtr<KFTPQueue::Transfer> m_curTransfer;
    
    void emitNewChildren(KFTPQueue::Transfer *parent);
private slots:
    void slotTransferCompleted();
    void slotRemoteCommandDone(CommandType type);
    
    void slotClientConnected(bool success);
    void slotClientDisconnected();
    void slotClientRetrySuccess();
signals:
    void dirScanDone();
};

typedef QPtrList<KFTPSession> KFTPSessionList;

/**
KFTPSessionManager provides access to sessions, their registration and deletion.

@author Jernej Kos
*/
class KFTPSessionManager : public QObject
{
Q_OBJECT
public:
    KFTPSessionManager(QObject *parent, QTabWidget *stat, KTabWidget *left, KTabWidget *right);
    
    // Spawn functions
    KFTPSession *spawnLocalSession(SessionSide side, bool forceNew = false);
    KFTPSession *spawnRemoteSession(SessionSide side, const KURL &remoteUrl, QDomNode site = QDomNode(), bool mustUnlock = false);
    
    // Session registration
    void registerSession(KFTPSession *session);
    void unregisterSession(KFTPSession *session);
    
    // Session search
    KFTPSession *find(KFTPClientThr *client);
    KFTPSession *find(KFTPFileDirView *fileView);
    KFTPSession *find(KFTPLogView *log);
    KFTPSession *find(const KURL &url, bool mustUnlock = false);
    KFTPSession *find(bool local);
    KFTPSession *find(KFTPQueue::Transfer *transfer);
    
    // GUI items
    QTabWidget *getStatTabs() { return m_statTabs; }
    KTabWidget *getTabs(SessionSide side);
    
    // Active session
    void setActive(KFTPSession *session);
    KFTPSession *getActive(SessionSide side);
private:
    KFTPSessionList m_sessionList;
    
    // These variables should be assigned right after construction
    QTabWidget *m_statTabs;
    KTabWidget *m_leftTabs;
    KTabWidget *m_rightTabs;
    
    // Currently active sessions (there can be only two)
    KFTPSession *m_leftActive;
    KFTPSession *m_rightActive;
private slots:
    void slotActiveChanged(QWidget *page);
public slots:
    void slotSessionCloseRequest(QWidget *page);
};

extern KFTPSessionManager *FTPSessionManager;

#endif
