/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftptransfer.h"
#include "kftptransferdir.h"
#include "kftpconfiguration.h"
#include "kftpsystemtray.h"
#include "kftpqueue.h"

#include <klocale.h>

namespace KFTPQueue {

Transfer::Transfer(QObject *parent, Type type)
 : QObject(parent), m_deleteMe(false), m_type(type)
{
  m_status = Stopped;
  m_size = 0;
  m_actualSize = 0;
  m_completed = 0;
  m_resumed = 0;
  m_speed = 0;
  
  // Add the transfer
  if (isChild())
    static_cast<TransferDir*>(parent)->addChildTransfer(this);
}


Transfer::~Transfer()
{
}

void Transfer::delayedExecute(int msec)
{
  /* Execute the transfer with delay - using a QTimer */
  if (msec > 1000)
    msec = 1000;

  QTimer::singleShot(msec, this, SLOT(execute()));
}

void Transfer::execute()
{
  qDebug("BUG: KFTPQueue::Transfer class should *NOT* be used as standalone class!");
}

Transfer *Transfer::getTransferParent()
{
  if (!isChild()) return 0L;
  
  return static_cast<Transfer*>(parent());
}

void Transfer::addActualSize(filesize_t size)
{
  m_actualSize += size;
  
  if (isChild())
    getTransferParent()->addActualSize(size);
    
  emit transferUpdate();
}

void Transfer::addSize(filesize_t size)
{
  m_size += size;
  m_actualSize += size;
  
  if (isChild())
    getTransferParent()->addSize(size);
  
  emit transferUpdate();
}

void Transfer::addCompleted(filesize_t completed)
{
  m_completed += completed;
  
  if (isChild())
    getTransferParent()->addCompleted(completed);
    
  emit transferUpdate();
}

void Transfer::setSpeed(filesize_t speed)
{
  m_speed = speed;
  
  if (isChild())
    getTransferParent()->setSpeed(speed);
    
  emit transferUpdate();
}

void Transfer::resetTransfer()
{
  m_status = Stopped;
  m_resumed = 0;
  m_completed = 0;
}

void Transfer::update()
{
  FTPQueueManager->doEmitUpdate();
  emit transferUpdate();
}

void Transfer::abort()
{
  m_aborting = true;
  
  emit transferAbort(m_id);
}

void Transfer::showTransCompleteBalloon()
{
  // Show a balloon :P
  if (FTPConfig().m_showBalloons && !FTPQueueManager->queueProcRunning()) {
    if (!FTPConfig().m_showBalloonWhenQueueEmpty || (FTPQueueManager->getQueueTransferList()->count() == 1 && !isChild())) {
      QString p_transCompleteStr = i18n("Transfer of the following files is complete:");
      p_transCompleteStr += "<br><i>";
      p_transCompleteStr += getSourceUrl().fileName();
      p_transCompleteStr += "</i>";
      s_sysTray->showBalloon(p_transCompleteStr);
    }
  }
}

};
#include "kftptransfer.moc"
