/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#ifndef KFTPQUEUEKFTPTRANSFER_H
#define KFTPQUEUEKFTPTRANSFER_H

#include "kftpclientthread.h"

#include <qobject.h>
#include <qtimer.h>

#include <kurl.h>

class KFTPSession;

namespace KFTPQueue {

enum TransferType {
    Download = 0,
    Upload = 1,
    FXP = 2
};

/**
This class is the base class for all transfers used in KFTPGrabber. It provides some basic methods that are extended by KFTPQueue::TransferFile/KFTPQueue::TransferDir for specific file or dir operations.

@author Jernej Kos
*/
class Transfer : public QObject
{
friend class TransferDir;
Q_OBJECT
public:
    enum Status {
      Running,
      Stopped,
      Connecting,
      DoingStat
    };
    
    enum Type {
      File,
      Directory
    };
    
    Transfer(QObject *parent, Type type);
    ~Transfer();
    
    /**
     * Returns the source KURL of this transfer.
     *
     * @return Source url
     */
    KURL getSourceUrl() { return m_sourceUrl; }
    
    /**
     * Returns the destination KURL of this transfer.
     *
     * @return Destination url
     */
    KURL getDestUrl() { return m_destUrl; }
    
    /**
     * Set the source KURL of this transfer.
     *
     * @param url Source url wannabe
     */
    void setSourceUrl(const KURL &url) { m_sourceUrl = url; }
    
    /**
     * Set the destination url of this transfer.
     *
     * @param url Destination url wannabe
     */
    void setDestUrl(const KURL &url) { m_destUrl = url; }
    
    /**
     * Returns the size of this file/directory.
     *
     * @return File/dir size
     */
    filesize_t getSize() { return m_size; }
    
    /**
     * Returns the actual size - that is usefull only if this is a directory since
     * it returns the current size of all its items (if some items were removed
     * from its first scan getSize() will return the initial size and getActualSize()
     * will return current size of all items). If this is not a directory, this
     * will return the same value as getSize();
     *
     * @return Actual directory size
     */
    filesize_t getActualSize() { return m_actualSize; }
    
    /**
     * Returns the already transfered file/dir size.
     *
     * @return Transfered file/dir size
     */
    filesize_t getCompleted() { return m_completed; }
    
    /**
     * Returns the number of bytes that have been resumed (using REST).
     *
     * @return Resume offset
     */
    filesize_t getResumed() { return m_resumed; }
    
    /**
     * Get current transfer speed or 0 if the transfer is stalled.
     *
     * @return Transfer speed
     */
    filesize_t getSpeed() { return m_speed; }
    
    /**
     * Adds size bytes to the current transfer size. This will also update all
     * parent transfers (if any).
     *
     * @param size Size to add
     */
    void addSize(filesize_t size);
    
    /**
     * Adds completed bytes to the current completed size. This will also update all
     * parent transfers (if any).
     *
     * @param completed Size to add
     */
    void addCompleted(filesize_t completed);
    
    /**
     * Set the current transfer speed. This will also update all parent transfers.
     *
     * @param speed Speed to set
     */
    void setSpeed(filesize_t speed);
    
    /**
     * Returns the KFTPQueue::Transfer::Type of this transfer. This can either be
     * File or Directory.
     *
     * @return Transfer type
     */
    Type getType() { return m_type; }
    
    /**
     * Return the KFTPQueue::TransferType -- that is if this transfer is an Upload, Download
     * or FXP transfer.
     *
     * @return Upload, Download or FXP
     */
    TransferType getTransferType() { return m_transferType; }
    
    /**
     * Set current KFTPQueue::TransferType -- that is Upload, Download or FXP
     *
     * @param type Upload, Download or FXP
     */
    void setTransferType(TransferType type) { m_transferType = type; }
    
    /**
     * Get transfer status.
     *
     * @return Status of this transfer
     */
    Status getStatus() { return m_status; }
    
    /**
     * Is this transfer a child of another transfer ?
     *
     * @return true if this transfer is a child of another KFTPQueue::Transfer
     */
    bool isChild() { return parent()->inherits("KFTPQueue::Transfer"); }
    
    /**
     * Is this transfer currently running ?
     *
     * @return true if this transfer's status is set to Running
     */
    bool isRunning() { return m_status == Running; }
    
    /**
     * Is this transfer a directory ?
     *
     * @return true if this transfer's type is set to Directory
     */
    bool isDir() { return m_type == Directory; }
    
    /**
     * Is this transfer marked for deletion ?
     *
     * @return true if this transfer is marked for deletion
     */
    bool isDeleteMarked() { return m_deleteMe; }
    
    /**
     * Get the transfer's parent transfer.
     *
     * @return Transfer's parent or NULL if isChild() returns false
     */
    Transfer *getTransferParent();
    
    /**
     * Set transfer's ID.
     *
     * @param Transfer's ID -- it *MUST* be unique.
     */
    void setId(long id) { m_id = id; }
    
    /**
     * Get transfer's ID.
     *
     * @return Transfer's unique ID number
     */
    long getId() { return m_id; }
    
    /**
     * Delays transfer execution for msec miliseconds. If this number is greater than
     * 1000, it will be set to 1000.
     *
     * @param msec Number of miliseconds to delay execution
     */
    void delayedExecute(int msec = 100);
    
    /**
     * Abort current transfer.
     */
    virtual void abort();
    
    /**
     * Just emits the transferUpdate() signal.
     */
    void emitUpdate() { emit transferUpdate(); }
public slots:
    virtual void execute();
protected:
    long m_id;
    bool m_deleteMe;
    bool m_aborting;
    
    Status m_status;
    Type m_type;
    TransferType m_transferType;
    
    /* Source/destination URL */
    KURL m_sourceUrl;
    KURL m_destUrl;
    
    /* File/dir information */
    filesize_t m_size;
    filesize_t m_actualSize;
    filesize_t m_completed;
    filesize_t m_resumed;
    filesize_t m_speed;
    
    void showTransCompleteBalloon();
    void resetTransfer();
    void update();
    
    void addActualSize(filesize_t size);
signals:
    void transferUpdate();
    
    void transferStart(long id);
    void transferComplete(long id);
    void transferAbort(long id);
};

}

#endif
