/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftptransferdir.h"
#include "kftpqueue.h"
#include "kftpsession.h"

namespace KFTPQueue {

TransferDir::TransferDir(QObject *parent)
  : Transfer(parent, Transfer::Directory), m_childIterator(QPtrListIterator<Transfer>(m_children))
{
  m_dirScanDone = false;
}

TransferDir::~TransferDir()
{
  // Abort any dir scans that might be in progress
  KFTPSession *session = FTPSessionManager->find(this);
  
  if (session) {
    session->abort();
  }
}

void TransferDir::execute()
{
  if (!m_dirScanDone && !isChild() && getRemainingChildren() == 0) {
    // We need to scan the directory if it contains some files =)
    KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, m_sourceUrl);
    session->scanDirectory(this);
    m_dirScanDone = true;
    
    connect(session, SIGNAL(dirScanDone()), this, SLOT(slotDirScanDone()));
    return;
  }
  
  m_status = Running;
  
  if (!m_childIterator.toFirst()) {
    // We are finished
    m_deleteMe = true;
    resetTransfer();
    
    emit transferComplete(m_id);
    update();
    return;
  }
  
  executeChild();
}

void TransferDir::abort()
{
  // If not running, just return
  if (!isRunning()) return;
  
  Transfer::abort();
  
  // Signal abort to all child transfers
  if (!m_deleteMe) {
    // Only if we actualy have children
    if (getRemainingChildren() > 0) {
      Transfer *i;
  
      for (i = m_children.first(); i; i = m_children.next()) {
        if (i->isRunning())
          i->abort();
      }
    }
    
    emit transferAbort(m_id);
  }
  
  m_dirScanDone = false;
  resetTransfer();
  update();
}

void TransferDir::executeChild()
{
  Transfer *exec = m_childIterator.current();

  if (!exec) {
    qDebug("--- ID: %ld", m_id);
    qDebug("No transfers left ?!? This is a BUG!");
    return;
  }
  
  connect(exec, SIGNAL(transferComplete(long)), this, SLOT(slotChildFinished()));
  connect(exec, SIGNAL(transferAbort(long)), this, SLOT(slotChildAborted()));
  
  exec->delayedExecute();
}

void TransferDir::addChildTransfer(Transfer *transfer)
{
  m_children.append(transfer);
  
  connect(transfer, SIGNAL(destroyed(QObject*)), this, SLOT(slotChildDestroyed(QObject*)));
}

void TransferDir::delChildTransfer(Transfer *transfer)
{
  m_children.remove(transfer);
}

void TransferDir::slotChildAborted()
{
  if (m_aborting) {
    // We are aborting as well, so we shouldn't take any further actions
    return;
  } else {
    slotChildFinished();
  }
}

void TransferDir::slotChildFinished()
{
  // Our child is finished -- if there are no children left, then we are finished
  // as well.
  if (!++m_childIterator) {
    // We are finished
    showTransCompleteBalloon();
    m_deleteMe = true;
    resetTransfer();
    
    emit transferComplete(m_id);
    FTPQueueManager->doEmitUpdate();
  } else {
    // Move to the next
    executeChild();
  }
}

Transfer *TransferDir::findChildTransfer(long id)
{
  QPtrListIterator<Transfer> it(m_children);
  Transfer *i;

  while ((i = it.current()) != 0) {
    ++it;
    
    if (i->getId() == id)
      return i;
      
    if (i->isDir()) {
      Transfer *tmp = static_cast<TransferDir*>(i)->findChildTransfer(id);
      
      if (tmp)
        return tmp;
    }
  }
  
  return NULL;
}

void TransferDir::removeMarkedChildren()
{
  QPtrListIterator<Transfer> it(m_children);
  Transfer *i;

  while ((i = it.current()) != 0) {
    ++it;
    
    if (i->isDir())
      static_cast<TransferDir*>(i)->removeMarkedChildren();
    
    if (i->isDeleteMarked())
      FTPQueueManager->removeTransfer(i);
  }
}

void TransferDir::moveChildUp(Transfer *child)
{
  if (m_children.findRef(child) != -1) {
    if (m_children.prev()) {
      int prevPos = m_children.at();
      m_children.remove(child);
      m_children.insert(prevPos, child);
    }
  }
}

void TransferDir::moveChildDown(Transfer *child)
{
  if (m_children.findRef(child) != -1) {
    if (m_children.next()) {
      int nextPos = m_children.at();
      m_children.remove(child);
      m_children.insert(nextPos, child);
    }
  }
}

bool TransferDir::canMoveChildUp(Transfer *child)
{
  if (child->isRunning()) return false;
  if (m_children.getFirst() == child) return false;
  
  return true;
}

bool TransferDir::canMoveChildDown(Transfer *child)
{
  if (child->isRunning()) return false;
  if (m_children.getLast() == child) return false;
  
  return true;
}

void TransferDir::slotChildDestroyed(QObject *child)
{
  // Remove the transfer
  delChildTransfer(static_cast<Transfer*>(child));
}

void TransferDir::slotDirScanDone()
{
  // Directory scanning has been completed, execute the transfer
  delayedExecute();
}

}
