/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftptransferfile.h"
#include "kftpsystemtray.h"
#include "kftpconfiguration.h"
#include "kftpsession.h"

#include <kmessagebox.h>
#include <klocale.h>
#include <kio/renamedlg.h>
#include <kdiskfreesp.h>

#include <qtimer.h>
#include <qfileinfo.h>

namespace KFTPQueue {

TransferFile::TransferFile(QObject *parent)
  : Transfer(parent, Transfer::File),
    m_srcSession(0),
    m_dstSession(0),
    m_updateTimer(0),
    m_dfTimer(0)
{
}

TransferFile::~TransferFile()
{
}

void TransferFile::execute()
{
  if (!m_srcSession) {
    m_srcSession = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, m_sourceUrl, QDomNode(), true);
    m_dstSession = FTPSessionManager->spawnRemoteSession(OPPOSITE_SIDE(m_srcSession->getSide()), m_destUrl, QDomNode(), true);
    
    // Try to lock both sessions
    if (!m_srcSession->canLock() || !m_dstSession->canLock()) {
      // This should not happen, since session manager will always return an unlocked session (as specified
      // by the mustUnlock parameter).
      qDebug("BUG: Unable to lock sessions, will not execute.");
      
      resetTransfer();
      return;
    } else {
      // Ok, sessions are available - lock them!
      m_srcSession->lock(this);
      m_dstSession->lock(this);
    }
    
    m_completed = 0;
    m_resumed = 0;
  
    // Connect events
    connect(remoteSession()->getClient(), SIGNAL(finished(CommandType)), this, SLOT(slotCommandFinished(CommandType)));
    connect(remoteSession()->getClient(), SIGNAL(resumedOffset(filesize_t)), this, SLOT(slotResumedOffset(filesize_t)));
    
    connect(m_srcSession->getClient(), SIGNAL(loginComplete(bool)), this, SLOT(slotClientConnected(bool)));
    connect(m_dstSession->getClient(), SIGNAL(loginComplete(bool)), this, SLOT(slotClientConnected(bool)));
  }
  
  // If any session is not yet connected, wait for the connection
  if (!m_srcSession->isConnected() || !m_dstSession->isConnected()) {
    m_status = Connecting;
    return;
  }
  
  // We should get file's stat information if this is requested
  if (m_status != DoingStat && !FTPConfig().m_overwriteFiles && m_transferType != FXP && (m_transferType != Download || QFileInfo(m_destUrl.path()).exists()))
  {
    m_status = DoingStat;

    switch(m_transferType) {
      case Download: remoteSession()->getClient()->stat(m_sourceUrl); break;
      case Upload: remoteSession()->getClient()->stat(m_destUrl); break;
      case FXP: break;
    }

    // When stat finishes execute will be recalled
    connect(remoteSession()->getClient(), SIGNAL(statResult(FTPEntry)), this, SLOT(slotFileStatDone(FTPEntry)));
    
    return;
  }
  
  // We are running now
  m_status = Running;
  
  // Init timer to follow the update
  m_updateTimer = new QTimer(this);
  connect(m_updateTimer, SIGNAL(timeout()), this, SLOT(slotTimerUpdate()));
  m_updateTimer->start(1000); // update status every 1 sec
  
  // Should we check for free space ?
  if (FTPConfig().m_diskCheckSpace) {
    m_dfTimer = new QTimer(this);
    connect(m_dfTimer, SIGNAL(timeout()), this, SLOT(slotTimerDiskFree()));
    m_dfTimer->start(FTPConfig().m_diskCheckInterval * 1000);
  }
  
  emit transferStart(m_id);
  
  switch(m_transferType) {
    case Download: {
      m_finishCmd = TCMD_GET;
      m_srcSession->getClient()->get(m_sourceUrl, m_destUrl);
      break;
    }
    case Upload: {
      m_finishCmd = TCMD_PUT;
      m_dstSession->getClient()->put(m_sourceUrl, m_destUrl);
      break;
    }
    case FXP: {
      // This is a FXP transfer, so there are no status messages - disable timers
      m_updateTimer->stop();
      m_updateTimer->QObject::disconnect();
      
      // Emit transfer update, so some stuff will be displayed by the queueview
      emit transferUpdate();

      m_finishCmd = TCMD_FXP;
      m_srcSession->getClient()->fxpTransfer(m_sourceUrl, m_destUrl, m_dstSession->getClient());
      break;
    }
  }
}

KFTPSession *TransferFile::remoteSession()
{
  if (m_srcSession->isRemote())
    return m_srcSession;
  else
    return m_dstSession;
}

void TransferFile::slotTimerUpdate()
{
  // Update the current stats
  if (!m_srcSession) {
    m_updateTimer->stop();
    m_updateTimer->QObject::disconnect();
    return;
  }

  addCompleted(remoteSession()->getClient()->getClient()->getProcessed() - m_completed);
  setSpeed(remoteSession()->getClient()->getClient()->getSpeed());
  
  update();
}

void TransferFile::slotTimerDiskFree()
{
  // Check for disk usage
  if (FTPConfig().m_diskCheckSpace) {
    KDiskFreeSp *df = KDiskFreeSp::findUsageInfo((getDestUrl().path()));
    connect(df, SIGNAL(foundMountPoint(const QString&, unsigned long, unsigned long, unsigned long)), this, SLOT(slotDiskFree(const QString&, unsigned long, unsigned long, unsigned long)));
  }
}

void TransferFile::slotDiskFree(const QString &mountPoint, unsigned long kBSize, unsigned long kBUsed, unsigned long kBAvail)
{
  if (FTPConfig().m_diskCheckSpace) {
    // Is there enough free space ?
    if (kBAvail < FTPConfig().m_diskMinFreeSpace) {
      QString p_transAbortStr = i18n("Transfer of the following files <b>has been aborted</b> because there is not enough free space left on '%1':").arg(mountPoint);
      p_transAbortStr += "<br><i>";
      p_transAbortStr += getSourceUrl().fileName();
      p_transAbortStr += "</i>";
      s_sysTray->showBalloon(p_transAbortStr);
      
      // Abort the transfer
      abort();
    }
  }
}

void TransferFile::slotCommandFinished(CommandType cmd)
{
  if (cmd == m_finishCmd) {
    m_updateTimer->stop();
    m_updateTimer->QObject::disconnect();
    
    showTransCompleteBalloon();
    m_deleteMe = true;
    addActualSize(-m_size);
    
    resetTransfer();
    emit transferComplete(m_id);
    
    FTPQueueManager->doEmitUpdate();
  }
}

void TransferFile::slotFileStatDone(FTPEntry e)
{
  // We have stat for the file
  if (m_status != DoingStat)
    return;

  if (e.name.isEmpty()) {
    // File does not exist
  } else {
    QString localPath = (m_transferType == Download) ? m_destUrl.path() : m_sourceUrl.path();
    QFileInfo p_fInfo(localPath);

    if (p_fInfo.exists()) {
      // Display a dialog
      filesize_t srcSize = (m_transferType == Download) ? e.size : p_fInfo.size();
      filesize_t dstSize = (m_transferType == Download) ? p_fInfo.size() : e.size;
      time_t srcTime = (m_transferType == Download) ? e.date : p_fInfo.created().toTime_t();
      time_t dstTime = (m_transferType == Download) ? p_fInfo.created().toTime_t() : e.date;
      
      FileExistsActions fa = (m_transferType == Download) ? FTPConfig().m_fileExistsDownActions : FTPConfig().m_fileExistsUpActions;
      FEAction act = fa.getActionForSituation(srcSize, srcTime, dstSize, dstTime);
      switch (act) {
        case FE_SKIP_ACT: {
          m_deleteMe = true;
          addActualSize(-m_size);
          
          resetTransfer();
          emit transferComplete(m_id);
          
          FTPQueueManager->doEmitUpdate();
          return;
        }
        case FE_OVERWRITE_ACT: {
          remoteSession()->getClient()->getClient()->setConfig("feat_resume", 0);
          break;
        }
        case FE_RESUME_ACT: {
          remoteSession()->getClient()->getClient()->setConfig("feat_resume", 1);
          break;
        }
        case FE_RENAME_ACT: {
          // TODO implement ranaming
        }
        case FE_USER_ACT: {
          QString p_newDestPath;
          KIO::RenameDlg_Result r = KIO::open_RenameDlg(
            i18n("File exists"),
            m_sourceUrl.prettyURL(),
            m_destUrl.prettyURL(),
            (KIO::RenameDlg_Mode) (KIO::M_OVERWRITE | KIO::M_RESUME),
            p_newDestPath,
            srcSize,
            dstSize,
            srcTime,
            dstTime
          );
    
          switch (r) {
            case KIO::R_RESUME: {
              remoteSession()->getClient()->getClient()->setConfig("feat_resume", 1);
              break;
            }
            case KIO::R_OVERWRITE: {
              remoteSession()->getClient()->getClient()->setConfig("feat_resume", 0);
              break;
            }
            case KIO::R_RENAME: {
              m_destUrl.setPath(p_newDestPath);
              update();
              break;
            }
            case KIO::R_CANCEL: {
              resetTransfer();
              
              emit transferComplete(m_id);
              FTPQueueManager->doEmitUpdate();
              return;
            }
            default: break;
          }
          
          break;
        }
      }
      
    }
  } // e.name check

  // Now, re-execute the transfer
  execute();
}

void TransferFile::slotResumedOffset(filesize_t offset)
{
  m_resumed = offset;
}

void TransferFile::slotClientConnected(bool success)
{
  if (success) {
    if (!m_srcSession->isConnected() || !m_dstSession->isConnected()) return;
    
    // Ok, both sessions are ready -- start the transfer
    m_status = Running;
    delayedExecute();
  } else {
    // Unable to login - abort transfer
    abort();
  }
}

void TransferFile::resetTransfer()
{
  // Unlock the sessions (they should be unlocked automaticly when the transferComplete signal
  // is emitted, but when a transfer is a child transfer, the next transfer may need the session
  // sooner). Also sessions should be unlocked when transfer aborts.
  if (m_srcSession && m_dstSession) {
    // Disconnect signals
    remoteSession()->getClient()->QObject::disconnect(this, SLOT(slotCommandFinished(CommandType)));
    remoteSession()->getClient()->QObject::disconnect(this, SLOT(slotFileStatDone(FTPEntry)));
    remoteSession()->getClient()->QObject::disconnect(this, SLOT(slotResumedOffset(filesize_t)));
    
    m_srcSession->getClient()->QObject::disconnect(this, SLOT(slotClientConnected(bool)));
    m_dstSession->getClient()->QObject::disconnect(this, SLOT(slotClientConnected(bool)));
  
    m_srcSession->unlock();
    m_dstSession->unlock();
  }
  
  m_srcSession = 0L;
  m_dstSession = 0L;
  m_finishCmd = TCMD_NONE;
  
  Transfer::resetTransfer();
}

void TransferFile::abort()
{
  // If not running, just return
  if (!isRunning()) return;
  
  Transfer::abort();
  
  if (m_updateTimer) {
    m_updateTimer->stop();
    m_updateTimer->QObject::disconnect();
    
    delete m_updateTimer;
    m_updateTimer = 0L;
  }
  
  if (m_dfTimer) {
    m_dfTimer->stop();
    m_dfTimer->QObject::disconnect();
    
    delete m_dfTimer;
    m_dfTimer = 0L;
  }

  // Abort any transfers
  m_srcSession->abort();
  m_dstSession->abort();
  
  // Update everything
  resetTransfer();
  update();
  
  if (isChild())
    disconnect(parent());
}

}
