/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2004 by the KFTPGrabber developers
 * Copyright (C) 2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftphowlthread.h"

#include <qapplication.h>

KFTPHowlThread::KFTPHowlThread()
 : QObject(), QThread()
{
  m_howlEnabled = false;
  
  qDebug("[HOWL] Thread init done.");
}

void KFTPHowlThread::run()
{
  QString type("_ftp._tcp.");
  qDebug("[HOWL] Thread run in progress...");
  
  // Begin with initialization
  if ((m_howlResult = sw_discovery_init(&m_howlDiscovery)) != SW_OKAY) {
    // Error while initing
    qDebug("[HOWL] Error in howl initialization (err=%d)!", (int) m_howlResult);
    return;
  }
  
  // Init the salt
  if (sw_discovery_salt(m_howlDiscovery, &m_howlSalt) != SW_OKAY) {
    qDebug("[HOWL] Error while getting salt!");
    return;
  }
  
  // Init the callbacks
  if (sw_discovery_browse(m_howlDiscovery, 0, type, NULL, &howlBrowser, this, &m_howlId) != SW_OKAY) {
    qDebug("[HOWL] Error in howl discovery!");
    return;
  }
  
  // Set the enabled flag
  m_howlEnabled = true;
  
  // Enter the event dispatch loop
  while (1) {
    sw_ulong msecs = 500;
    if (sw_salt_step(m_howlSalt, &msecs) != SW_OKAY) {
      qWarning("[HOWL] Error in salt operations, aborting ZeroConf discovery...");
      return;
    }
  }
}

bool KFTPHowlThread::isHowlEnabled()
{
  return m_howlEnabled;
}

sw_result HOWL_API KFTPHowlThread::howlBrowser (
        sw_discovery discovery,
        sw_discovery_oid oid,
        sw_discovery_browse_status status,
        sw_uint32 interface_index,
        sw_const_string name,
        sw_const_string type,
        sw_const_string domain,
        sw_opaque extra)
{
  sw_discovery_resolve_id rid;

  switch (status) {
    case SW_DISCOVERY_BROWSE_ADD_SERVICE: {
      // New Zeroconf service has been published
      if (sw_discovery_resolve(discovery, interface_index, name, type, domain, howlResolver, extra, &rid) != SW_OKAY) {
        qDebug("[HOWL] Resolve failed!");
      }
      break;
    }
    case SW_DISCOVERY_BROWSE_REMOVE_SERVICE: {
      // Find the service and remove it from our list of services
      QValueList<KFTPHowlService> *list = &((KFTPHowlThread*) extra)->m_services;
      QValueList<KFTPHowlService>::iterator i;
      for (i = list->begin(); i != list->end(); ++i) {
        if ((*i).getServiceName() == name) {
          list->remove(i);
          
          // Signal the update
          emit ((KFTPHowlThread*) extra)->_howlListUpdate();
          break;
        }
      }
      break;
    }
    default: {
      // Do nothing :)
    }
  }

  return SW_OKAY;
}

sw_result HOWL_API KFTPHowlThread::howlResolver (
        sw_discovery discovery,
        sw_discovery_oid oid,
        sw_uint32 interface_index,
        sw_const_string name,
        sw_const_string type,
        sw_const_string domain,
        sw_ipv4_address address,
        sw_port port,
        sw_octets text_record,
        sw_ulong text_record_len,
        sw_opaque extra)
{
  char name_buf[16];

  sw_discovery_cancel(discovery, oid);
  
  // Add the service to our list
  ((KFTPHowlThread*) extra)->m_services.append(KFTPHowlService(name, sw_ipv4_address_name(address, name_buf, 16), port));

  // Signal update
  emit ((KFTPHowlThread*) extra)->_howlListUpdate();

  return SW_OKAY;
}

KFTPHowlEventForwarder::KFTPHowlEventForwarder()
  : QObject()
{
  // Connect the thread signal
  connect(&m_thread, SIGNAL(_howlListUpdate()), this, SLOT(slotListUpdate()));
}

void KFTPHowlEventForwarder::slotListUpdate()
{
  // Create the event and dispatch it to ourself
  HowlEvent *e = new HowlEvent();
  qApp->postEvent(this, e);
}

void KFTPHowlEventForwarder::customEvent(QCustomEvent *e)
{
  // We just forward events :)
  if (e->type() == HOWL_EVENT_ID) {
    emit howlListUpdate();
  }
}

