/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpapi.h"
#include "kftpconfiguration.h"

#include <kapplication.h>
#include <kconfig.h>
#include <kemailsettings.h>
#include <klocale.h>

#include <qdir.h>
#include <qregexp.h>

using namespace KFTPQueue;

/*
 *                             KFTPFileViewColorizer class
 */

KFTPFileViewColorizer *FTPFileColorizer = 0L;

KFTPFileViewColorizer::KFTPFileViewColorizer(QObject *parent)
  : QObject(parent)
{
}

void KFTPFileViewColorizer::addFilePattern(const QString &pattern, const QColor &color)
{
  PatternStruct _ps;
  _ps.pattern = pattern;
  _ps.color = color;
  m_filePatterns.append(_ps);
}

QColor KFTPFileViewColorizer::getColor(const QString &name)
{
  if (!FTPConfig().m_enableHighlight) return QColor();

  QRegExp e;
  e.setWildcard(true);
  for (unsigned int i = 0; i <= m_filePatterns.count(); i++) {
    PatternStruct _ps = *m_filePatterns.at(i);
    e.setPattern(_ps.pattern);
    if (e.exactMatch(name))
      return _ps.color;
  }

  return QColor();
}

/*
 *                                KFTPConfiguration class
 */

KFTPConfiguration::KFTPConfiguration()
  : QObject()
{
  m_fileExistsDownActions.setTypeText(i18n("Download"));
  m_fileExistsUpActions.setTypeText(i18n("Upload"));
  
  read();
}

void KFTPConfiguration::emitChange()
{
  emit configChanged();
}

void KFTPConfiguration::loadToColorizer()
{
  FTPFileColorizer->m_filePatterns.clear();

  for (QStringList::Iterator i = m_filterList.begin(); i != m_filterList.end(); ++i) {
    QString pat = (*i).section(':', 0, 0);
    QString col = (*i).section(':', 1, 1);

    FTPFileColorizer->addFilePattern(pat, QColor(col));
  }
}

char KFTPConfiguration::ftpMode(const QString &filename)
{
  // Get FTP mode (binary/ascii)
  switch (m_transMode) {
    case 'A': return 'A'; break;
    case 'I': return 'I'; break;
    case 'X':
    default: {
      char mode = 'I';
      QRegExp e;
      e.setWildcard(true);
      for (QStringList::iterator i = m_asciiList.begin(); i != m_asciiList.end(); ++i) {
        e.setPattern((*i));
        if (e.exactMatch(filename)) {
          mode = 'A';
          break;
        }
      }
    
      return mode;
    }
  }
}

bool KFTPConfiguration::isOnSkipList(const QString &filename)
{
  // Is a file on skiplist ?
  QRegExp e;
  e.setWildcard(true);
  for (QStringList::iterator i = m_skipList.begin(); i != m_skipList.end(); ++i) {
    e.setPattern((*i));
    
    if (e.exactMatch(filename))
      return true;
  }

  return false;
}

QString KFTPConfiguration::getGlobalMail()
{
  KEMailSettings kes;
  kes.setProfile( kes.defaultProfileName() );
  return kes.getSetting( KEMailSettings::EmailAddress );
}

void KFTPConfiguration::read()
{
  KConfig *conf = kapp->config();

  // Main Window
  conf->setGroup("MainWindow");
  m_size = conf->readSizeEntry("size", new QSize(800, 500));
  m_position = conf->readPointEntry("position", new QPoint(0,0));
  
  // General
  conf->setGroup("General");
  m_retryCount = conf->readNumEntry("retryCount", 10);
  m_retryTime = conf->readNumEntry("retryTime", 60);
  m_useCache = conf->readBoolEntry("useCache", false);
  m_showBalloons = conf->readBoolEntry("showBalloons", true);
  m_showBalloonWhenQueueEmpty = conf->readBoolEntry("showBalloonWhenQueueEmpty", false);
  m_showRetrySuccessBalloon = conf->readBoolEntry("showRetrySuccessBalloon", true);
  m_confirmExit = conf->readBoolEntry("confirmExit", true);
  m_encryptBookmarks = conf->readBoolEntry("encryptBookmarks", false);
  m_defLocalDir = conf->readEntry("defLocalDir", QDir::homeDirPath());
  m_startMinimized = conf->readBoolEntry("startMinimized", false);
  m_showSplash = conf->readBoolEntry("showSplash", true);
  
  // Actions
  conf->setGroup("Actions");
  QString tmp = conf->readEntry("downloadActions", "4;4;4;4;4;4;4;4;4;");
  tmp >> m_fileExistsDownActions;
  
  tmp = conf->readEntry("uploadActions", "4;4;4;4;4;4;4;4;4;");
  tmp >> m_fileExistsUpActions;

  // Filters
  conf->setGroup("Filters");
  m_filterList = conf->readListEntry("filterList");
  m_enableHighlight = conf->readBoolEntry("enableHighlight", true);
  m_asciiList = conf->readListEntry("asciiList");
  m_enableSkipList = conf->readBoolEntry("enableSkipList", true);
  m_skipEmptyFiles = conf->readBoolEntry("skipEmptyFiles", false);
  m_skipEmptyDirs = conf->readBoolEntry("skipEmptyDirs", false);
  m_applySkipToDirs = conf->readBoolEntry("applySkipToDirs", false);
  m_skipNoQueue = conf->readBoolEntry("skipNoQueue", true);
  m_skipList = conf->readListEntry("skipList");

  // Log
  conf->setGroup("Log");
  m_logFont = conf->readFontEntry("logFont");
  m_logCommandsColor = conf->readColorEntry("logCommandsColor", new QColor(67, 170, 23));
  m_logResponsesColor = conf->readColorEntry("logResponsesColor", new QColor(0, 0, 255));
  m_logMultilineColor = conf->readColorEntry("logMultilineColor", new QColor(148, 188, 22));
  m_logErrorColor = conf->readColorEntry("logErrorColor", new QColor(255, 0, 0));
  m_logStatusColor = conf->readColorEntry("logStatusColor", new QColor(0, 0, 0));
  m_logSaveToFile = conf->readBoolEntry("saveToFile");
  m_logOutputFilename = conf->readEntry("outputFilename", "");

  // Transfers
  conf->setGroup("Transfers");
  m_overwriteFiles = conf->readBoolEntry("overwriteFiles", false);
  m_portForceIp = conf->readBoolEntry("portForceIp", false);
  m_portIp = conf->readEntry("portIp", "");
  m_queueOnDND = conf->readBoolEntry("queueOnDND", true);
  m_lowercaseDownload = conf->readBoolEntry("lowercaseDownload", false);
  m_diskCheckSpace = conf->readBoolEntry("diskCheckSpace", false);
  m_diskCheckInterval = conf->readNumEntry("diskCheckInterval", 60);
  m_diskMinFreeSpace = conf->readNumEntry("diskMinFreeSpace", 500);
  m_globalMail = conf->readBoolEntry("globalMail", true);
  m_anonMail = conf->readEntry("anonMail", getGlobalMail());
    
  // Display
  conf->setGroup("Display");
  m_showTree = conf->readBoolEntry("showTree", true);
  m_showHiddenFiles = conf->readBoolEntry("showHiddenFiles", true);
  m_showSizeInBytes = conf->readBoolEntry("showSizeInBytes", true);

  // Commit to colorizer so it will actualy colorize :)
  if (FTPFileColorizer == 0L) {
    // Create the colorizer
    FTPFileColorizer = new KFTPFileViewColorizer();
  }
  loadToColorizer();
}

void KFTPConfiguration::write()
{
  KConfig *conf = kapp->config();

  // Main Window
  conf->setGroup("MainWindow");
  conf->writeEntry("size", KFTPAPI::getInstance()->mainWindow()->size());
  conf->writeEntry("position", KFTPAPI::getInstance()->mainWindow()->pos());
  
  // General
  conf->setGroup("General");
  conf->writeEntry("retryCount", m_retryCount);
  conf->writeEntry("retryTime", m_retryTime);
  conf->writeEntry("useCache", m_useCache);
  conf->writeEntry("showBalloons", m_showBalloons);
  conf->writeEntry("showBalloonWhenQueueEmpty", m_showBalloonWhenQueueEmpty);
  conf->writeEntry("showRetrySuccessBalloon", m_showRetrySuccessBalloon);
  conf->writeEntry("confirmExit", m_confirmExit);
  conf->writeEntry("encryptBookmarks", m_encryptBookmarks);
  conf->writeEntry("defLocalDir", m_defLocalDir);
  conf->writeEntry("startMinimized", m_startMinimized);
  conf->writeEntry("showSplash", m_showSplash);
  
  // Actions
  conf->setGroup("Actions");
  QString tmp;
  tmp << m_fileExistsDownActions;
  conf->writeEntry("downloadActions", tmp);
  
  tmp << m_fileExistsUpActions;
  conf->writeEntry("uploadActions", tmp);

  // Filters
  conf->setGroup("Filters");
  conf->writeEntry("filterList", m_filterList);
  conf->writeEntry("enableHighlight", m_enableHighlight);
  conf->writeEntry("asciiList", m_asciiList);
  conf->writeEntry("enableSkipList", m_enableSkipList);
  conf->writeEntry("skipEmptyFiles", m_skipEmptyFiles);
  conf->writeEntry("skipEmptyDirs", m_skipEmptyDirs);
  conf->writeEntry("applySkipToDirs", m_applySkipToDirs);
  conf->writeEntry("skipNoQueue", m_skipNoQueue);
  conf->writeEntry("skipList", m_skipList);

  // Log
  conf->setGroup("Log");
  conf->writeEntry("logFont", m_logFont);
  conf->writeEntry("logCommandsColor", m_logCommandsColor);
  conf->writeEntry("logResponsesColor", m_logResponsesColor);
  conf->writeEntry("logMultilineColor", m_logMultilineColor);
  conf->writeEntry("logErrorColor", m_logErrorColor);
  conf->writeEntry("logStatusColor", m_logStatusColor);
  conf->writeEntry("saveToFile", m_logSaveToFile);
  conf->writeEntry("outputFilename", m_logOutputFilename);

  // Transfers
  conf->setGroup("Transfers");
  conf->writeEntry("overwriteFiles", m_overwriteFiles);
  conf->writeEntry("portForceIp", m_portForceIp);
  conf->writeEntry("portIp", m_portIp);
  conf->writeEntry("queueOnDND", m_queueOnDND);
  conf->writeEntry("lowercaseDownload", m_lowercaseDownload);
  conf->writeEntry("diskCheckSpace", m_diskCheckSpace);
  conf->writeEntry("diskCheckInterval", m_diskCheckInterval);
  conf->writeEntry("diskMinFreeSpace", m_diskMinFreeSpace);
  conf->writeEntry("globalMail", m_globalMail);
  conf->writeEntry("anonMail", m_anonMail);
  
  // Display
  conf->setGroup("Display");
  conf->writeEntry("showTree", m_showTree);
  conf->writeEntry("showHiddenFiles", m_showHiddenFiles);
  conf->writeEntry("showSizeInBytes", m_showSizeInBytes);
}

KFTPConfiguration& FTPConfig() {
    static KFTPConfiguration conf;
    return conf;
}

