/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2004 by the KFTPGrabber developers
 * Copyright (C) 2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpscripteditor.h"

#ifdef HAVE_QSA

#include "kftpscripting.h"
#include "kftplistview.h"
#include "misc.h"

#include <klocale.h>
#include <ktabwidget.h>
#include <kmessagebox.h>

#include <qsplitter.h>
#include <qlayout.h>

// QSA includes
#include <qsinterpreter.h>
#include <qsscript.h>

using namespace KFTPGrabberBase;

KFTPScriptEditor::KFTPScriptEditor(QWidget *parent, const char *name)
 : KDialogBase(Plain, i18n("Script and event editor"), Ok|Cancel, Ok, parent, name)
{
  setupGUI();
  fillEvents();
  
  // Connect slots
  connect(m_eventView, SIGNAL(executed(QListViewItem*)), this, SLOT(slotEventClicked(QListViewItem*)));
}


KFTPScriptEditor::~KFTPScriptEditor()
{
}

void KFTPScriptEditor::setupGUI()
{
  setInitialSize(QSize(800, 400));
  
  QGridLayout *layout = new QGridLayout(plainPage());
  QSplitter *splitter = new QSplitter(plainPage());
  layout->addWidget(splitter, 0, 0);
  setMainWidget(splitter);
  
  // Create the event list view
  m_eventView = new KFTPListView(splitter);
  splitter->moveToFirst(m_eventView);
  splitter->setResizeMode(m_eventView, QSplitter::KeepSize);
  
  m_eventView->setMinimumWidth(200);
  m_eventView->setFullWidth(true);
  m_eventView->addColumn(i18n("Event name"));
  m_eventView->setRootIsDecorated(true);
  m_eventView->setEmptyListText(i18n("No events"));
  m_eventView->setItemsRenameable(false);
  
  // Create the tab widget
  KTabWidget *tabs = new KTabWidget(splitter);
  splitter->moveToLast(tabs);
  
  QSProject *project = KFTPScripting::getInstance()->m_project;
  
  m_editor = project->createEditor(project->script("main.qs"), tabs);
  tabs->addTab(m_editor, i18n("Script editor"));
}

void KFTPScriptEditor::fillEvents()
{
  // Just fill in all the events names
  
  // GUI EVENTS
  KListViewItem *evGui = new KListViewItem(m_eventView, "GUI related events");
  evGui->setOpen(true);
  evGui->setPixmap(0, loadSmallPixmap("folder"));
  new KListViewItem(evGui, "OnAppStartup");
  new KListViewItem(evGui, "OnAppExit");
  
  // TRANSFER EVENTS
  KListViewItem *evTransfer = new KListViewItem(m_eventView, "Transfer events");
  evTransfer->setOpen(true);
  evTransfer->setPixmap(0, loadSmallPixmap("folder"));
  new KListViewItem(evTransfer, "OnTransferComplete", "(QID)");
  new KListViewItem(evTransfer, "OnTransferAbort", "(QID)");
  new KListViewItem(evTransfer, "OnTransferStart", "(QID)");
  new KListViewItem(evTransfer, "OnTransferQueue", "(QID)");
  
  // SESSION EVENTS
  KListViewItem *evSession = new KListViewItem(m_eventView, "Session events");
  evSession->setOpen(true);
  evSession->setPixmap(0, loadSmallPixmap("folder"));
  new KListViewItem(evSession, "OnSessionSpawn", "(session)");
  new KListViewItem(evSession, "OnSessionDestroy", "(session)");
  
  // CLIENT EVENTS
  KListViewItem *evClient = new KListViewItem(m_eventView, "Client events");
  evClient->setOpen(true);
  evClient->setPixmap(0, loadSmallPixmap("folder"));
  new KListViewItem(evClient, "OnClientConnected", "(client)");
  new KListViewItem(evClient, "OnClientDisconnected", "(client)");
  new KListViewItem(evClient, "OnRawCommand", "(client, command)");
}

void KFTPScriptEditor::beforeShow()
{
  // Revert the contents so it will be as it actually is
  m_editor->revert();
  m_oldCode = m_editor->script()->code();
}

void KFTPScriptEditor::slotOk()
{
  // Check the code before commiting
  if (!m_editor->interpreter()->checkSyntax(m_editor->text())) {
    if (KMessageBox::warningYesNo(this, i18n("This code cannot be saved, because it contains syntax errors! Do you want to return to editing?"), i18n("Syntax error")) == KMessageBox::No) {
      m_editor->script()->setCode(m_oldCode);
      accept();
    }
  } else {
    m_editor->commit();
    accept();
  }
}

void KFTPScriptEditor::slotCancel()
{
  m_editor->script()->setCode(m_oldCode);
  reject();
}

void KFTPScriptEditor::slotEventClicked(QListViewItem *item)
{
  // If there is no selected item, return
  if (!item || item->text(0).left(2) != "On")
    return;
  
  // First commit current changes, if there are syntax errors, the code will
  // be dropped when user presses the Ok button ;)
  m_editor->commit();
    
  // Then add some stuff
  QSProject *project = m_editor->project();
  QSInterpreter *interpreter = m_editor->interpreter();
  
  if (interpreter->functions().findIndex(item->text(0)) == -1) {
    QSScript *sc = project->script("main.qs");
    
    // Add a newline at the beginning, otherwise the insterted function
    // will be misinserted ;-)
    sc->addCode("\n\n");
    
    // Create new function
    QString funcName = item->text(0);
    if (!item->text(1).isEmpty()) {
      funcName += item->text(1);
      
      // Add the code
      QString code = "function ";
      code += funcName;
      code += "\n{\n\n}\n";
      
      // Add the code with parameters
      sc->addCode(code);
      
      return;
    }
    
    sc->addFunction(funcName);
  }
}

#include "kftpscripteditor.moc"

#endif

