/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2004 by the KFTPGrabber developers
 * Copyright (C) 2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpscripting.h"
#include "kftpscripteditor.h"

#ifdef HAVE_QSA

#include "misc.h"
#include "kftpapi.h"
#include "kftpqueue.h"
#include "kftpgrabber.h"

#include <qsscript.h>
#include <qsinterpreter.h>
#include <qsutilfactory.h>

/* KFTPGrabber QSA interfaces */
#include "transfersinterface.h"

using namespace KFTPGrabberScript;

KFTPScripting *KFTPScripting::m_instance = 0L;

KFTPScripting *KFTPScripting::getInstance()
{
  if (!m_instance)
    m_instance = new KFTPScripting();

  return m_instance;
}

KFTPScripting::KFTPScripting(QObject *parent, const char *name)
 : QObject(parent, name)
{
  // Create the QSA project
  m_project = new QSProject(this);
  m_project->interpreter()->setErrorMode(QSInterpreter::Notify);
  m_scriptEditor = 0L;
  
  addInterfaces();
  connectQsaSignals();
  
  // Create the main script
  m_project->createScript("main.qs");  
}

KFTPScripting::~KFTPScripting()
{
  m_instance = 0L;
  
  if (m_scriptEditor != 0L) {
    delete m_scriptEditor;
    m_scriptEditor = 0L;
  }
}

void KFTPScripting::addInterfaces()
{
  // Create the interfaces
  TransfersInterface *transIface = new TransfersInterface(this, "transferManager");
  
  // Add some object factories & wrapper factories
  m_project->interpreter()->addObjectFactory(new QSUtilFactory());
  m_project->interpreter()->addWrapperFactory(new Wrappers());
  
  // Add them to the script engine
  m_project->addObject(transIface);
}

void KFTPScripting::loadFromFile()
{
  QString file = KFTPGrabberBase::getStoreDir("scripts");
  m_project->load(file);
}

void KFTPScripting::saveToFile()
{
  m_project->save();
}

void KFTPScripting::showScriptEditor()
{
  if (m_scriptEditor == 0L) {
    m_scriptEditor = new KFTPScriptEditor();
  }
  
  // Exec the dialog
  m_scriptEditor->beforeShow();
  m_scriptEditor->exec();
}

void KFTPScripting::connectQsaSignals()
{
  KFTPGrabber *mainWin = (KFTPGrabber*) KFTPAPI::getInstance()->mainWindow();
  
  connect(mainWin, SIGNAL(qsaAppStartup()), this, SLOT(qsaAppStartup()));
  connect(mainWin, SIGNAL(qsaAppExit()), this, SLOT(qsaAppExit()));
  
  connect(FTPQueueManager, SIGNAL(qsaTransferComplete(long)), this, SLOT(qsaTransferComplete(long)));
  connect(FTPQueueManager, SIGNAL(qsaTransferStart(long)), this, SLOT(qsaTransferStart(long)));
  connect(FTPQueueManager, SIGNAL(qsaTransferAbort(long)), this, SLOT(qsaTransferAbort(long)));
}

void KFTPScripting::qsaAppStartup()
{
  if (m_project->interpreter()->functions().findIndex("OnAppStartup") != -1)
    m_project->interpreter()->call("OnAppStartup");
}

void KFTPScripting::qsaAppExit()
{
  if (m_project->interpreter()->functions().findIndex("OnAppExit") != -1)
    m_project->interpreter()->call("OnAppExit");
}

void KFTPScripting::qsaTransferComplete(long id)
{
  if (m_project->interpreter()->functions().findIndex("OnTransferComplete") != -1)
    m_project->interpreter()->call("OnTransferComplete", QSArgumentList((long long int) id));
}

void KFTPScripting::qsaTransferStart(long id)
{
  if (m_project->interpreter()->functions().findIndex("OnTransferStart") != -1)
    m_project->interpreter()->call("OnTransferStart", QSArgumentList((long long int) id));
}

void KFTPScripting::qsaTransferAbort(long id)
{
  if (m_project->interpreter()->functions().findIndex("OnTransferAbort") != -1)
    m_project->interpreter()->call("OnTransferAbort", QSArgumentList((long long int) id));
}

#include "kftpscripting.moc"

#endif

