/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpconfiguration.h"
#include "kftpconfigdialog.h"
#include "misc.h"

#include "kftpconfiglog.h"
#include "kftpconfigtransfers.h"
#include "kftpconfiggeneral.h"
#include "kftpconfigdisplay.h"

#include <klocale.h>
#include <kfontdialog.h>
#include <kcolorbutton.h>
#include <kurlrequester.h>
#include <klineedit.h>
#include <knuminput.h>

#include <qframe.h>
#include <qlayout.h>
#include <qcheckbox.h>

using namespace KFTPGrabberBase;

KFTPConfigDialog::KFTPConfigDialog(QWidget *parent, const char *name)
  : KDialogBase(IconList, i18n("Preferences"), Ok|Apply|Cancel, Ok, parent, name, true, true)
{
  // "General" page
  QFrame *frame = addPage(i18n("General"), i18n("General preferences"), loadPanelPixmap("kfm"));
  QVBoxLayout *frameLayout = new QVBoxLayout(frame, 0, 0);
  m_configGeneral = new KFTPConfigGeneral(frame);
  m_configGeneral->defLocalDir->setMode(KFile::Directory | KFile::ExistingOnly | KFile::LocalOnly );
  frameLayout->addWidget(m_configGeneral);
  
  // "Actions" page
  frame = addPage(i18n("Actions"), i18n("Actions"), loadPanelPixmap("kfm"));
  frameLayout = new QVBoxLayout(frame, 0, 0);
  frameLayout->addWidget(FTPConfig().m_fileExistsDownActions.getConfigWidget(frame));
  frameLayout->addSpacing(KDialog::spacingHint());
  frameLayout->addWidget(FTPConfig().m_fileExistsUpActions.getConfigWidget(frame));
  frameLayout->addItem(new QSpacerItem(41, 300, QSizePolicy::Minimum, QSizePolicy::Expanding));
  
  // "Transfers" page
  frame = addPage(i18n("Transfers"), i18n("Transfer settings"), loadPanelPixmap("queue"));
  frameLayout = new QVBoxLayout(frame, 0, 0);
  m_configTrans = new KFTPConfigTransfers(frame);
  frameLayout->addWidget(m_configTrans);

  // "Log" page
  frame = addPage(i18n("Log"), i18n("Log"), loadPanelPixmap("log"));
  frameLayout = new QVBoxLayout(frame, 0, 0);
  m_configLog = new KFTPConfigLog(frame);
  frameLayout->addWidget(m_configLog);
  
  // "Display" page
  frame = addPage(i18n("Display"), i18n("Display"), loadPanelPixmap("background"));
  frameLayout = new QVBoxLayout(frame, 0, 0);
  m_configDisplay = new KFTPConfigDisplay(frame);
  frameLayout->addWidget(m_configDisplay);
  
  // Set all dialog elements according to the stored configuration
  updateDialog();
  
  connect(this, SIGNAL(applyClicked()), this, SLOT(slotUpdateConfig()));
  connect(this, SIGNAL(okClicked()), this, SLOT(slotUpdateConfig()));
}

void KFTPConfigDialog::slotUpdateConfig()
{
  // Save settings to the configuration

  // LOG
  FTPConfig().m_logFont = m_configLog->m_curFont;
  FTPConfig().m_logCommandsColor = m_configLog->commandColor->color();
  FTPConfig().m_logResponsesColor = m_configLog->responseColor->color();
  FTPConfig().m_logMultilineColor = m_configLog->multilineColor->color();
  FTPConfig().m_logErrorColor = m_configLog->errorColor->color();
  FTPConfig().m_logStatusColor = m_configLog->statusColor->color();
  FTPConfig().m_logSaveToFile = m_configLog->saveToFile->isChecked();
  FTPConfig().m_logOutputFilename = m_configLog->fileName->url();

  // TRANSFERS
  FTPConfig().m_portForceIp = m_configTrans->transForceIp->isChecked();
  FTPConfig().m_portIp = m_configTrans->transHostIp->text();
  FTPConfig().m_overwriteFiles = m_configTrans->transOverwrite->isChecked();
  FTPConfig().m_queueOnDND = m_configTrans->transQueueDND->isChecked();
  FTPConfig().m_lowercaseDownload = m_configTrans->transLowercase->isChecked();
  FTPConfig().m_diskCheckSpace = m_configTrans->diskCheckFreeSpace->isChecked();
  FTPConfig().m_diskCheckInterval = m_configTrans->diskCheckInterval->value();
  FTPConfig().m_diskMinFreeSpace = m_configTrans->diskFreeSpace->value();
  FTPConfig().m_globalMail = m_configTrans->globalMail->isChecked();
  FTPConfig().m_anonMail = m_configTrans->anonMail->text();
  
  // GENERAL
  FTPConfig().m_defLocalDir = m_configGeneral->defLocalDir->url();
  FTPConfig().m_retryTime = m_configGeneral->defRetryDelay->value();
  FTPConfig().m_retryCount = m_configGeneral->defNumRetries->value();
  FTPConfig().m_encryptBookmarks = m_configGeneral->encryptBookmarks->isChecked();
  FTPConfig().m_showBalloons = m_configGeneral->showBalloons->isChecked();
  FTPConfig().m_showBalloonWhenQueueEmpty = m_configGeneral->showBalloonsWhenEmpty->isChecked();
  FTPConfig().m_showRetrySuccessBalloon = m_configGeneral->showRetryBalloon->isChecked();
  FTPConfig().m_confirmExit = m_configGeneral->confirmExit->isChecked();
  FTPConfig().m_startMinimized = m_configGeneral->startMinimized->isChecked();
  FTPConfig().m_showSplash = m_configGeneral->showSplash->isChecked();
  
  // ACTIONS
  FTPConfig().m_fileExistsDownActions.updateConfig();
  FTPConfig().m_fileExistsUpActions.updateConfig();
  
  // DISPLAY
  FTPConfig().m_showHiddenFiles = m_configDisplay->showHidden->isChecked();
  FTPConfig().m_showTree = m_configDisplay->showTree->isChecked();
  FTPConfig().m_showSizeInBytes = m_configDisplay->showInBytes->isChecked();
  
  FTPConfig().emitChange();
}

void KFTPConfigDialog::updateDialog()
{
  // LOG
  m_configLog->m_curFont = FTPConfig().m_logFont;
  m_configLog->fontButton->setText(m_configLog->m_curFont.family());
  m_configLog->commandColor->setColor(FTPConfig().m_logCommandsColor);
  m_configLog->responseColor->setColor(FTPConfig().m_logResponsesColor);
  m_configLog->multilineColor->setColor(FTPConfig().m_logMultilineColor);
  m_configLog->errorColor->setColor(FTPConfig().m_logErrorColor);
  m_configLog->statusColor->setColor(FTPConfig().m_logStatusColor);
  m_configLog->saveToFile->setChecked(FTPConfig().m_logSaveToFile);
  m_configLog->fileName->setURL(FTPConfig().m_logOutputFilename);
  m_configLog->slotSaveToFileClicked();

  // TRANSFERS
  m_configTrans->transForceIp->setChecked(FTPConfig().m_portForceIp);
  m_configTrans->transHostIp->setText(FTPConfig().m_portIp);
  m_configTrans->transOverwrite->setChecked(FTPConfig().m_overwriteFiles);
  m_configTrans->transQueueDND->setChecked(FTPConfig().m_queueOnDND);
  m_configTrans->transLowercase->setChecked(FTPConfig().m_lowercaseDownload);
  m_configTrans->diskCheckFreeSpace->setChecked(FTPConfig().m_diskCheckSpace);
  m_configTrans->diskCheckInterval->setValue(FTPConfig().m_diskCheckInterval);
  m_configTrans->diskFreeSpace->setValue(FTPConfig().m_diskMinFreeSpace);
  m_configTrans->globalMail->setChecked(FTPConfig().m_globalMail);
  m_configTrans->anonMail->setText(FTPConfig().m_anonMail);
  m_configTrans->slotForceIpClicked();
  m_configTrans->slotFreeSpaceClicked();
  m_configTrans->slotGlobalMailClicked();
  
  // GENERAL
  m_configGeneral->defLocalDir->setURL(FTPConfig().m_defLocalDir);
  m_configGeneral->defRetryDelay->setValue(FTPConfig().m_retryTime);
  m_configGeneral->defNumRetries->setValue(FTPConfig().m_retryCount);
  m_configGeneral->encryptBookmarks->setChecked(FTPConfig().m_encryptBookmarks);
  m_configGeneral->showBalloons->setChecked(FTPConfig().m_showBalloons);
  m_configGeneral->showBalloonsWhenEmpty->setChecked(FTPConfig().m_showBalloonWhenQueueEmpty);
  m_configGeneral->showRetryBalloon->setChecked(FTPConfig().m_showRetrySuccessBalloon);
  m_configGeneral->confirmExit->setChecked(FTPConfig().m_confirmExit);
  m_configGeneral->startMinimized->setChecked(FTPConfig().m_startMinimized);
  m_configGeneral->showSplash->setChecked(FTPConfig().m_showSplash);
  
  // ACTIONS
  FTPConfig().m_fileExistsDownActions.updateWidget();
  FTPConfig().m_fileExistsUpActions.updateWidget();
  
  // DISPLAY
  m_configDisplay->showHidden->setChecked(FTPConfig().m_showHiddenFiles);
  m_configDisplay->showTree->setChecked(FTPConfig().m_showTree);
  m_configDisplay->showInBytes->setChecked(FTPConfig().m_showSizeInBytes);
}


#include "kftpconfigdialog.moc"
