/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpfiledirtreeview.h"

#include "misc.h"

#include <qheader.h>

using namespace KFTPGrabberBase;

/*
 *                         KFTPFileDirTreeViewItem class
 */

KFTPFileDirTreeViewItem::KFTPFileDirTreeViewItem(KFTPFileDirTreeView *parent, const KURL &url)
  : QListViewItem(parent)
{
  // Set the url
  m_url = url;
  m_dirty = false;

  // Set the name
  setText(0, path2Name(url.path(-1)));
}

KFTPFileDirTreeViewItem::KFTPFileDirTreeViewItem(QListViewItem *parent, const KURL &url)
  : QListViewItem(parent)
{
  // Set the url
  m_url = url;
  m_dirty = false;

  // Set the name
  setText(0, path2Name(url.path(-1)));
}

int KFTPFileDirTreeViewItem::compare(QListViewItem *i, int col, bool) const
{
  // Hidden files must be on top
  if (m_url.fileName().at(0) == '.')
    return -1;

  return QListViewItem::compare(i, col, false);
}

/*
 *                          KFTPFileDirTreeView class
 */

KFTPFileDirTreeView::KFTPFileDirTreeView(QWidget *parent, const char *name)
 : KListView(parent, name)
{
  // Create the columns
  addColumn(i18n("Directory"));

  // General tree settings
  setMinimumWidth(150);
  setRootIsDecorated(false);
  header()->hide();

  // Reset the view
  resetView(KURL("/"));
  m_noItemOpen = false;

  // Connect signals
  connect(this, SIGNAL(clicked(QListViewItem*)), this, SLOT(slotClicked(QListViewItem*)));
  connect(this, SIGNAL(doubleClicked(QListViewItem*)), this, SLOT(slotDoubleClicked(QListViewItem*)));
}


KFTPFileDirTreeView::~KFTPFileDirTreeView()
{
}

void KFTPFileDirTreeView::resetView(const KURL url, const QString &rootText)
{
  // Clear the view
  clear();
  header()->resizeSection(0, 0);

  // Create the root item
  KFTPFileDirTreeViewItem *rootItem = new KFTPFileDirTreeViewItem(this, remoteUrl("/", url));
  rootItem->setText(0, rootText);
  rootItem->setPixmap(0, loadSmallPixmap("folder_red"));
  setOpen(rootItem, true);
}

int KFTPFileDirTreeView::openURL(const KURL url, QListViewItem *parent)
{
  // Root item should always be open
  setOpen(firstChild(), true);

  // Go trough all items in the list and try to find the url
  QListViewItem *item;
  if (!parent)
    item = firstChild();
  else
    item = parent;

  while (item) {
    // Check if the item is correct
    if (static_cast<KFTPFileDirTreeViewItem*>(item)->m_url.path(-1) == url.path(-1)) {
      // We have found it
      ensureItemVisible(item);
      setCurrentItem(item);
      setSelected(item, true);

      // Change viewport
      QRect r = itemRect(item);
      if (r.isValid()) {
        int x, y;
        viewportToContents(contentsX(), r.y(), x, y);
        setContentsPos(x, y);
      }

      return 1;
    }

    if (item->firstChild()) {
      // If item has children, go after them
      if (openURL(url, item->firstChild()) == 1)
        return 1;
    }

    item = item->nextSibling();
  }

  return 0;
}

QListViewItem *KFTPFileDirTreeView::findItem(QListViewItem *parent, const QString &name)
{
  QListViewItem *item = parent->firstChild();

  while (item) {
    if (item->text(0) == name)
      return item;

    item = item->nextSibling();
  }

  // If nothing is found, parent should be returned
  return parent;
}

void KFTPFileDirTreeView::createFolder(const KURL url, QPixmap icon)
{
  // Creates a new folder at url
  QListViewItem *top = firstChild();
  int numDirs = url.path(1).contains('/', false);

  // URL for items
  KURL itemURL = url;
  itemURL.setPath("/");

  for (int i = 1; i < numDirs; i++) {
    QString sect = url.path().section('/', i, i);
    itemURL.setPath(url.path().section('/', 0, i));

    if (findItem(top, sect) == top) {
      // Item not yet created
      top = new KFTPFileDirTreeViewItem(top, itemURL);
      top->setPixmap(0, icon);
    } else {
      // Item is already present
      top = findItem(top, sect);
    }
    
    // Mark it as dirty
    static_cast<KFTPFileDirTreeViewItem*>(top)->m_dirty = true;

    // Open it
    if (!m_noItemOpen)
      setOpen(top, true);
  }

  // Root item should always be open
  setOpen(firstChild(), true);
}

void KFTPFileDirTreeView::endUpdate(const KURL url)
{
  // Remove all items in the selected dir, not marked as dirty
  KFTPFileDirTreeViewItem *top = static_cast<KFTPFileDirTreeViewItem*>(firstChild());
  KFTPFileDirTreeViewItem *tmp = 0L;
  int numDirs = url.path(1).contains('/', false);
  
  if (url.path() == "/") {
    KFTPFileDirTreeViewItem *i = static_cast<KFTPFileDirTreeViewItem*>(top->firstChild());
    while (i) {
      tmp = static_cast<KFTPFileDirTreeViewItem*>(i->nextSibling());
      
      if (!i->m_dirty)
        delete i;
      else
        i->m_dirty = false;
      
      i = tmp;
    }
    
    return;
  }

  // URL for items
  KURL itemURL = url;
  itemURL.setPath("/");

  for (int i = 1; i < numDirs; i++) {
    QString sect = url.path().section('/', i, i);
    itemURL.setPath(url.path().section('/', 0, i));

    if (findItem(top, sect) == top) {
      // Item not yet created
      return;
    } else {
      // Item is already present
      top = static_cast<KFTPFileDirTreeViewItem*>(findItem(top, sect));

      // Check for URL match
      if (itemURL.path(-1) == url.path(-1)) {
        // URL match, delete the item's children
        KFTPFileDirTreeViewItem *i = static_cast<KFTPFileDirTreeViewItem*>(top->firstChild());
        while (i) {
          tmp = static_cast<KFTPFileDirTreeViewItem*>(i->nextSibling());
          
          if (!i->m_dirty)
            delete i;
          else
            i->m_dirty = false;
            
          i = tmp;
        }
        return;
      }
    }
  }
}

void KFTPFileDirTreeView::clearFolder(const KURL url)
{
  // Remove url's children
  QListViewItem *top = firstChild();
  int numDirs = url.path(1).contains('/', false);
  
  if (url.path() == "/") {
    QListViewItem *i = top->firstChild();
    while (i) {
      delete i;
      i = top ->firstChild();
    }
    
    return;
  }

  // URL for items
  KURL itemURL = url;
  itemURL.setPath("/");

  for (int i = 1; i < numDirs; i++) {
    QString sect = url.path().section('/', i, i);
    itemURL.setPath(url.path().section('/', 0, i));

    if (findItem(top, sect) == top) {
      // Item not yet created
      return;
    } else {
      // Item is already present
      top = findItem(top, sect);

      // Check for URL match
      if (itemURL.path(-1) == url.path(-1)) {
        // URL match, delete the item's children
        QListViewItem *i = top->firstChild();
        while (i) {
          delete i;
          i = top->firstChild();
        }
        return;
      }
    }
  }
}

void KFTPFileDirTreeView::removeFolder(const KURL url)
{
  // Removes a folder at url
  QListViewItem *top = firstChild();
  int numDirs = url.path(1).contains('/', false);

  // URL for items
  KURL itemURL = url;
  itemURL.setPath("/");

  for (int i = 1; i < numDirs; i++) {
    QString sect = url.path().section('/', i, i);
    itemURL.setPath(url.path().section('/', 0, i));

    if (findItem(top, sect) == top) {
      // Item not yet created
      return;
    } else {
      // Item is already present
      top = findItem(top, sect);

      // Check for URL match
      if (itemURL.path(-1) == url.path(-1)) {
        // URL match, delete the item
        delete top;
        return;
      }
    }
  }
}

void KFTPFileDirTreeView::slotClicked(QListViewItem *item)
{
  if (!item)
    return;
    
  m_noItemOpen = true;
  emit pathChanged(static_cast<KFTPFileDirTreeViewItem*>(item)->m_url);
  m_noItemOpen = false;
}

void KFTPFileDirTreeView::slotDoubleClicked(QListViewItem *item)
{
  if (!item)
    return;
    
  setOpen(item, !item->isOpen());
}

#include "kftpfiledirtreeview.moc"
