/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpfiledirviewactions.h"
#include "kftpfiledirview.h"
#include "kftpfiledirlistview.h"
#include "kftpquickconnectdialog.h"
#include "kftpbookmarks.h"
#include "kftpqueue.h"
#include "misc.h"
#include "kftpapi.h"
#include "kftpfilepropsplugin.h"
#include "kftpsession.h"

#include <kapplication.h>
#include <kmainwindow.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kinputdialog.h>
#include <kpropertiesdialog.h>
#include <kio/job.h>
#include <kshred.h>

#include <qdir.h>
#include <qclipboard.h>

using namespace KFTPGrabberBase;

KFTPFileDirViewActions::KFTPFileDirViewActions(QObject *parent, const char *name)
 : QObject(parent, name)
{
  m_dirView = (KFTPFileDirView*)parent;
}

KActionCollection *KFTPFileDirViewActions::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

void KFTPFileDirViewActions::initActions()
{
  // Create all the actions;
  m_goUpAction = KStdAction::up(this, SLOT(slotGoUp()), actionCollection(), genID());
  m_goBackAction = KStdAction::back(this, SLOT(slotGoBack()), actionCollection(), genID());
  m_goForwardAction = KStdAction::forward(this, SLOT(slotGoForward()), actionCollection(), genID());
  m_goHomeAction = KStdAction::home(this, SLOT(slotGoHome()), actionCollection(), genID());

  m_reloadAction = KStdAction::redisplay(this, SLOT(slotReload()), actionCollection(), genID());
  m_reloadAction->setText(i18n("&Reload"));
  
  m_abortAction = new KAction(i18n("&Abort"), "stop", KShortcut(), this, SLOT(slotAbort()), actionCollection());
  m_toggleTreeViewAction = new KToggleAction(i18n("&Show tree view"), "view_tree", KShortcut(), this, SLOT(slotShowHideTree()), actionCollection());
  
  m_transferAction = new KAction(i18n("&Transfer"), KShortcut(), this, SLOT(slotTransfer()), actionCollection());
  m_queueTransferAction = new KAction(i18n("&Queue transfer"), "queue", KShortcut(), this, SLOT(slotQueueTransfer()), actionCollection());
  m_createDirAction = new KAction(i18n("&Create directory"), "folder_new", KShortcut(), this, SLOT(slotCreateDir()), actionCollection());
  m_renameAction = new KAction(i18n("&Rename"), KShortcut(), this, SLOT(slotRename()), actionCollection());
  m_deleteAction = new KAction(i18n("&Delete"), KShortcut(), this, SLOT(slotDelete()), actionCollection());
  m_propsAction = new KAction(i18n("&Properties"), KShortcut(), this, SLOT(slotProps()), actionCollection());
  m_setFilterAction = new KAction(i18n("&Set filter..."), "filter", KShortcut(), this, SLOT(slotSetFilter()), actionCollection());
  m_clearFilterAction = new KAction(i18n("Clear filter"), KShortcut(), this, SLOT(slotClearFilter()), actionCollection());
  m_shredAction = new KAction(i18n("&Shred"), "editshred", KShortcut(), this, SLOT(slotShred()), actionCollection());
  
  m_copyAction = KStdAction::copy(this, SLOT(slotCopy()), actionCollection(), genID());
  m_pasteAction = KStdAction::paste(this, SLOT(slotPaste()), actionCollection(), genID());
  
  m_moreActions = new KActionMenu(i18n("&More actions"), this);
  m_rawCmdAction = new KAction(i18n("Send &raw command..."), "openterm", KShortcut(), this, SLOT(slotRawCmd()), actionCollection());
  
  m_moreActions->insert(m_rawCmdAction);

  m_siteChangeAction = new KActionMenu(i18n("&Change site"), "goto", this);
  m_quickConnectAction = new KAction(i18n("&Quick connect..."), "connect_creating", KShortcut(), this, SLOT(slotQuickConnect()), actionCollection());
  m_connectAction = new KActionMenu(i18n("&Connect to"), this);
  m_disconnectAction = new KAction(i18n("&Disconnect"), "connect_no", KShortcut(), this, SLOT(slotDisconnect()), actionCollection());
  m_localAction = new KAction(i18n("View &local"), "gohome", KShortcut(), this, SLOT(slotLocal()), actionCollection());
  m_offlineAction = new KToggleAction(i18n("&Offline mode"), KShortcut(), this, SLOT(slotOffline()), actionCollection());

  m_siteChangeAction->insert(m_quickConnectAction);
  m_siteChangeAction->insert(m_connectAction);
  m_siteChangeAction->insert(m_disconnectAction);
  m_siteChangeAction->popupMenu()->insertSeparator();
  m_siteChangeAction->insert(m_localAction);
  m_siteChangeAction->insert(m_offlineAction);
  m_siteChangeAction->setStickyMenu(true);
  m_siteChangeAction->setDelayed(false);

  // Populate bookmarks
  FTPBookmarks->populateBookmarksMenu(m_connectAction, FTPBookmarks->getXMLDocument().documentElement(), false, m_dirView);
}

void KFTPFileDirViewActions::updateActions()
{
  // Enable/disable actions
  m_goUpAction->setEnabled(m_dirView->m_list->canMoveTo(KFTPFileDirListView::Up));
  m_goBackAction->setEnabled(m_dirView->m_list->canMoveTo(KFTPFileDirListView::Back));
  m_goForwardAction->setEnabled(m_dirView->m_list->canMoveTo(KFTPFileDirListView::Forward));
  m_goHomeAction->setEnabled(m_dirView->m_list->canMoveTo(KFTPFileDirListView::Home));

  m_reloadAction->setEnabled(m_dirView->m_list->canMoveTo(KFTPFileDirListView::Reload));
  m_abortAction->setEnabled(m_dirView->m_ftpClient->getClient()->isBusy());
  m_toggleTreeViewAction->setEnabled(true);

  m_localAction->setEnabled(!m_dirView->m_list->m_curURL.isLocalFile());
  m_quickConnectAction->setEnabled(m_dirView->m_list->m_curURL.isLocalFile() && !m_offlineAction->isChecked());
  m_connectAction->setEnabled(m_dirView->m_list->m_curURL.isLocalFile() || m_offlineAction->isChecked());
  m_disconnectAction->setEnabled(m_dirView->m_ftpClient->getClient()->isConnected() && !m_offlineAction->isChecked());
  m_offlineAction->setEnabled(true);

  // We can queue a transfer if we canTransfer() or we are in offline mode and have a remote URL
  // open or "the other" is in offline mode and has a remote URL open
  if (m_dirView->m_list->canTransfer() || (m_dirView->m_list->m_offlineMode && !m_dirView->m_list->m_curURL.isLocalFile()) ||
      (m_dirView->m_list->m_companion && (m_dirView->m_list->m_companion->m_offlineMode &&
      !m_dirView->m_list->m_companion->m_curURL.isLocalFile())))
  {
    m_queueTransferAction->setEnabled(true);
  } else {
    m_queueTransferAction->setEnabled(false);
  }

  m_transferAction->setEnabled(m_dirView->m_list->canTransfer());
  m_createDirAction->setEnabled(m_dirView->m_list->canLocalActions());
  m_renameAction->setEnabled(m_dirView->m_list->canLocalActions());
  m_deleteAction->setEnabled(m_dirView->m_list->canLocalActions());
  m_propsAction->setEnabled(!m_dirView->m_list->m_offlineMode);
  m_shredAction->setEnabled(m_dirView->m_list->m_curURL.isLocalFile());

  m_rawCmdAction->setEnabled(!m_dirView->m_list->m_curURL.isLocalFile() && 
                             m_dirView->m_ftpClient->getClient()->getFeatures() & SF_RAW_COMMAND);
  
  m_setFilterAction->setEnabled(true);
  m_clearFilterAction->setEnabled(!m_dirView->m_list->m_filter.isNull());

  // Repopulate bookmarks menu on updates
  m_connectAction->popupMenu()->clear();
  FTPBookmarks->populateBookmarksMenu(m_connectAction, FTPBookmarks->getXMLDocument().documentElement(), false, m_dirView);
}

void KFTPFileDirViewActions::slotQuickConnect()
{
  // Create/get the new dialog
  KFTPQuickConnectDialog* qcDialog = new KFTPQuickConnectDialog(m_dirView, "qc dialog");

  if (qcDialog->exec() == KDialogBase::Accepted) {
    // Get the url and connect
    if (m_dirView->m_ftpClient->getClient()->isConnected()) {
      if (KMessageBox::warningYesNo(0, i18n("Do you want to drop current connection ?")) == KMessageBox::No)
        return;

      m_dirView->m_ftpClient->KFTPClientThr::disconnect();
    }

    m_dirView->m_list->setHomeURL(KURL());
    FTPBookmarks->setFTPOptions(qcDialog->getServerURL(), m_dirView->m_ftpClient);
    m_dirView->m_ftpClient->connectToHost(qcDialog->getServerURL());
  }

  delete qcDialog;
}

void KFTPFileDirViewActions::slotDisconnect()
{
  if (m_dirView->m_ftpClient->getClient()->isConnected()) {
    if (KMessageBox::warningYesNo(0, i18n("Do you want to drop current connection ?")) == KMessageBox::No)
      return;

    m_dirView->m_ftpClient->KFTPClientThr::disconnect();
  }
}

void KFTPFileDirViewActions::slotLocal()
{
  if (m_dirView->m_ftpClient->getClient()->isConnected()) {
    if (KMessageBox::warningYesNo(0, i18n("Do you want to drop current connection ?")) == KMessageBox::No)
      return;

    m_dirView->m_ftpClient->KFTPClientThr::disconnect();
  }

  m_dirView->m_list->openURL(KURL(QDir::homeDirPath()));

  if (m_offlineAction->isChecked()) {
    m_offlineAction->setChecked(false);
    m_dirView->m_list->m_offlineMode = false;
    updateActions();
    
    // Trick the session into thinking that the "remote client" has disconnected
    FTPSessionManager->find(m_dirView)->slotClientDisconnected();
  }
}

void KFTPFileDirViewActions::slotOffline()
{
  if (m_dirView->m_ftpClient->getClient()->isConnected()) {
    if (KMessageBox::warningYesNo(0, i18n("Do you want to drop current connection ?")) == KMessageBox::No) {
      m_offlineAction->setChecked(false);
      m_dirView->m_list->m_offlineMode = false;
      return;
    }

    m_dirView->m_ftpClient->KFTPClientThr::disconnect();
    m_dirView->m_list->openURL(KURL(QDir::homeDirPath()));
  }

  m_dirView->m_list->m_offlineMode = m_offlineAction->isChecked();
  updateActions();

  if (!m_offlineAction->isChecked()) {
    // Go back to local dir
    m_dirView->m_list->openURL(KURL(QDir::homeDirPath()));
    
    // Trick the session into thinking that the "remote client" has disconnected
    FTPSessionManager->find(m_dirView)->slotClientDisconnected();
  }
}

void KFTPFileDirViewActions::slotGoUp()
{
  m_dirView->m_list->moveTo(KFTPFileDirListView::Up);
}

void KFTPFileDirViewActions::slotGoBack()
{
  m_dirView->m_list->moveTo(KFTPFileDirListView::Back);
}

void KFTPFileDirViewActions::slotGoForward()
{
  m_dirView->m_list->moveTo(KFTPFileDirListView::Forward);
}

void KFTPFileDirViewActions::slotReload()
{
  m_dirView->m_list->moveTo(KFTPFileDirListView::Reload);
}

void KFTPFileDirViewActions::slotGoHome()
{
  m_dirView->m_list->moveTo(KFTPFileDirListView::Home);
}

void KFTPFileDirViewActions::slotTransfer()
{
    // Queue a transfer
  ListItems selection = m_dirView->m_list->getCurrentSelectionLI();

  // Determine transfer type
  bool c_local = m_dirView->m_list->m_companion->m_curURL.isLocalFile();
  bool w_local = m_dirView->m_list->m_curURL.isLocalFile();
  KFTPQueue::TransferType transType = KFTPQueue::Download;

  if (w_local && !c_local) /* Local --> Remote (UPLOAD) */
    transType = KFTPQueue::Upload;

  if (!w_local && c_local) /* Remote --> Local (DOWNLOAD) */
    transType = KFTPQueue::Download;

  if (!w_local && !c_local) /* Remote --> Remote (FXP TRANSFER) */
    transType = KFTPQueue::FXP;

  KFTPQueue::Transfer *m_transfer = 0L;

  for (unsigned int i = 0; i < selection.count(); i++) {
    KFTPFileDirListViewItem *item = selection.at(i);
    KURL destURL = m_dirView->m_list->m_companion->m_curURL;
    destURL.addPath(item->m_dirObject.name());

    m_transfer = FTPQueueManager->spawnTransfer(
      item->m_url,
      destURL,
      item->m_dirObject.size(),
      item->m_directory,
      transType,
      selection.count() == 1,
      true,
      0L,
      true
    );
  }

  // Execute transfer
  if (m_transfer)
    m_transfer->delayedExecute();
}

void KFTPFileDirViewActions::slotQueueTransfer()
{
  // Queue a transfer
  ListItems selection = m_dirView->m_list->getCurrentSelectionLI();

  // Determine transfer type
  bool c_local = m_dirView->m_list->m_companion->m_curURL.isLocalFile();
  bool w_local = m_dirView->m_list->m_curURL.isLocalFile();
  KFTPQueue::TransferType transType = KFTPQueue::Download;

  if (w_local && !c_local) /* Local --> Remote (UPLOAD) */
    transType = KFTPQueue::Upload;

  if (!w_local && c_local) /* Remote --> Local (DOWNLOAD) */
    transType = KFTPQueue::Download;

  if (!w_local && !c_local) /* Remote --> Remote (FXP TRANSFER) */
    transType = KFTPQueue::FXP;

  for (unsigned int i = 0; i < selection.count(); i++) {
    KFTPFileDirListViewItem *item = selection.at(i);
    KURL destURL = m_dirView->m_list->m_companion->m_curURL;
    destURL.addPath(item->m_dirObject.name());

    FTPQueueManager->spawnTransfer(
      item->m_url,
      destURL,
      item->m_dirObject.size(),
      item->m_directory,
      transType,
      selection.count() == 1
    );
  }
}

void KFTPFileDirViewActions::slotCreateDir()
{
  // Create new directory
  bool ok;
  QString newDirName = KInputDialog::getText(i18n("Create directory"), i18n("Directory name"), "", &ok);
  
  if (ok) {
    if (m_dirView->m_list->m_curURL.isLocalFile()) {
      KIO::mkdir(KURL(m_dirView->m_list->m_curURL.path(1) + newDirName));
    } else {
      KURL newDir = m_dirView->m_list->m_curURL;
      newDir.addPath(newDirName);
      m_dirView->m_ftpClient->mkdir(newDir);
    }
  }
}

void KFTPFileDirViewActions::slotShred()
{
  // Shred the file
  if (KMessageBox::questionYesNo(0, i18n("Are you sure you want to SHRED this file?"), i18n("Shred file")) == KMessageBox::No)
    return;
  
  KURL::List selection = m_dirView->m_list->getCurrentSelection();
  KShred::shred((*selection.at(0)).path());
  
  slotReload();
}

void KFTPFileDirViewActions::slotRename()
{
  // Rename the first file in the current selection
  m_dirView->m_list->rename(m_dirView->m_list->selectedItems().at(0), 0);
}

void KFTPFileDirViewActions::slotDelete()
{
  // Delete a file or directory
  KURL::List selection = m_dirView->m_list->getCurrentSelection();

  if (KMessageBox::questionYesNo(0, i18n("Are you sure you want to delete this file(s)?"), i18n("Delete file")) == KMessageBox::No)
    return;

  // Go trough all files and delete them
  if (selection[0].isLocalFile()) {
    KIO::del(selection);
  } else {
    for (KURL::List::Iterator i = selection.begin(); i != selection.end(); ++i) {
      if (!(*i).isLocalFile()) {
        m_dirView->m_ftpClient->remove(KURL((*i).url()));
      }
    }
  }
}

void KFTPFileDirViewActions::slotCopy()
{
  QClipboard *cb = QApplication::clipboard();
  
  cb->setData(m_dirView->m_list->dragObject(), QClipboard::Clipboard);
}

void KFTPFileDirViewActions::slotPaste()
{
  // Decode the data and try to init transfer
  KIO::MetaData p_meta;
  KURL::List p_urls;
  
  if (KURLDrag::decode(QApplication::clipboard()->data(), p_urls, p_meta)) {
    // Add destination url and call the QueueManager
    p_meta.insert("DestURL", m_dirView->m_list->m_curURL.url());
    KURLDrag *drag = new KURLDrag(p_urls, p_meta, m_dirView->m_list, name());
    FTPQueueManager->insertTransfer(drag);
  }
}

void KFTPFileDirViewActions::slotProps()
{
  // Show file properties
  KURL::List selection = m_dirView->m_list->getCurrentSelection();
  KURL itemURL = *selection.at(0);

  if (selection.count() == 0) {
    itemURL = m_dirView->m_list->m_curURL;
    
    if (!itemURL.isLocalFile())
      return;
  }

  // Show the dialog
  if (itemURL.isLocalFile()) {
    KPropertiesDialog *propsDialog = new KPropertiesDialog(new KFileItem(itemURL, 0, 0));
    propsDialog->exec();
  } else {
    ListItems selItems = m_dirView->m_list->getCurrentSelectionLI();
    KFTPSession *session = FTPSessionManager->find(m_dirView);
    FTPDirectoryItem i = selItems.at(0)->m_dirObject;
    
    KPropertiesDialog *propsDialog = new KPropertiesDialog(itemURL.fileName());
    propsDialog->insertPlugin(new KFTPFilePropsPlugin(propsDialog, i, itemURL));
    propsDialog->insertPlugin(new KFTPFilePermissionsPropsPlugin(propsDialog, i, session, itemURL));
    propsDialog->exec();
  }
}

void KFTPFileDirViewActions::slotSetFilter()
{
  bool ok;
  QString filterStr = KInputDialog::getText(i18n("Filter"), i18n("Filter"), m_dirView->m_list->m_filter, &ok);

  if (ok) {
    m_dirView->m_list->setFilter(filterStr);
  }
}

void KFTPFileDirViewActions::slotClearFilter()
{
  m_dirView->m_list->setFilter(QString::null);
}

void KFTPFileDirViewActions::slotAbort()
{
  m_dirView->m_ftpClient->getClient()->abort();
}

void KFTPFileDirViewActions::slotRawCmd()
{
  bool ok;
  QString rawCmd = KInputDialog::getText(i18n("Send raw command"), i18n("Command"), "", &ok);

  if (ok) {
    m_dirView->m_ftpClient->rawCommand(rawCmd);
  }
}

void KFTPFileDirViewActions::slotShowHideTree()
{
  m_dirView->setTreeVisible(m_toggleTreeViewAction->isChecked());
}

#include "kftpfiledirviewactions.moc"
