/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpconfiguration.h"
#include "kftpfiltereditor.h"
#include "kftpaddpatterndialog.h"
#include "misc.h"

#include <klistview.h>
#include <klocale.h>
#include <kpushbutton.h>
#include <kmessagebox.h>
#include <klineeditdlg.h>

#include <qcheckbox.h>

using namespace KFTPGrabberBase;

KFTPFilterEditor::KFTPFilterEditor(QWidget *parent, const char *name)
 : KDialogBase(parent, name, true, i18n("Filter editor"), Ok|Cancel, Ok)
{
  // Create the main widget
  m_editorLayout = new KFTPFilterEditorLayout(this);

  // Set the dialog options
  setMainWidget(m_editorLayout);
  setInitialSize(QSize(500, 200));

  // Set the listview options
  m_editorLayout->patternList->setAllColumnsShowFocus(true);
  m_editorLayout->patternList->setColumnWidth(0, 100);

  // Set the options
  m_editorLayout->enabledCheck->setChecked(FTPConfig().m_enableHighlight);

  m_editorLayout->skipEnable->setChecked(FTPConfig().m_enableSkipList);
  m_editorLayout->skipEmptyFiles->setChecked(FTPConfig().m_skipEmptyFiles);
  m_editorLayout->skipEmptyDirs->setChecked(FTPConfig().m_skipEmptyDirs);
  m_editorLayout->skipApplyToDirs->setChecked(FTPConfig().m_applySkipToDirs);
  m_editorLayout->skipNoQueue->setChecked(FTPConfig().m_skipNoQueue);

  highlightLoadPatterns();
  asciiLoadExtensions();
  skipLoadList();

  // Connect the slots
  connect(m_editorLayout->addPatternButton, SIGNAL(clicked()), this, SLOT(slotAddHighlight()));
  connect(m_editorLayout->removePatternButton, SIGNAL(clicked()), this, SLOT(slotRemoveHighlight()));
  connect(m_editorLayout->editPatternButton, SIGNAL(clicked()), this, SLOT(slotEditHighlight()));

  connect(m_editorLayout->addExtButton, SIGNAL(clicked()), this, SLOT(slotAddAscii()));
  connect(m_editorLayout->removeExtButton, SIGNAL(clicked()), this, SLOT(slotRemoveAscii()));

  connect(m_editorLayout->addSkipButton, SIGNAL(clicked()), this, SLOT(slotAddSkip()));
  connect(m_editorLayout->removeSkipButton, SIGNAL(clicked()), this, SLOT(slotRemoveSkip()));
}


KFTPFilterEditor::~KFTPFilterEditor()
{
}

void KFTPFilterEditor::slotAddSkip()
{
  if (!m_editorLayout->newSkipPat->text().stripWhiteSpace().isEmpty()) {
    new QListViewItem(m_editorLayout->skipList, m_editorLayout->newSkipPat->text().stripWhiteSpace());
    m_editorLayout->newSkipPat->clear();
  }
}

void KFTPFilterEditor::slotRemoveSkip()
{
  if (!m_editorLayout->skipList->selectedItem())
    return;

  delete m_editorLayout->skipList->selectedItem();
}

void KFTPFilterEditor::skipLoadList()
{
  // Load the skiplist
  QStringList p_skipList = FTPConfig().m_skipList;

  for (QStringList::iterator i = p_skipList.begin(); i != p_skipList.end(); ++i) {
    new QListViewItem(m_editorLayout->skipList, (*i));
  }
}

QStringList KFTPFilterEditor::skipToStringList()
{
  QStringList p_skipList;
  QListViewItem *i = m_editorLayout->skipList->firstChild();
  while (i) {
    p_skipList.append(i->text(0));

    i = i->itemBelow();
  }

  return p_skipList;
}

void KFTPFilterEditor::slotAddAscii()
{
  if (!m_editorLayout->newExtension->text().stripWhiteSpace().isEmpty()) {
    new QListViewItem(m_editorLayout->extensionList, m_editorLayout->newExtension->text().stripWhiteSpace());
    m_editorLayout->newExtension->clear();
  }
}

void KFTPFilterEditor::slotRemoveAscii()
{
  if (!m_editorLayout->extensionList->selectedItem())
    return;

  delete m_editorLayout->extensionList->selectedItem();
}

void KFTPFilterEditor::slotAddHighlight()
{
  KFTPAddPatternDialog *p_addPatternDialog = new KFTPAddPatternDialog(this, "add pattern dialog");
  if (p_addPatternDialog->exec() == Accepted) {
    // FIXME replace if name exists
    QString pat = p_addPatternDialog->m_addLayout->patName->text();
    QString col = p_addPatternDialog->m_addLayout->patColor->color().name();

    QListViewItem *p_item = new QListViewItem(m_editorLayout->patternList, pat, " " + col);
    p_item->setPixmap(1, createColorPixmap(col));
  }
}

void KFTPFilterEditor::slotRemoveHighlight()
{
  if (!m_editorLayout->patternList->selectedItem())
    return;

  if (KMessageBox::warningYesNo(0, i18n("Are you sure that you want to remove this pattern ?")) == KMessageBox::Yes)
  {
    delete m_editorLayout->patternList->selectedItem();
  }
}

void KFTPFilterEditor::slotEditHighlight()
{
  if (!m_editorLayout->patternList->selectedItem())
    return;

  KFTPAddPatternDialog *p_addPatternDialog = new KFTPAddPatternDialog(this, "add pattern dialog");

  QListViewItem *p_item = m_editorLayout->patternList->selectedItem();
  p_addPatternDialog->m_addLayout->patName->setText(p_item->text(0));
  p_addPatternDialog->m_addLayout->patColor->setColor(QColor(p_item->text(1).stripWhiteSpace()));

  if (p_addPatternDialog->exec() == Accepted) {
    p_item->setText(0, p_addPatternDialog->m_addLayout->patName->text());
    p_item->setText(1, " " + p_addPatternDialog->m_addLayout->patColor->color().name());
    p_item->setPixmap(1, createColorPixmap(p_addPatternDialog->m_addLayout->patColor->color().name()));
  }
}

void KFTPFilterEditor::asciiLoadExtensions()
{
  // Load the ascii extensions
  QStringList p_extensions = FTPConfig().m_asciiList;

  for (QStringList::iterator i = p_extensions.begin(); i != p_extensions.end(); ++i) {
    new QListViewItem(m_editorLayout->extensionList, (*i));
  }

}

QStringList KFTPFilterEditor::asciiToStringList()
{
  QStringList p_extensions;
  QListViewItem *i = m_editorLayout->extensionList->firstChild();
  while (i) {
    p_extensions.append(i->text(0));

    i = i->itemBelow();
  }

  return p_extensions;
}

void KFTPFilterEditor::highlightLoadPatterns()
{
  // Load the patterns
  QStringList p_patterns = FTPConfig().m_filterList;

  for (QStringList::Iterator i = p_patterns.begin(); i != p_patterns.end(); ++i) {
    QString pat = (*i).section(':', 0, 0);
    QString col = (*i).section(':', 1, 1);

    QListViewItem *p_item = new QListViewItem(m_editorLayout->patternList, pat, " " + col);
    p_item->setPixmap(1, createColorPixmap(col));
  }
}

QStringList KFTPFilterEditor::highlightToStringList()
{
  QStringList p_patterns;
  QListViewItem *i = m_editorLayout->patternList->firstChild();
  while (i) {
    p_patterns.append(i->text(0)+":"+i->text(1).stripWhiteSpace());

    i = i->itemBelow();
  }

  return p_patterns;
}

void KFTPFilterEditor::slotOk()
{
  // Ok clicked, save the settings
  FTPConfig().m_enableHighlight = m_editorLayout->enabledCheck->isChecked();
  FTPConfig().m_filterList = highlightToStringList();
  FTPConfig().loadToColorizer();

  FTPConfig().m_asciiList = asciiToStringList();

  FTPConfig().m_enableSkipList = m_editorLayout->skipEnable->isChecked();
  FTPConfig().m_skipEmptyFiles = m_editorLayout->skipEmptyFiles->isChecked();
  FTPConfig().m_skipEmptyDirs = m_editorLayout->skipEmptyDirs->isChecked();
  FTPConfig().m_applySkipToDirs = m_editorLayout->skipApplyToDirs->isChecked();
  FTPConfig().m_skipNoQueue = m_editorLayout->skipNoQueue->isChecked();
  FTPConfig().m_skipList = skipToStringList();

  // Close the dialog
  accept();
}

#include "kftpfiltereditor.moc"
