/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2004 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include <qlayout.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qgroupbox.h>
#include <qspinbox.h>
#include <qsplitter.h>
#include <qtextedit.h>
#include <qtabwidget.h>

#include <klocale.h>
#include <kiconloader.h>
#include <klineedit.h>
#include <kpassdlg.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kcombobox.h>
#include <kurlrequester.h>
#include <kguiitem.h>

#include "kftpapi.h"
#include "kftpbookmarks.h"
#include "kftpbookmarkeditor.h"
#include "kftpbookmarkeditorlistview.h"
#include "kftpbookmarkimportwizard.h"
#include "kftpbookmarkeditortls.h"
#include "kftpconfiguration.h"
#include "misc.h"

using namespace KFTPGrabberBase;

KActionCollection *KFTPBookmarkEditor::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

KFTPBookmarkEditor::KFTPBookmarkEditor(QWidget *parent, const char *name)
 : KDialogBase(parent, name, true, "FTP Bookmark editor", User1|Ok|Apply|Cancel, Ok, true, KGuiItem(i18n("Import..."))),
   m_activeItem(0L)
{
  // We operate on a copy of the bookmarks
  m_bookmarks = new KFTPBookmarks(*FTPBookmarks);
  
  QHBoxLayout *winLayout = new QHBoxLayout(this, 0, KDialog::spacingHint());

  QSplitter *splitter = new QSplitter(this);
  setMainWidget(splitter);
  winLayout->addWidget(splitter);
  
  // Create the list view for editing bookmarks
  m_tree = new KFTPBookmarkEditorListView(m_bookmarks, splitter);

  m_properties = new ftpSiteProperties(splitter);
  
  initDialog();

  connect(this, SIGNAL(user1Clicked()), this, SLOT(slotImportWizard()));
  
  connect(m_tree, SIGNAL(bookmarkClicked(QListViewItem*)), this, SLOT(slotTreeClicked(QListViewItem*)));
  connect(m_tree, SIGNAL(bookmarkNew(KFTPBookmarkEditorItem*, QDomElement)), this, SLOT(slotNewAction(KFTPBookmarkEditorItem*, QDomElement)));
  connect(m_tree, SIGNAL(bookmarkDuplicated(KFTPBookmarkEditorItem*, QDomNode)), this, SLOT(slotDuplicateAction(KFTPBookmarkEditorItem*, QDomNode)));
  connect(m_tree, SIGNAL(bookmarkDeleted()), this, SLOT(slotDeleteAction()));
  connect(m_tree, SIGNAL(bookmarkMoved()), this, SLOT(refresh()));
  connect(m_tree, SIGNAL(categoryRenamed()), this, SLOT(refresh()));

  // Get the new bookmark data
  m_tree->fillBookmarkData(m_bookmarks->getXMLDocument().documentElement());
}

KFTPBookmarkEditor::KFTPBookmarkEditor(KFTPBookmarkEditorItem *item, QWidget *parent, const char *name)
 : KDialogBase(parent, name, true, "FTP Bookmark editor", Ok|Apply|Cancel, Ok, true),
   m_activeItem(item),
   m_tree(0L)
{
  setCaption(m_activeItem->text(0));
  
  // We operate on a copy of the bookmarks
  m_bookmarks = new KFTPBookmarks(*FTPBookmarks);
  
  QHBoxLayout *winLayout = new QHBoxLayout(this, 0, KDialog::spacingHint());
  
  m_properties = new ftpSiteProperties(this);
  setMainWidget(m_properties);
  winLayout->addWidget(m_properties);
  
  initDialog();
  
  m_activeNode = m_activeItem->m_node;
  showActiveNode();
}

KFTPBookmarkEditor::~KFTPBookmarkEditor()
{
  delete m_bookmarks;
}

void KFTPBookmarkEditor::initDialog()
{
  m_properties->tabList->setTabIconSet(m_properties->general, loadSmallIcon("kfm"));
  m_properties->tabList->setTabIconSet(m_properties->advanced, loadSmallIcon("contents"));
  m_properties->protoAdvanced->setIconSet(loadSmallIcon("configure"));
  m_properties->lookGroup->setEnabled(false);
  m_properties->retryGroup->setEnabled(false);
  m_properties->defLocalDir->setMode(KFile::Directory | KFile::ExistingOnly | KFile::LocalOnly );
    
  connect(m_properties->anonLogin, SIGNAL(clicked()), this, SLOT(slotChangeAnonLogin()));
  connect(m_properties->useLook, SIGNAL(clicked()), this, SLOT(slotChangeActiveLookGroup()));
  connect(m_properties->doRetry, SIGNAL(clicked()), this, SLOT(slotChangeActiveRetryGroup()));
  connect(m_properties->protoAdvanced, SIGNAL(clicked()), this, SLOT(slotProtoAdvancedClicked()));
  connect(m_properties->serverProtocol, SIGNAL(activated(int)), this, SLOT(slotChangeServerProtocol(int)));
  
  connect(this, SIGNAL(applyClicked()), this, SLOT(slotSaveActiveNode()));
  connect(this, SIGNAL(okClicked()), this, SLOT(slotSaveActiveNode()));
  
  // Deactivate everything until an item gets selected
  m_properties->setEnabled(false);
}

void KFTPBookmarkEditor::refresh()
{
  m_tree->clear();
  m_tree->fillBookmarkData(m_bookmarks->getXMLDocument().documentElement());
}

void KFTPBookmarkEditor::slotImportWizard()
{
  KFTPBookmarkImportWizard *wizard = new KFTPBookmarkImportWizard(this);
  
  if (wizard->exec() == QDialog::Accepted) {
    // Refresh bookmarks
    m_bookmarks->setBookmarks(FTPBookmarks);
    
    m_tree->clear();
    m_tree->fillBookmarkData(m_bookmarks->getXMLDocument().documentElement());
  }
  
  delete wizard;
}

void KFTPBookmarkEditor::slotNewAction(KFTPBookmarkEditorItem *item, QDomElement newServer)
{
  // Display the new server's properties
  m_activeItem = item;
  m_activeNode = newServer;
  showActiveNode();
  
  // Set Focus to servername
  m_properties->serverName->selectAll();
  m_properties->serverName->setFocus();
}

void KFTPBookmarkEditor::slotDuplicateAction(KFTPBookmarkEditorItem *item, QDomNode copy)
{
  m_activeItem = item;
    
  // Display its properties
  m_activeNode = copy;
  showActiveNode();
}

void KFTPBookmarkEditor::slotDeleteAction()
{
  m_activeItem = 0L;
  m_properties->setEnabled(false);
}

void KFTPBookmarkEditor::slotSaveActiveNode()
{
  if ( !m_activeItem ) {
    FTPBookmarks->setBookmarks(m_bookmarks);
    return;
  }
  
  // Save all the data for the active node
  if (m_properties->serverName->text() != m_activeNode.toElement().attribute("name", "") &&
      m_bookmarks->domServerExistsByName(m_properties->serverName->text()))
  {
    // No duplicates!
    KMessageBox::sorry(this, i18n("This server name is already in use!\n\nChoose another ;-)"),
                       i18n("Name already in use"));
    return;
  }

  QDomElement activeElement = m_activeNode.toElement();
  activeElement.setAttribute("name", m_properties->serverName->text());

  QString port, retryTime, retryCount;
  port.setNum(m_properties->serverPort->value());
  retryTime.setNum(m_properties->retryDelay->value());
  retryCount.setNum(m_properties->retryCount->value());
  
  // Save the properties from the form
  m_bookmarks->domSetProperty(activeElement, "host", m_properties->serverAddress->text());
  m_bookmarks->domSetProperty(activeElement, "port", port);
  
  m_bookmarks->domSetProperty(activeElement, "protocol", m_properties->serverProtocol->currentItem() == SP_SFTP ? "sftp" : "ftp");
  m_bookmarks->domSetProperty(activeElement, "use_tls", m_properties->serverProtocol->currentItem() == SP_SSL_EXPLICIT ? "true" : "false" );
  m_bookmarks->domSetProperty(activeElement, "use_implicit", m_properties->serverProtocol->currentItem() == SP_SSL_IMPLICIT ? "true" : "false" );
  
  m_bookmarks->domSetOption(activeElement, "anonlogin", m_properties->anonLogin->isChecked());
  m_bookmarks->domSetProperty(activeElement, "downuser", m_properties->downUser->text());
  m_bookmarks->domSetProperty(activeElement, "downpass", encodePassword(m_properties->downPass->password()));
  
  // Only save the local dir if it differs from the default one
  if (m_properties->defLocalDir->url() != FTPConfig().m_defLocalDir)
    m_bookmarks->domSetProperty(activeElement, "deflocalpath", m_properties->defLocalDir->url());
    
  m_bookmarks->domSetProperty(activeElement, "defremotepath", m_properties->defRemoteDir->text());
  m_bookmarks->domSetProperty(activeElement, "description", m_properties->description->text());
  if (m_properties->useLook->isChecked()) {
    m_bookmarks->domSetProperty(activeElement, "lookuser", m_properties->lookUser->text());
    m_bookmarks->domSetProperty(activeElement, "lookpass", encodePassword(m_properties->lookPass->text()));
  }
  if (m_properties->doRetry->isChecked()) {
    m_bookmarks->domSetProperty(activeElement, "retrytime", retryTime);
    m_bookmarks->domSetProperty(activeElement, "retrycount", retryCount);
  }
  m_bookmarks->domSetOption(activeElement, "disablePASV", m_properties->disablePASV->isChecked());
  m_bookmarks->domSetOption(activeElement, "disableEPSV", m_properties->disableEPSV->isChecked());
  m_bookmarks->domSetProperty(activeElement, "dist_ftpd", m_properties->distributedFtp->isChecked() ? "true" : "false");

  // Update the ListView node
  m_activeItem->setText(0, m_properties->serverName->text());

  if (m_tree)
    FTPBookmarks->setBookmarks(m_bookmarks);
  
  showActiveNode();
  FTPBookmarks->domSaveFile();
}

void KFTPBookmarkEditor::showActiveNode()
{
  if ( m_activeItem->m_type == BT_CATEGORY ) {
    m_properties->setEnabled(false);
    return;
  }

  m_properties->setEnabled(true);
    
  // Show all the data for the active node
  QString serverName = m_activeNode.toElement().attribute("name");
  QString serverHost = m_bookmarks->domGetProperty(m_activeNode, "host");
  QString serverPort = m_bookmarks->domGetProperty(m_activeNode, "port");
  QString serverProtocol = m_bookmarks->domGetProperty(m_activeNode, "protocol");
  bool useTLS = m_bookmarks->domGetProperty(m_activeNode, "use_tls") == "true" ? true : false;
  bool useImplicit = m_bookmarks->domGetProperty(m_activeNode, "use_implicit") == "true" ? true : false;
  QString downUser = m_bookmarks->domGetProperty(m_activeNode, "downuser");
  QString downPass = m_bookmarks->domGetProperty(m_activeNode, "downpass");
  
  QString defLPath = m_bookmarks->domGetProperty(m_activeNode, "deflocalpath");
  if ( defLPath.isEmpty() )
    defLPath = FTPConfig().m_defLocalDir;
  
  QString defRPath = m_bookmarks->domGetProperty(m_activeNode, "defremotepath");
  QString description = m_bookmarks->domGetProperty(m_activeNode, "description");
  QString lookUser = m_bookmarks->domGetProperty(m_activeNode, "lookuser");
  QString lookPass = m_bookmarks->domGetProperty(m_activeNode, "lookpass");
  QString retryTime = m_bookmarks->domGetProperty(m_activeNode, "retrytime");
  QString retryCount = m_bookmarks->domGetProperty(m_activeNode, "retrycount");
  bool distributedFtpd = m_bookmarks->domGetProperty(m_activeNode, "dist_ftpd") == "true" ? true : false;
  
  if (serverPort.isEmpty())
    serverPort = "21";
    
  // Set the properties in the form
  m_properties->serverName->setText(serverName);
  m_properties->serverAddress->setText(serverHost);
  
  if (serverProtocol == "sftp") m_properties->serverProtocol->setCurrentItem(SP_SFTP);
  else if ( useTLS ) m_properties->serverProtocol->setCurrentItem(SP_SSL_EXPLICIT);
  else if ( useImplicit ) m_properties->serverProtocol->setCurrentItem(SP_SSL_IMPLICIT);
  else m_properties->serverProtocol->setCurrentItem(SP_FTP);
  slotChangeServerProtocol(m_properties->serverProtocol->currentItem());
  
  m_properties->serverPort->setValue(serverPort.toInt());
  m_properties->downUser->erase();
  m_properties->downPass->erase();
  m_properties->anonLogin->setChecked(m_bookmarks->domGetOption(m_activeNode, "anonlogin").toInt());
  slotChangeAnonLogin();
  
  m_properties->downUser->setText(downUser);
  m_properties->downPass->insert(downPass);
  m_properties->defLocalDir->setURL(defLPath);
  m_properties->defRemoteDir->setText(defRPath);
  m_properties->description->setText(description);

  if (!lookUser.isEmpty() && !lookPass.isEmpty()) {
    m_properties->useLook->setChecked(true);
    m_properties->lookGroup->setEnabled(true);
    m_properties->lookUser->setText(lookUser);
    m_properties->lookPass->erase();
    m_properties->lookPass->insert(lookPass);
  } else {
    m_properties->useLook->setChecked(false);
    slotChangeActiveLookGroup();
  }

  if ( !retryTime.isEmpty() || !retryCount.isEmpty() ) {
    m_properties->doRetry->setChecked(true);
    m_properties->retryGroup->setEnabled(true);
    m_properties->retryDelay->setValue(retryTime.toInt());
    m_properties->retryCount->setValue(retryCount.toInt());
  } else {
    m_properties->doRetry->setChecked(false);
    m_properties->retryDelay->setValue(FTPConfig().m_retryTime);
    m_properties->retryCount->setValue(FTPConfig().m_retryCount);
    slotChangeActiveRetryGroup();
  }

  m_properties->disablePASV->setChecked(m_bookmarks->domGetOption(m_activeNode, "disablePASV").toInt());
  m_properties->disableEPSV->setChecked(m_bookmarks->domGetOption(m_activeNode, "disableEPSV").toInt());
  m_properties->distributedFtp->setChecked(distributedFtpd);
}

void KFTPBookmarkEditor::slotTreeClicked(QListViewItem *item)
{
  m_activeItem = (KFTPBookmarkEditorItem*)item;
  
  if ( !m_activeItem || m_activeItem->m_type == BT_CATEGORY ) {
    // Deactivate all controls
     m_properties->setEnabled(false);
     return;
  }

  // Find the node and set it active
  if (m_activeItem->m_type == BT_CATEGORY) {
    // Category
    m_activeNode.clear();
  } else {
    // Server
    m_activeNode = m_activeItem->m_node;
  }
  
  // Show the active node (if any)
  showActiveNode();
}

void KFTPBookmarkEditor::slotChangeActiveLookGroup()
{
  if (m_properties->useLook->isChecked()) {
    m_properties->lookGroup->setEnabled(true);
  } else {
    m_properties->lookGroup->setEnabled(false);
    m_properties->lookUser->setText("");
    m_properties->lookPass->setText("");
  }
}

void KFTPBookmarkEditor::slotChangeActiveRetryGroup()
{
  if (m_properties->doRetry->isChecked()) {
    m_properties->retryGroup->setEnabled(true);
  } else {
    m_properties->retryGroup->setEnabled(false);
    m_properties->retryCount->setValue(10);
    m_properties->retryDelay->setValue(60);
  }
}

void KFTPBookmarkEditor::slotChangeAnonLogin()
{
  static QString tmpUser, tmpPass;
  
  if (m_properties->anonLogin->isChecked()) {
    m_properties->downUser->setEnabled(false);
    m_properties->downPass->setEnabled(false);
    tmpUser = m_properties->downUser->text();
    tmpPass = m_properties->downPass->text();
    m_properties->downUser->setText("anonymous");
    m_properties->downPass->setText(FTPConfig().m_anonMail);
  } else {
    m_properties->downUser->setEnabled(true);
    m_properties->downPass->setEnabled(true);
    m_properties->downUser->setText(tmpUser);
    m_properties->downPass->setText(tmpPass);
  }
}

void KFTPBookmarkEditor::slotProtoAdvancedClicked()
{
  KFTPBookmarkEditorTLS *tlsDialog = new KFTPBookmarkEditorTLS(this);

  // Initialize dialog
  QChar tlsMode = m_bookmarks->domGetProperty(m_activeNode, "tls_mode").at(0);
  bool useCert = m_bookmarks->domGetProperty(m_activeNode, "use_cert") == "true" ? true : false;
  QString certPath = m_bookmarks->domGetProperty(m_activeNode, "tls_cert_path");
  
  tlsDialog->setTLSMode(tlsMode);
  tlsDialog->setCertChecked(useCert);
  if (useCert)
    tlsDialog->setCertPath(certPath);
  tlsDialog->slotChangeActiveX509Group();
  
  if ( tlsDialog->exec() ) {
    // Save the modifications
    m_bookmarks->domSetProperty(m_activeNode.toElement(), "tls_mode", tlsDialog->getTLSMode());
    m_bookmarks->domSetProperty(m_activeNode.toElement(), "use_cert", tlsDialog->isCertChecked() ? "true" : "false");
    
    if (tlsDialog->isCertChecked())
      m_bookmarks->domSetProperty(m_activeNode.toElement(),"tls_cert_path", tlsDialog->getCertPath());
  }
  
  delete tlsDialog;
}

void KFTPBookmarkEditor::slotChangeServerProtocol(int item)
{
  // Enable/Disable the SSL/TLS settings if needed
  m_properties->protoAdvanced->setEnabled( item == 1 || item == 2 );
  
  // Set the default port
  switch (item) {
   case SP_SSL_IMPLICIT:
     m_properties->serverPort->setValue(993);
     break;
   case SP_SFTP:
     m_properties->serverPort->setValue(22);
     break;
   default:
     m_properties->serverPort->setValue(21);
     break;
   }
}
