/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2004 by the KFTPGrabber developers
 * Copyright (C) 2004 Jernej Kos <kostko@jweb-network.net>
 * Copyright (C) 2004 Markus Brueffer <markus@brueffer.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpapi.h"
#include "kftpbookmarkeditorlistview.h"
#include "kftpbookmarks.h"
#include "kftpsession.h"
#include "misc.h"

#include <qdragobject.h>

#include <kaction.h>
#include <kinputdialog.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kdebug.h>

using namespace KFTPGrabberBase;

KFTPBookmarkEditorItem::KFTPBookmarkEditorItem(QListView *parent, const QString &t1)
  : KListViewItem(parent, t1)
{
}

KFTPBookmarkEditorItem::KFTPBookmarkEditorItem(QListViewItem *parent, const QString &t1)
  : KListViewItem(parent, t1)
{
}

int KFTPBookmarkEditorItem::compare(QListViewItem *i, int col, bool) const
{
  if (m_type != ((KFTPBookmarkEditorItem*)i)->m_type) {
    if (m_type == 0)
      return -1; // category
    else
      return 1; // server
  }

  return KListViewItem::compare(i, col, false);
}

QString KFTPBookmarkEditorItem::getXMLPath()
{
  QDomNode n = m_node;
  QString path = "";
  
  while (n != ((KFTPBookmarkEditorListView*)listView())->getBookmarks()->getXMLDocument().documentElement()) {
    path = "/" + n.toElement().attribute("name") + path;
    
    n = n.parentNode();
  }
  
  return path;
}

KFTPBookmarkEditorListView::KFTPBookmarkEditorListView(KFTPBookmarks *bookmarks, QWidget *parent, const char *name)
 : KFTPListView(parent, name),
   m_autoUpdate(false),
   m_connectBookmark(false),
   m_editMenuItem(false),
   m_bookmarks(bookmarks),
   m_activeItem(0L)
{
  // Set some stuff
  setMinimumWidth(150);
  setFullWidth(true);
  addColumn(i18n("Bookmarks"));
  setRootIsDecorated(true);
  setEmptyListText(i18n("No bookmarks."));
  setItemsRenameable(false);
  
  // Drag & drop
  setDragEnabled(true);
  setAcceptDrops(true);
  setDropVisualizer(false);
  
  // Init auto open timer
  m_openTimer = new QTimer(this);
  
  connect(this, SIGNAL(dropped(QDropEvent*, QListViewItem*)), this, SLOT(slotDropped(QDropEvent*, QListViewItem*)));
  connect(this, SIGNAL(clicked(QListViewItem*)), this, SLOT(slotClicked(QListViewItem*)));
  connect(this, SIGNAL(doubleClicked(QListViewItem*)), this, SLOT(slotDoubleClicked(QListViewItem*)));
  connect(this, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint&, int)), this, SLOT(slotContextMenu(QListViewItem*, const QPoint&, int)));
  
  /* Init the actions */
  m_newAction = new KAction(i18n("&New"), "filenew", KShortcut(), this, SLOT(slotNewAction()), actionCollection(), "bookmark_new");
  m_renameAction = new KAction(i18n("&Rename"), KShortcut(), this, SLOT(slotRenameAction()), actionCollection(), "bookmark_rename");
  m_deleteAction = new KAction(i18n("&Delete"), "editdelete", KShortcut(), this, SLOT(slotDeleteAction()), actionCollection(), "bookmark_delete");
  m_subCatAction = new KAction(i18n("&Create subcategory"), "folder_new", KShortcut(), this, SLOT(slotSubCatAction()), actionCollection(), "bookmark_subcat");
  m_copyAction = new KAction(i18n("&Duplicate"), KShortcut(), this, SLOT(slotDuplicateAction()), actionCollection(), "bookmark_duplicate");
  
  // Set the starting category to document root
  m_activeCategory = m_bookmarks->getXMLDocument().documentElement();
}

KActionCollection *KFTPBookmarkEditorListView::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

void KFTPBookmarkEditorListView::setAutoUpdate(bool value)
{
  m_autoUpdate = value;
  
  if (value)
    connect(m_bookmarks, SIGNAL(update()), this, SLOT(slotAutoUpdate()));
    
  slotAutoUpdate();
}

void KFTPBookmarkEditorListView::setConnectBookmark(bool value)
{
  m_connectBookmark = value;
  
  if (value)
    connect(this, SIGNAL(executed(QListViewItem*)), this, SLOT(slotBookmarkExecuted(QListViewItem*)));
}

void KFTPBookmarkEditorListView::setEditMenuItem(bool value)
{ 
  m_editMenuItem = value;
}

void KFTPBookmarkEditorListView::slotBookmarkExecuted(QListViewItem *item)
{
  if (!item || !m_connectBookmark || static_cast<KFTPBookmarkEditorItem*>(item)->m_type == BT_CATEGORY)
    return;
    
  QDomNode selServer = static_cast<KFTPBookmarkEditorItem*>(item)->m_node;
  
  KURL serverURL;
  serverURL.setProtocol(m_bookmarks->domGetProperty(selServer, "protocol"));
  serverURL.setHost(m_bookmarks->domGetProperty(selServer, "host"));
  serverURL.setPort(m_bookmarks->domGetProperty(selServer, "port").toInt());
  if (!m_bookmarks->domGetProperty(selServer, "lookuser").isEmpty()) {
    // If there is look account set, use it
    serverURL.setUser(m_bookmarks->domGetProperty(selServer, "lookuser"));
    serverURL.setPass(m_bookmarks->domGetProperty(selServer, "lookpass"));
  } else {
    // Otherwise use download account
    serverURL.setUser(m_bookmarks->domGetProperty(selServer, "downuser"));
    serverURL.setPass(m_bookmarks->domGetProperty(selServer, "downpass"));
  }
  
  // Check something
  if (!serverURL.hasUser()) {
    KMessageBox::sorry(0, i18n("No username specified for '%1'!").arg(serverURL.host()), i18n("Error connecting"));
    return;
  }
  
  // Just spawn a new session
  KFTPSession *session = FTPSessionManager->spawnRemoteSession(IGNORE_SIDE, serverURL, selServer);
  m_bookmarks->setFTPOptions(selServer, session->getClient());
}

void KFTPBookmarkEditorListView::slotAutoUpdate()
{
  if (m_autoUpdate) {
    // Update the bookmark list
    fillBookmarkData();
  }
}

QDragObject *KFTPBookmarkEditorListView::dragObject()
{
  // Determine hotspot and pixmap
  QPoint hotspot;
  QPixmap pixmap = *selectedItem()->pixmap(0);
  hotspot.setX(pixmap.width() / 2);
  hotspot.setY(pixmap.height() / 2);
  
  QString path = static_cast<KFTPBookmarkEditorItem*>(selectedItem())->getXMLPath();
  m_dragObject = new QStoredDrag("application/x-qlistviewitem", this);
  QByteArray data;
  QDataStream arg(data, IO_WriteOnly);
  arg << path;

  // Start the drag
  static_cast<QStoredDrag*>(m_dragObject)->setEncodedData(data);
  m_dragObject->setPixmap(pixmap, hotspot);
  
  return m_dragObject;
}

bool KFTPBookmarkEditorListView::acceptDrag(QDropEvent *e) const
{
  // If it is a local drag, accept it
  return e->source() == static_cast<QWidget*>(const_cast<KFTPBookmarkEditorListView*>(this));
}

void KFTPBookmarkEditorListView::contentsDragEnterEvent(QDragEnterEvent *e)
{
  if (!acceptDrag(e)) {
    e->ignore();
    return;
  }
  e->acceptAction();
  
  m_currentBeforeDropItem = selectedItem();
  QListViewItem *item = itemAt(contentsToViewport(e->pos()));
  
  if (item) {
    m_dropItem = item;
    m_openTimer->start(750);
  } else {
    m_dropItem = 0L;
  }
}

void KFTPBookmarkEditorListView::contentsDragMoveEvent(QDragMoveEvent *e)
{
  if (!acceptDrag(e)) {
    e->ignore();
    return;
  }
  e->acceptAction();
  
  QListViewItem *item = itemAt(contentsToViewport(e->pos()));
  if (item) {
    setSelected(item, true);
    if (item != m_dropItem) {
      m_openTimer->stop();
      m_dropItem = item;
      m_openTimer->start(750);
    }
  } else {
    if (selectedItem()) {
      setSelected(selectedItem(), false);
    }
    
    m_openTimer->stop();
    m_dropItem = 0L;
  }
}

void KFTPBookmarkEditorListView::contentsDragLeaveEvent(QDragLeaveEvent*)
{
  if (m_currentBeforeDropItem) {
    setSelected(m_currentBeforeDropItem, true);
    ensureItemVisible(m_currentBeforeDropItem);
  } else {
    setSelected(m_dropItem, false);
  }
  
  m_openTimer->stop();
  m_dropItem = 0L;
}

void KFTPBookmarkEditorListView::slotOpenTimer()
{
  m_openTimer->stop();
  if (m_dropItem && m_dropItem->isExpandable()) {
    // Expand the category
    m_dropItem->setOpen(true);
  }
}

void KFTPBookmarkEditorListView::slotDropped(QDropEvent *e, QListViewItem*)
{
  QString newPath = "/";
  
  if (selectedItem()) {
    QListViewItem *newParent = 0L;
    
    // Get the new path
    if (static_cast<KFTPBookmarkEditorItem*>(selectedItem())->m_type == 0)
      newParent = selectedItem();
    else
      newParent = selectedItem()->parent();
    
    if (newParent)
      newPath = static_cast<KFTPBookmarkEditorItem*>(newParent)->getXMLPath();
  }
    
  QString oldPath;
  
  // Decode the data
  QDataStream arg(e->encodedData("application/x-qlistviewitem"), IO_ReadOnly);
  arg >> oldPath;
  
  // Resolve nodes
  QDomNode original = m_bookmarks->domFindNodeByXMLPath(oldPath);
  QDomNode dest = m_bookmarks->domFindNodeByXMLPath(newPath);
  
  // Move it
  dest.appendChild(original);
  emit bookmarkMoved();
  
  // Notify the bookmark manager
  m_bookmarks->doEmitUpdate();
}

void KFTPBookmarkEditorListView::fillBookmarkData(QDomNode parent, KFTPBookmarkEditorItem *item)
{
  // Fill the tree with data
  if (parent.isNull()) {
    // Clear the tree
    clear();
    
    // Get the root element from bookmarks
    parent = m_bookmarks->getXMLDocument().documentElement();
  }

  QDomNode n = parent.firstChild();
  
  while (!n.isNull()) {
    if (n.toElement().tagName() == "category") {
      // Add a submenu
      KFTPBookmarkEditorItem *cat;
      if (!item)
        cat = new KFTPBookmarkEditorItem(this, n.toElement().attribute("name", "No name"));
      else
        cat = new KFTPBookmarkEditorItem(item, n.toElement().attribute("name", "No name"));

      cat->m_type = 0;
      cat->m_node = n;

      cat->setPixmap(0, loadSmallPixmap("bookmark_folder"));
      fillBookmarkData(n, cat);
    }
    if (n.toElement().tagName() == "server") {
      KFTPBookmarkEditorItem *serv;
      if (!item)
        serv = new KFTPBookmarkEditorItem(this, n.toElement().attribute("name", "No name"));
      else
        serv = new KFTPBookmarkEditorItem(item, n.toElement().attribute("name", "No name"));

      serv->m_type = 1;
      serv->m_node = n;

      serv->setPixmap(0, loadSmallPixmap("ftp"));
    }
    n = n.nextSibling();
  }
}

void KFTPBookmarkEditorListView::slotClicked(QListViewItem *item)
{
  m_activeItem = (KFTPBookmarkEditorItem*)item;
  
  if ( m_activeItem )
  {
    if (m_activeItem->m_type == BT_CATEGORY) {
      // Category
      m_activeCategory = m_activeItem->m_node;
      m_activeNode.clear();
    } else {
      // Server
      m_activeNode = m_activeItem->m_node;
      m_activeCategory = m_activeNode.parentNode();
    }
  }

  emit bookmarkClicked(item);
}

void KFTPBookmarkEditorListView::slotDoubleClicked(QListViewItem *item)
{
  if (!item)
    return;
    
  setOpen(item, !item->isOpen());
}

KFTPBookmarks *KFTPBookmarkEditorListView::getBookmarks()
{
  return m_bookmarks;
}

void KFTPBookmarkEditorListView::slotRenameAction()
{
  // Rename a subcategory
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  QDomNode cat = item->m_node;
  
  bool ok;
  QString newName = KInputDialog::getText(i18n("Category name"), i18n("Rename category"), item->text(0), &ok, this);
  
  if (ok) {
    // Ok, let's rename it
    m_bookmarks->domRenameCategory(item->m_node, newName);    
    item->setText(0, newName);
  }
}

void KFTPBookmarkEditorListView::slotSubCatAction()
{
  // Create a subcategory
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  QDomNode cat = item ? item->m_node : m_bookmarks->getXMLDocument().documentElement();
  
  bool ok;
  QString newName = KInputDialog::getText(i18n("New category name"), i18n("New category"), "", &ok, this);
  
  if (ok) {
    // Let's create the sub category
    m_bookmarks->domCreateCategory(cat, newName);
    emit categoryRenamed();
  }
}

void KFTPBookmarkEditorListView::slotNewAction()
{
  // Create the new node
  QDomElement newServer = m_bookmarks->getXMLDocument().createElement("server");
  newServer.toElement().setAttribute("name", i18n("New Server"));
  m_activeCategory.appendChild(newServer);
       
  // Create the ListViewItem
  KFTPBookmarkEditorItem *serv = 0L;
  if ( !m_activeItem || ( !m_activeItem->parent() && m_activeItem->m_type == BT_SERVER ) )
    serv = new KFTPBookmarkEditorItem(this, i18n("New Server"));
  else {
    if ( m_activeItem->m_type == BT_CATEGORY ) {
      serv = new KFTPBookmarkEditorItem(m_activeItem, i18n("New Server"));
      m_activeItem->setOpen(true);
    }
    else
      serv = new KFTPBookmarkEditorItem(m_activeItem->parent(), i18n("New Server"));
  }

  serv->m_type = BT_SERVER;
  serv->m_node = newServer;

  serv->setPixmap(0, loadSmallPixmap("ftp"));

  // Set the new ListViewItem as active
  setSelected(serv, true);
  m_activeItem = serv;
    
  // Display its properties
  m_activeNode = newServer;
  emit bookmarkNew(serv, newServer);
}

void KFTPBookmarkEditorListView::slotDuplicateAction()
{
  // Copy a bookmark
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  
  // Clone the node
  QDomNode server = item->m_node;
  QDomNode copy = server.cloneNode();
  
  // Select a new name
  QString newName = i18n("Copy of") + " " + server.toElement().attribute("name");
    
  // Append the copy to original's parent
  copy.toElement().setAttribute("name", newName);
  server.parentNode().appendChild(copy);
  
  // Create the ListViewItem
  KFTPBookmarkEditorItem *serv = 0L;
  if (item->parent())
    serv = new KFTPBookmarkEditorItem(item->parent(), newName);
  else
    serv = new KFTPBookmarkEditorItem(this, newName);
  
  serv->m_type = BT_SERVER;
  serv->m_node = copy;

  serv->setPixmap(0, loadSmallPixmap("ftp"));

  // Set the new ListViewItem as active
  setSelected(serv, true);
  m_activeItem = serv;
    
  // Display its properties
  m_activeNode = copy;
  emit bookmarkDuplicated(serv, copy);
}

void KFTPBookmarkEditorListView::slotDeleteAction()
{
  // Delete a server or a category        
  KFTPBookmarkEditorItem *item = static_cast<KFTPBookmarkEditorItem*>(selectedItems().at(0));
  
  // What do we have here ?
  if (item->m_type == BT_CATEGORY) {
    // Category
    if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove this category?")) == KMessageBox::Yes) {
      m_bookmarks->domDeleteCategory(item->m_node);
    } else {
      return;
    }
  } else {
    // Server
    if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove this server?")) == KMessageBox::Yes) {
      m_bookmarks->domDeleteServer(item->m_node);
    } else {
      return;
    }
  }
  
  if ( !m_autoUpdate )
    delete item;
  
  m_activeItem = 0L;
  emit bookmarkDeleted();
}

void KFTPBookmarkEditorListView::slotContextMenu(QListViewItem *item, const QPoint &pos, int)
{
  slotClicked(item);

  KPopupMenu *context = new KPopupMenu(this);
  
  if (item)
  {
    context->insertTitle(item->text(0));

    // Server or category
    if (static_cast<KFTPBookmarkEditorItem*>(item)->m_type == BT_CATEGORY) {
      m_newAction->plug(context);    
      m_deleteAction->plug(context);
      m_subCatAction->plug(context);
      m_renameAction->plug(context);
    } else {
      if ( m_editMenuItem )
        actionCollection()->action("bookmark_edit2")->plug(context);
      m_newAction->plug(context);    
      m_copyAction->plug(context);
      m_deleteAction->plug(context);
    }
  } else {
    // Nothing selected
    m_newAction->plug(context);    
    m_subCatAction->plug(context);
  }
  
  context->exec(pos);
  
  delete context;
}


#include "kftpbookmarkeditorlistview.moc"
