/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "kftpqueueeditor.h"
#include "kftpserverlineedit.h"
#include "kftpbookmarks.h"
#include "kftpqueueeditorlayout.h"

#include <klineedit.h>
#include <kpassdlg.h>
#include <kcombobox.h>
#include <klocale.h>

#include <qspinbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qtabwidget.h>

KFTPQueueEditor::KFTPQueueEditor(QWidget *parent, const char *name)
: KDialogBase(parent, name, true, "Edit queue", KDialogBase::Ok | KDialogBase::Cancel,
              KDialogBase::Ok, true)
{
  m_layout = new KFTPQueueEditorLayout(this);
  setMainWidget(m_layout);

  connect(m_layout->sourcePath, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->destPath, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->serverHost, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->serverUser, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->serverPass, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->fxpHost, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->fxpUser, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));
  connect(m_layout->fxpPass, SIGNAL(textChanged(const QString&)), this, SLOT(slotTextChanged()));

  connect(m_layout->serverName, SIGNAL(serverChanged(QDomNode)), this, SLOT(slotSourceServerChanged(QDomNode)));
  connect(m_layout->fxpName, SIGNAL(serverChanged(QDomNode)), this, SLOT(slotDestServerChanged(QDomNode)));

  connect(m_layout->transferType, SIGNAL(activated(int)), this, SLOT(slotTransferModeChanged(int)));

  setMaximumHeight(250);
  setInitialSize(QSize(500, 250));

  enableButtonOK(false);
}

void KFTPQueueEditor::resetTabs()
{
  m_layout->serverTab->setTabEnabled(m_layout->tab, false);
  m_layout->serverTab->setTabEnabled(m_layout->tab_2, false);
}

void KFTPQueueEditor::resetServerData()
{
  // Source
  m_layout->serverName->clear();
  m_layout->serverHost->setText("");
  m_layout->serverPort->setValue(21);
  m_layout->serverUser->setText("");
  m_layout->serverPass->erase();

  // Destination
  m_layout->fxpName->clear();
  m_layout->fxpHost->setText("");
  m_layout->fxpPort->setValue(21);
  m_layout->fxpUser->setText("");
  m_layout->fxpPass->erase();
}

void KFTPQueueEditor::slotTransferModeChanged(int index)
{
  if (m_lastTransferType == index)
    return;
  else
    m_lastTransferType = (KFTPQueue::TransferType) index;

  resetTabs();
  resetServerData();
  switch (index) {
    case 0: {
      // Download - source: remote dest: local
      m_layout->serverTab->setTabEnabled(m_layout->tab, true);
      m_layout->serverTab->showPage(m_layout->tab);
      break;
    }
    case 1: {
      // Upload - source: local dest: remote
      m_layout->serverTab->setTabEnabled(m_layout->tab_2, true);
      m_layout->serverTab->showPage(m_layout->tab_2);
      break;
    }
    case 2: {
      // FXP Download/Upload - source: remote dest: remote
      m_layout->serverTab->setTabEnabled(m_layout->tab, true);
      m_layout->serverTab->setTabEnabled(m_layout->tab_2, true);
      m_layout->serverTab->showPage(m_layout->tab);
      break;
    }
  }

  slotTextChanged();
}

bool KFTPQueueEditor::sourceIsValid()
{
  if (m_lastTransferType == 1) return true;

  if (m_layout->serverHost->text().stripWhiteSpace().isEmpty() || m_layout->serverUser->text().stripWhiteSpace().isEmpty() ||
      QString(m_layout->serverPass->password()).stripWhiteSpace().isEmpty())
    return false;
  else
    return true;
}

bool KFTPQueueEditor::destIsValid()
{
  if (m_lastTransferType == 0) return true;

  if (m_layout->fxpHost->text().stripWhiteSpace().isEmpty() || m_layout->fxpUser->text().stripWhiteSpace().isEmpty() ||
      QString(m_layout->fxpPass->password()).stripWhiteSpace().isEmpty())
    return false;
  else
    return true;
}

void KFTPQueueEditor::slotTextChanged()
{
  if (m_layout->sourcePath->text().stripWhiteSpace().isEmpty() || m_layout->destPath->text().stripWhiteSpace().isEmpty() ||
      m_layout->sourcePath->text().left(1) != "/" || m_layout->destPath->text().left(1) != "/" ||
      !sourceIsValid() || !destIsValid() )
    enableButtonOK(false);
  else
    enableButtonOK(true);
}

void KFTPQueueEditor::setData(KFTPQueue::Transfer *transfer)
{
  KURL sUrl, dUrl;
  m_layout->sourcePath->setText(transfer->getSourceUrl().path());
  m_layout->destPath->setText(transfer->getDestUrl().path());

  // Source
  sUrl = transfer->getSourceUrl();
  if (!sUrl.isLocalFile()) {
    if (FTPBookmarks->domServerExistsByHost(sUrl.host()))
      m_layout->serverName->setCurrentServer(FTPBookmarks->domFindServerByHost(sUrl.host()));
    else
      m_layout->serverName->clear();
    m_layout->serverHost->setText(sUrl.host());
    m_layout->serverPort->setValue(sUrl.port());
    m_layout->serverUser->setText(sUrl.user());
    m_layout->serverPass->insert(sUrl.pass());
  } else {
    m_layout->serverTab->setTabEnabled(m_layout->tab, false);
  }

  // Destination
  dUrl = transfer->getDestUrl();
  if (!dUrl.isLocalFile()) {
    if (FTPBookmarks->domServerExistsByHost(dUrl.host()))
      m_layout->fxpName->setCurrentServer(FTPBookmarks->domFindServerByHost(dUrl.host()));
    else
      m_layout->fxpName->clear();
    m_layout->fxpHost->setText(dUrl.host());
    m_layout->fxpPort->setValue(dUrl.port());
    m_layout->fxpUser->setText(dUrl.user());
    m_layout->fxpPass->insert(dUrl.pass());
  } else {
    m_layout->serverTab->setTabEnabled(m_layout->tab_2, false);
  }

  // Transfer type
  m_lastTransferType = transfer->getTransferType();
  m_layout->transferType->setCurrentItem(m_lastTransferType);

  m_transfer = transfer;
}

void KFTPQueueEditor::saveData()
{
  KURL sUrl, dUrl;

  if (m_lastTransferType != 1) {
    sUrl.setProtocol("ftp");
    sUrl.setHost(m_layout->serverHost->text());
    sUrl.setPort(m_layout->serverPort->value());
    sUrl.setUser(m_layout->serverUser->text());
    sUrl.setPass(m_layout->serverPass->password());
  } else {
    sUrl.setProtocol("file");
  }
  sUrl.setPath(m_layout->sourcePath->text());

  if (m_lastTransferType != 0) {
    dUrl.setProtocol("ftp");
    dUrl.setHost(m_layout->fxpHost->text());
    dUrl.setPort(m_layout->fxpPort->value());
    dUrl.setUser(m_layout->fxpUser->text());
    dUrl.setPass(m_layout->fxpPass->password());
  } else {
    dUrl.setProtocol("file");
  }
  dUrl.setPath(m_layout->destPath->text());

  m_transfer->setSourceUrl(sUrl);
  m_transfer->setDestUrl(dUrl);
  m_transfer->setTransferType(m_lastTransferType);
}

void KFTPQueueEditor::slotSourceServerChanged(QDomNode server)
{
  m_layout->serverHost->setText(FTPBookmarks->domGetProperty(server, "host"));
  m_layout->serverPort->setValue(FTPBookmarks->domGetProperty(server, "port").toInt());
  m_layout->serverUser->setText(FTPBookmarks->domGetProperty(server, "downuser"));
  m_layout->serverPass->erase();
  m_layout->serverPass->insert(FTPBookmarks->domGetProperty(server, "downpass"));
}

void KFTPQueueEditor::slotDestServerChanged(QDomNode server)
{
  m_layout->fxpHost->setText(FTPBookmarks->domGetProperty(server, "host"));
  m_layout->fxpPort->setValue(FTPBookmarks->domGetProperty(server, "port").toInt());
  m_layout->fxpUser->setText(FTPBookmarks->domGetProperty(server, "downuser"));
  m_layout->fxpPass->erase();
  m_layout->fxpPass->insert(FTPBookmarks->domGetProperty(server, "downpass"));
}
