/*
 * This file is part of the KFTPGrabber project
 *
 * Copyright (C) 2003-2004 by the KFTPGrabber developers
 * Copyright (C) 2003-2004 Jernej Kos <kostko@jweb-network.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * is provided AS IS, WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, and
 * NON-INFRINGEMENT.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include "misc.h"

#include "kftpqueueview.h"
#include "kftpqueue.h"
#include "kftpapi.h"
#include "kftpqueueeditor.h"

#include <kpopupmenu.h>
#include <kmessagebox.h>
#include <kiconloader.h>
#include <kmimetype.h>
#include <klocale.h>
#include <kio/job.h>

using namespace KFTPGrabberBase;

KFTPQueueViewItem::KFTPQueueViewItem(KFTPQueue::Transfer *transfer, QListView *parent)
  : QObject(parent), QListViewItem(parent)
{
  m_queueTransfer = transfer;

  connect(transfer, SIGNAL(transferUpdate()), this, SLOT(slotUpdateItemRequested()));
  refresh();
}

KFTPQueueViewItem::KFTPQueueViewItem(KFTPQueue::Transfer *transfer, QListViewItem *parent)
  : QObject(), QListViewItem(parent)
{
  m_queueTransfer = transfer;

  connect(transfer, SIGNAL(transferUpdate()), this, SLOT(slotUpdateItemRequested()));
  refresh();
}

void KFTPQueueViewItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int alignment)
{
    QColorGroup _cg( cg );
    QColor c = _cg.text();

    QColor n_color;
    if (m_queueTransfer) {
      switch (m_queueTransfer->getStatus()) {
        case KFTPQueue::Transfer::Running: n_color.setRgb(255, 0, 0); break;
        case KFTPQueue::Transfer::Connecting: n_color.setRgb(0, 0, 255); break;
        default: break;
      }
    }
    
    if (n_color.isValid())
      _cg.setColor(QColorGroup::Text, n_color);

    QListViewItem::paintCell( p, _cg, column, width, alignment );
    _cg.setColor( QColorGroup::Text, c );
}

void KFTPQueueViewItem::refresh()
{
  if (!m_queueTransfer)
    return;

  KFTPQueue::Transfer *transfer = m_queueTransfer;

  // Speed
  QString speed;
  if ((!transfer->isDir() || !isOpen()) && transfer->isRunning()) {
    speed.sprintf("%ld KB/s", (transfer->getSpeed() / 1024));
    if (transfer->getSpeed() > 1024*1024)
      speed.sprintf("%ld MB/s", (transfer->getSpeed() / 1024) / 1024);
  
    if (transfer->getSpeed() == 0 && transfer->getTransferType() != KFTPQueue::FXP) {
      speed = i18n("stalled");
    }
  }

  // ETA
  QString eta;

  if (transfer->isRunning() && transfer->getSpeed() > 0) {
    eta = KIO::calculateRemaining(transfer->getActualSize(),
                                  0,
                                  transfer->getSpeed()).toString("hh:mm:ss");
  } else {
    eta = "";
  }

  // Set the columns
  setText(0, transfer->getSourceUrl().fileName());
  setText(1, KIO::convertSize(transfer->getActualSize()));
  setText(2, transfer->getSourceUrl().prettyURL());
  setText(3, transfer->getDestUrl().prettyURL());
  setText(5, speed);
  setText(6, eta);
  setText(7, QString::number(transfer->getId()));

  // Progress
  int progress;
  int r_progress;

  if (transfer->getSize() == 0)
    progress = 0;
  else
    progress = transfer->getCompleted()*100/transfer->getSize();
  
  // We can't get progress for FXP transfers
  if (transfer->getTransferType() == KFTPQueue::FXP)
    progress = -1;

  if (transfer->getResumed() == 0)
    r_progress = 0;
  else
    r_progress = transfer->getResumed()*100/transfer->getSize();

  if (m_queueTransfer->getStatus() == KFTPQueue::Transfer::Connecting) {
    // Transfer is not yet connected,
    setText(4, i18n("Connecting..."));
  } else if (progress >= 0) {
    setPixmap(4, createProgressPixmap(progress, r_progress));

    QString progressText;
    progressText.sprintf("%d %%", progress);
    setText(4, progressText);
  } else if (transfer->isRunning() && transfer->getTransferType() == KFTPQueue::FXP) {
    setText(4, i18n("Unknown (FXP)"));
    setText(5, i18n("running"));
  }

  // Transfer type pixmap
  setPixmap(2, loadSmallPixmap(transfer->getSourceUrl().isLocalFile() ? "system" : "network"));
  setPixmap(3, loadSmallPixmap(transfer->getDestUrl().isLocalFile() ? "system" : "network"));

  // Icon
  QString iconText;
  if (transfer->isDir()) {
    iconText = "folder";
  } else {
    KMimeType::Ptr theType = KMimeType::findByURL("/" + transfer->getSourceUrl().path(), 0, false, true);
    iconText = theType->icon(QString::null, false);
  }

  setPixmap(0, loadSmallPixmap(iconText));
}

void KFTPQueueViewItem::slotUpdateItemRequested()
{
  refresh();
}

void KFTPQueueViewItem::setTransfer(KFTPQueue::Transfer *transfer)
{
  m_queueTransfer = transfer;
}

KFTPQueue::Transfer *KFTPQueueViewItem::getTransfer()
{
  return m_queueTransfer;
}

int KFTPQueueViewItem::compare(QListViewItem *i, int col, bool) const
{
  if (!m_queueTransfer) return 1;
  
  KFTPQueue::Transfer *f_transfer = static_cast<KFTPQueueViewItem*>(i)->getTransfer();

  if (f_transfer && f_transfer->isChild() && f_transfer->parent() == m_queueTransfer->parent()) {
    KFTPQueue::TransferDir *f_parent = static_cast<KFTPQueue::TransferDir*>(f_transfer->getTransferParent());

    if (f_parent->getChildrenList().findRef(f_transfer) <
        f_parent->getChildrenList().findRef(m_queueTransfer))
      return 1;
    else
      return -1;
  }
  
  if (FTPQueueManager->getQueueTransferList()->findRef(f_transfer) <
      FTPQueueManager->getQueueTransferList()->findRef(m_queueTransfer))
    return 1;
  else
    return -1;
}

/*
                                                           CLASS KFTPQUEUEVIEW
*/

KActionCollection *KFTPQueueView::actionCollection()
{
  return KFTPAPI::getInstance()->mainWindow()->actionCollection();
}

KFTPQueueView::KFTPQueueView(QWidget *parent, const char *name)
 : KFTPListView(parent, name)
{
  // Create the columns
  addColumn(i18n("Name"), 150);
  addColumn(i18n("Size"), 75);
  addColumn(i18n("Source"), 250);
  addColumn(i18n("Destination"), 250);
  addColumn(i18n("Progress"), 140);
  addColumn(i18n("Speed"), 70);
  addColumn(i18n("ETA"), 80);

  // Text when there is nothing queued
  setEmptyListText(i18n("You don't have any files in the queue."));

  // Multi-select
  setSelectionModeExt(FileManager);
  setAllColumnsShowFocus(true);
  setRootIsDecorated(true);
  setEnableContext(true);

  // The signals
  connect(this, SIGNAL(contextMenu(KListView*, QListViewItem*, const QPoint&)), this,
          SLOT(contextMenuRequested(KListView*, QListViewItem*, const QPoint&)));

  // Let us be up-to-date
  connect(FTPQueueManager, SIGNAL(queueUpdate()), this, SLOT(slotQueueUpdate()));
  connect(FTPQueueManager, SIGNAL(transferRemoved(long)), this, SLOT(slotTransferRemoved(long)));
  connect(FTPQueueManager, SIGNAL(newTransfer(KFTPQueue::Transfer*)), this, SLOT(slotTransferAdded(KFTPQueue::Transfer*)));
  
  // Create the context menu actions
  initActions();
}

KFTPQueueView::~KFTPQueueView()
{
}

void KFTPQueueView::setEnableContext(bool value)
{
  // Enable/disable context menu
  m_enableContext = value;
}

void KFTPQueueView::initActions()
{
  // Create all the actions
  m_launchAction = new KAction(i18n("&Start transfer"), "launch", KShortcut(), this, SLOT(slotLaunch()), actionCollection(), "launch");
  m_abortAction = new KAction(i18n("&Abort transfer"), KShortcut(), this, SLOT(slotAbort()), actionCollection(), "abort");
  m_removeAction = new KAction(i18n("&Remove"), "remove", KShortcut(), this, SLOT(slotRemove()), actionCollection(), "remove");
  m_removeAllAction = new KAction(i18n("Remove &all"), KShortcut(), this, SLOT(slotRemoveAll()), actionCollection(), "removeAll");
  m_moveUpAction = new KAction(i18n("Move &up"), "up", KShortcut(), this, SLOT(slotMoveUp()), actionCollection(), "moveUp");
  m_moveDownAction = new KAction(i18n("Move &down"), "down", KShortcut(), this, SLOT(slotMoveDown()), actionCollection(), "moveDown");
  m_editAction = new KAction(i18n("&Change transfer info"), KShortcut(), this, SLOT(slotEdit()), actionCollection(), "changeTransfer");
  
  m_removeAction->setEnabled(false);
}

void KFTPQueueView::updateActions(KFTPQueueViewItem *item)
{
  // Update all the actions
  if (!item)
    return;
    
  bool parentRunning = false;
  if (item->getTransfer()->isChild())
    parentRunning = item->getTransfer()->getTransferParent()->isRunning();
  
  m_launchAction->setEnabled(!item->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning());
  m_abortAction->setEnabled(item->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning());
  m_removeAction->setEnabled(!item->getTransfer()->isRunning() && !FTPQueueManager->queueProcRunning() && !parentRunning);
  m_removeAllAction->setEnabled(!FTPQueueManager->queueProcRunning());
  m_moveUpAction->setEnabled(FTPQueueManager->canBeMovedUp(item->getTransfer()->getId()));
  m_moveDownAction->setEnabled(FTPQueueManager->canBeMovedDown(item->getTransfer()->getId()));
  m_editAction->setEnabled(!item->getTransfer()->isRunning() && !item->getTransfer()->isChild() && !item->getTransfer()->isDir());
}

void KFTPQueueView::slotTransferAdded(KFTPQueue::Transfer *transfer)
{
  if (transfer->isChild()) {
    // This transfer should be inserted under some other transfer
    QListViewItem *parent = findItem(QString::number(transfer->getTransferParent()->getId()), 7);
    
    if (parent) {
      new KFTPQueueViewItem(transfer, parent);
    }
  } else {
    new KFTPQueueViewItem(transfer, this);
  }
}

void KFTPQueueView::slotTransferRemoved(long id)
{
  // Delete the transfer
  QListViewItem *i = findItem(QString::number(id), 7);
  
  if (i)
    delete i;
}

void KFTPQueueView::slotQueueUpdate()
{
  /* Do a resort */
  sort();
}

void KFTPQueueView::contextMenuRequested(KListView* l, QListViewItem* i, const QPoint& p)
{
  if (!i || !m_enableContext) return;

  KPopupMenu *contextMenu = new KPopupMenu(this);
  
  // Populate context menu
  m_launchAction->plug(contextMenu);
  m_abortAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_removeAction->plug(contextMenu);
  m_removeAllAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_moveUpAction->plug(contextMenu);
  m_moveDownAction->plug(contextMenu);
  contextMenu->insertSeparator();
  m_editAction->plug(contextMenu);
  
  // Update the actions
  m_curSelection = static_cast<KFTPQueueViewItem*>(i);
  updateActions(m_curSelection);
  
  // Show the context menu
  contextMenu->exec(p);
}

void KFTPQueueView::slotLaunch()
{
  m_curSelection->getTransfer()->execute();
}

void KFTPQueueView::slotAbort()
{
  m_curSelection->getTransfer()->abort();
}

void KFTPQueueView::slotRemove()
{
  if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove queued file(s)?")) == KMessageBox::Yes) {
    FTPQueueManager->setEmitUpdate(false);
    
    for (QListViewItem *item = selectedItems().first(); item; item = selectedItems().next()) {
      KFTPQueueViewItem *s_item = static_cast<KFTPQueueViewItem*>(item);
      FTPQueueManager->removeTransfer(s_item->getTransfer());
    }
    
    FTPQueueManager->setEmitUpdate(true);
    FTPQueueManager->doEmitUpdate();
  }
}

void KFTPQueueView::slotRemoveAll()
{
  if (KMessageBox::questionYesNo(this, i18n("Are you sure you want to remove ALL queued files?")) == KMessageBox::Yes) {
    FTPQueueManager->clearQueue();
  }
}

void KFTPQueueView::slotMoveUp()
{
  FTPQueueManager->moveTransferUp(m_curSelection->getTransfer()->getId());
}

void KFTPQueueView::slotMoveDown()
{
  FTPQueueManager->moveTransferDown(m_curSelection->getTransfer()->getId());
}

void KFTPQueueView::slotEdit()
{
  KFTPQueueEditor *editor = new KFTPQueueEditor(this);
  editor->setData(m_curSelection->getTransfer());
  
  // Show the queue editor
  if (editor->exec() == QDialog::Accepted) {
    editor->saveData();
    FTPQueueManager->doEmitUpdate();
    m_curSelection->refresh();
  }
}

